// C_Control

#ifndef _CONTROL_H_
#define _CONTROL_H_

#define MAX_BUTTONS 10

#include "common.h"
#include "timer.h"
#include "input_defines.h"

#include <deque>

#ifdef WIN32

 #include <initguid.h>
 #define DIRECTINPUT_VERSION 0x0800
 #include <dinput.h>

#else

 #ifdef __APPLE__
  #include "fssimplewindow.h"
  #include "ysjoyreader.h"
 #else
  #include <X11/X.h>
  #include <X11/Xlib.h>
  #include <X11/Xutil.h>
  #include <X11/keysym.h>

  #include <fcntl.h>
  #include <linux/joystick.h>

  #define LINUX_JOYSTICK_DEVNAME "/dev/input/js" //i_iPadDevice in constructor is added at the end

 #endif
#endif

class C_Control {
public:
#ifdef WIN32
	C_Control(HWND i_hWnd, int i_iPadDevice = 0, bool i_bExclusiveMouse = false);
	void XLateMouseEvent(UINT iMessage, WPARAM wParam, LPARAM lParam); //must be run to make mouse event work
	void ReleaseMouse(); //to be done on WM_ENTERSIZEMOVE and WM_ENTERMENULOOP if mouse exclusive
	void AcquireMouse(); //to be done on WM_EXITSIZEMOVE and WM_EXITMENULOOP
#else
	C_Control(int i_iPadDevice = 0, bool i_bExclusiveMouse = false);
 #ifdef __APPLE__
	void PollInput(); //must be run to make keyboard and mouse work
 #else
	void PollInput(XEvent *pKeyMouseEvent); //must be run to make keyboard and mouse work

	void SetDpyWnd(Display *i_pDisplay, Window i_wndRoot); //to be done if mouse exclusive
 #endif
#endif
	~C_Control();
	static C_Control *GetControlObject() {return s_pclTheControl;};

	void Update();

	//keyboard
	int IsPressed(int i_iKey);
	int IsPressedOnce(int i_iKey);

	void SetRate(int i_iMsRate, int i_iMsDelay)
		{m_iRepeatRate = i_iMsRate; m_iRepeatDelay = i_iMsDelay;};
	int IsPressedRepeat(int i_iKey);

	int GetAscii(int i_iDiIndex);
	int GetAscii2(int i_iDiIndex);

	//mouse
	int GetMouseEvent(int *o_iX, int *o_iY, int *o_iExtraVal);
	void GetMouseMovement(int *o_iX, int *o_iY);             //pixels since last
	void GetMouseMovementSmooth(double *o_dX, double *o_dY); //pixels/frame 
	void GetMousePos(int *o_iX, int *o_iY);
	void GetMouseButton(int *o_iLeft, int *o_iRight, int *o_iMiddle); //mouse button state
	bool MouseIsPressed(int i_iButton); //can be 0, 1, 2 (Left, Right, Middle)
	bool MouseIsPressedOnce(int i_iButton);

	//joystick
	void SetJoystickKeyMapping(S_JoystickMapping *i_pstMapping);
	void GetJoystickState(S_JoystickState *o_pstState);        //for joystick operation
	void GetJoystickStateDigital(S_JoystickState *o_pstState); //for gamepad operation
private:
	C_Timer             *m_pclTimer;
	static C_Control    *s_pclTheControl;
	bool                m_bExclusiveMouse;

#ifdef WIN32
	LPDIRECTINPUT        m_pclDI;
	//keyboard
	LPDIRECTINPUTDEVICE  m_pclDIK;
	HKL                  m_hKeyLayout; //getascii2
	//mouse
	LPDIRECTINPUTDEVICE  m_pclDIM;
	DIMOUSESTATE         m_stMBuf;
	//pad
	LPDIRECTINPUTDEVICE8 m_pclDIP;
	DIJOYSTATE           m_stPBuf;

	bool ReadJoystickState();
#else
	S_JoystickState      m_stJStateLast; //extra copy of joystick state to get working keyboard mapping,
		//since we are dealing with key events and not reading the keyboard state
	void ReadJoystickState();

	//mouse in exclusive mode
	int m_iMStateXBufLast2, m_iMStateYBufLast2;
 #ifdef __APPLE__
	YsJoyReader m_clJoystick[1];
	bool m_bJoystickValid;
 #else
	int                  m_hJoystick;
	int ReadJoystickEvent(S_JSEvent *pstJSE);
	void XLateKeyEvent(XEvent *pKeyEvent);
	void XLateMouseEvent(XEvent *pMouseEvent);

	//x system
	Display *m_pDisplay;
	Window m_wndRoot;
 #endif
#endif

	//keyboard
	void InitAsciiLookup();
	void UpdateKeys();
	int                  ASCIILOOKUP[256];
	int                  ASCIILOOKUPSHIFTED[256];
	char                 m_acKBuf[256];

	uint32_t             m_iRepeatRate;
	uint32_t             m_iRepeatDelay;
	uint32_t             m_aiRepeatCount[256];
	double               m_adRepeat[256];
	char                 m_acPressed[256];

	//mouse
	double               m_dMouseTimer;
	double               m_dCurX,  m_dCurY;
	double               m_dStepX, m_dStepY;
	char                 m_iMButtonStateBuf[3];
	char                 m_iMButtonState[3];
	int                  m_iMStateXBuf, m_iMStateYBuf;
	int                  m_iMStateXBufLast, m_iMStateYBufLast; //mouse pos in screen coords
	std::deque<S_MouseEvent> *m_pclMouseEvents;

	//pad
	int                  m_iNumButtons;
	S_JoystickMapping    *m_pstPMapping;
	S_JoystickState      m_stJState;
};

#endif
