// C_Map (for mapeditor)

//class to handle a map
// - save to disk
// - load from disk
// - draw
// - functions useful for the editor

#include "e_tilemap.h"
#include "fileresource.h"
#include "global.h"

#include <stdio.h>
//constructor to create a map from an existing file
//constructor to create an empty map
C_Map::C_Map(char *i_szGridpath, S_MapHeader *i_pstMapInfo, S_Rect *i_pstDrawarea, bool i_bNew)
{
	int i;
	m_szGridpath[0] = 0;
	if(i_szGridpath) strcpy(m_szGridpath, i_szGridpath);
	memcpy(&m_stHeader, i_pstMapInfo, sizeof(m_stHeader));
	m_iWidth = m_stHeader.iWidth;
	m_iHeight = m_stHeader.iHeight;
	m_iTileSize = m_stHeader.iTileSize;

	//allocate memory for map, and set defaults
	m_iLastX = m_iLastY = 0;
	m_stDrawarea.x = 0;
	m_stDrawarea.y = 0;
	m_stDrawarea.width = 0;
	m_stDrawarea.height = 0;
	if(i_pstDrawarea) memcpy(&m_stDrawarea, i_pstDrawarea, sizeof(m_stDrawarea));
	m_pGrid = new char*[m_iHeight];
	for(i=0; i<m_iHeight; i++) {
		m_pGrid[i] = new char[m_iWidth];
	}

	if(!i_bNew) {
		Load();
		LoadTileSet();
	} else {
		Clear();
		LoadTileSet();
	}

	//save the area entered by the user if wrap around mode should change later
	m_iDrawareaWidthSave = m_stDrawarea.width;
	m_iDrawareaHeightSave = m_stDrawarea.height;
	SetMapWrapping(false, false);

	//get graphwrapper interface
	m_pclGraph = C_GraphWrapperGL::GetGraphWrapperObject();
}

//destructor
C_Map::~C_Map()
{
	int i;

	for(i=0; i<m_iHeight; i++) {
		delete[] m_pGrid[i];
	}
	delete[] m_pGrid;

	delete[] m_stTileRects;
	delete m_pclTilesImg;
	delete m_pclTilesImg_Sysmem;
}

//set the draw area
void C_Map::SetScreenRect(S_Rect *i_pstDrawarea)
{
	m_stDrawarea = *i_pstDrawarea;
	m_iDrawareaWidthSave = m_stDrawarea.width;
	m_iDrawareaHeightSave = m_stDrawarea.height;
	SetMapWrapping(m_iWrappX, m_iWrappY);
}

//get tileset
void C_Map::LoadTileSet()
{
	int i, j;
	bool bResult;

	m_pclTilesImg = new C_Image(m_stHeader.szTileset, C_Global::szResourceFile, &bResult);
	if(!bResult) { //not in resource, try normal file
		char szTilename[MAX_PATH];
		sprintf(szTilename, "%s/%s", C_Global::szDataPath, m_stHeader.szTileset);
		delete m_pclTilesImg;
		m_pclTilesImg = new C_Image(szTilename, NULL, &bResult);
		if(!bResult) {
			//default to ts_alien.tga
			delete m_pclTilesImg; m_pclTilesImg = NULL;
			m_pclTilesImg = new C_Image("ts_alien.tga", C_Global::szResourceFile, &bResult);
			if(!bResult) {
				delete m_pclTilesImg; m_pclTilesImg = NULL;
				return;
			}
		}
	}
	m_pclTilesImg_Sysmem = new C_Image(m_pclTilesImg);

	//analyse tileset (get rectangles)
	int iWidth, iHeight, iNumX, iNumY, iTileNum;
	m_pclTilesImg->GetInfo(&iWidth, &iHeight, NULL);
	iNumX = iWidth / m_iTileSize;
	iNumY = iHeight / m_iTileSize;
	m_iNumTiles = iNumX * iNumY;
	m_stTileRects = new S_Rect[m_iNumTiles];

	iTileNum = 0;
	for(j=0; j<iNumY; j++) {
		for(i=0; i<iNumX; i++) {
			m_stTileRects[iTileNum].x = i * m_iTileSize;
			m_stTileRects[iTileNum].y = j * m_iTileSize;
			m_stTileRects[iTileNum].width = m_iTileSize;
			m_stTileRects[iTileNum].height = m_iTileSize;
			iTileNum++;
		}
	}
}

//clear all elements in the map to index 0
void C_Map::Clear()
{
	int x, y;

	for(y=0; y<m_iHeight; y++) {
		for(x=0; x<m_iWidth; x++) {
			m_pGrid[y][x] = 0;
		}
	}
}

//crops one tile from the given side
//0=left, 1=right, 2=top, 3=bottom
void C_Map::Crop(int iSide)
{
	int x, y, w, h;

	int xmod = 0;
	int ymod = 0;
	w = m_iWidth;
	h = m_iHeight;
	switch(iSide) {
		case 0: xmod = 1; w--; break;
		case 1: xmod = 0; w--; break;
		case 2: ymod = 1; h--; break;
		case 3: ymod = 0; h--; break;
	}

	//create new grid
	char **pGrid = new char*[h];
	for(y=0; y<h; y++) {
		pGrid[y] = new char[w];
	}

	//copy selected part
	for(y=0; y<h; y++) {
		for(x=0; x<w; x++) {
			pGrid[y][x] = m_pGrid[y+ymod][x+xmod];
		}
	}

	//replace grid
	for(y=0; y<m_iHeight; y++) {
		delete[] m_pGrid[y];
	}
	delete[] m_pGrid;
	m_pGrid = pGrid;
	m_stHeader.iHeight = m_iHeight = h;
	m_stHeader.iWidth  = m_iWidth  = w;
	m_iLastX = m_iLastY = 0;
	SetMapWrapping(m_iWrappX, m_iWrappY);
}

//extend one tile from the given side
//0=left, 1=right, 2=top, 3=bottom
void C_Map::Extend(int iSide)
{
	int x, y, w, h;

	int xmod = 0;
	int ymod = 0;
	w = m_iWidth;
	h = m_iHeight;
	switch(iSide) {
		case 0: xmod = 1; w++; break;
		case 1: xmod = 0; w++; break;
		case 2: ymod = 1; h++; break;
		case 3: ymod = 0; h++; break;
	}

	//create new grid
	char **pGrid = new char*[h];
	for(y=0; y<h; y++) {
		pGrid[y] = new char[w];
	}
	//set default
	for(y=0; y<h; y++) {
		for(x=0; x<w; x++) {
			pGrid[y][x] = 0;
		}
	}

	//copy from old map
	for(y=0; y<m_iHeight; y++) {
		for(x=0; x<m_iWidth; x++) {
			pGrid[y + ymod][x + xmod] = m_pGrid[y][x];
		}
	}

	//replace grid
	for(y=0; y<m_iHeight; y++) {
		delete[] m_pGrid[y];
	}
	delete[] m_pGrid;
	m_pGrid = pGrid;
	m_stHeader.iHeight = m_iHeight = h;
	m_stHeader.iWidth = m_iWidth = w;
	m_iLastX = m_iLastY = 0;
	SetMapWrapping(m_iWrappX, m_iWrappY);
}

//checks if the values goes outside the limits, and take actions
bool C_Map::WrapAroundTest(int *o_iX, int *o_iY, int i_iWidth, int i_iHeight)
{
	bool bResult = true;

	if(*o_iX>=i_iWidth) {
		if(m_iWrappX) *o_iX %= i_iWidth;
		else {
			*o_iX = i_iWidth;
			bResult = false;
		}
	}
	if(*o_iY>=i_iHeight) {
		if(m_iWrappY) *o_iY %= i_iHeight;
		else {
			*o_iY = i_iHeight;
			bResult = false;
		}
	}
	if(*o_iX<0) {
		if(m_iWrappX) *o_iX = i_iWidth + (*o_iX % i_iWidth);
		else {
			*o_iX = 0;
			bResult = false;
		}
	}
	if(*o_iY<0) {
		if(m_iWrappY) *o_iY = i_iHeight + (*o_iY % i_iHeight);
		else {
			*o_iY = 0;
			bResult = false;
		}
	}

	return bResult;
}

//set if the map shall stop or wrap around when scrolling past the borders
void C_Map::SetMapWrapping(bool i_bHorizontal, bool i_bVertical)
{
	m_iWrappX = i_bHorizontal;
	m_iWrappY = i_bVertical;

	m_stDrawarea.width = m_iDrawareaWidthSave;
	m_stDrawarea.height = m_iDrawareaHeightSave;
	//if the show area is larger than the map, and wrap around
	// shall not be used, we need to limit the showarea
	if(!m_iWrappX && m_iWidth*m_iTileSize< m_iDrawareaWidthSave) m_stDrawarea.width = m_iWidth*m_iTileSize;
	if(!m_iWrappY && m_iHeight*m_iTileSize< m_iDrawareaHeightSave) m_stDrawarea.height = m_iHeight*m_iTileSize;

	//calc maximum upper left coordinate we can scroll to
	// without getting outside the map (used if no wrap around)
	m_iPixMaxX = m_stDrawarea.x + m_stDrawarea.width;
	m_iPixMaxY = m_stDrawarea.y + m_stDrawarea.height;
	m_iScrollMaxX = m_iWidth*m_iTileSize - m_stDrawarea.width;
	m_iScrollMaxY = m_iHeight*m_iTileSize - m_stDrawarea.height;
	if(m_iScrollMaxX < 0) m_iScrollMaxX = 0;
	if(m_iScrollMaxY < 0) m_iScrollMaxY = 0;
}

//checks if the given pos is within the map
bool C_Map::IsValidPos(int i_iMapTileX, int i_iMapTileY)
{
	return ((i_iMapTileX < m_iWidth) && (i_iMapTileY < m_iHeight) && (i_iMapTileX >= 0) && (i_iMapTileY >= 0));
}

//sets a mapelement to the given value
void C_Map::SetElement(int i_iMapTileX, int i_iMapTileY, char i_cVal)
{
	if(!IsValidPos(i_iMapTileX, i_iMapTileY)) return;
	m_pGrid[i_iMapTileY][i_iMapTileX] = i_cVal;
}

//gets a mapelement from the given pos
int C_Map::GetElement(int i_iMapTileX, int i_iMapTileY)
{
	if(!IsValidPos(i_iMapTileX, i_iMapTileY)) return -1;
	return m_pGrid[i_iMapTileY][i_iMapTileX];
}

//sets a mapelement to the given value
void C_Map::SetElementAbsolute(int i_iMapTileX, int i_iMapTileY, char i_cVal)
{
	if(!WrapAroundTest(&i_iMapTileX, &i_iMapTileY, m_iWidth, m_iHeight)) return;

	m_pGrid[i_iMapTileY][i_iMapTileX] = i_cVal;
}

//gets a mapelement from the given pos
int C_Map::GetElementAbsolute(int i_iMapTileX, int i_iMapTileY)
{
	if(!WrapAroundTest(&i_iMapTileX, &i_iMapTileY, m_iWidth, m_iHeight)) return -1;

	return m_pGrid[i_iMapTileY][i_iMapTileX];
}

//moves the showpoint of the map relative to the last
// position and draws the result
void C_Map::MoveRelative(int i_iX, int i_iY)
{
	MoveAbsolute(m_iLastX + i_iX, m_iLastY + i_iY);
}

//moves the showpoint of the map to a new position,
// given in world pixel coordinates
void C_Map::MoveAbsolute(int i_iX, int i_iY)
{
	int iCmpx = m_iWrappX ? m_iWidth*m_iTileSize  : m_iScrollMaxX;
	int iCmpy = m_iWrappY ? m_iHeight*m_iTileSize : m_iScrollMaxY;
	WrapAroundTest(&i_iX, &i_iY, iCmpx, iCmpy);

	m_iLastX = i_iX; m_iLastY = i_iY;
}

//draws all the currently visible tiles in the map on screen
void C_Map::Draw()
{
	int iTempX, iTempY;
	int iBaseTileX, iBaseOffsetX;
	int iTileX, iTileY;
	int iOffsetX, iOffsetY;
	int iScreenX = m_stDrawarea.x, iScreenY = m_stDrawarea.y;
	S_Rect stRect;
	S_Rect *pstRect;

	iBaseTileX = iTileX = m_iLastX / m_iTileSize;
	iBaseOffsetX = iOffsetX = m_iLastX % m_iTileSize;
	iTileY = m_iLastY / m_iTileSize;
	iOffsetY = m_iLastY % m_iTileSize;
	do {
		do {
			pstRect = &m_stTileRects[(int)m_pGrid[iTileY][iTileX]];
			stRect.x = pstRect->x + iOffsetX;
			stRect.y = pstRect->y + iOffsetY;
			iTempX = m_iPixMaxX - iScreenX;
			iTempY = m_iPixMaxY - iScreenY;
			stRect.width = (m_iTileSize - iOffsetX)<iTempX ? (m_iTileSize - iOffsetX):iTempX;
			stRect.height = (m_iTileSize - iOffsetY)<iTempY ? (m_iTileSize - iOffsetY):iTempY;

			m_pclGraph->Blt(m_pclTilesImg, &stRect, iScreenX, iScreenY);

			iTileX++;
			if(m_iWrappX) iTileX %= m_iWidth;
			iScreenX += stRect.width;
			iOffsetX = 0;
		} while (iScreenX<m_iPixMaxX && iTileX<m_iWidth);
		iOffsetX = iBaseOffsetX;
		iOffsetY = 0;
		iTileX = iBaseTileX;
		iTileY++;
		if(m_iWrappY) iTileY %= m_iHeight;

		iScreenX = m_stDrawarea.x;
		iScreenY += stRect.height;
	} while (iScreenY<m_iPixMaxY && iTileY<m_iHeight);
}

//saves the map grid data to the file associated with the map-obj
bool C_Map::Save()
{
	int i;
	FILE *pstFile;

	char szMapname[MAX_PATH];
	sprintf(szMapname, "%s/%s", m_szGridpath, m_stHeader.szGridfile);

	pstFile = fopen(szMapname, "wb");
	if(!pstFile) return false;
	for(i=0; i<m_iHeight; i++) {
		fwrite(m_pGrid[i], m_iWidth*sizeof(**m_pGrid), 1, pstFile);
	}
	fclose(pstFile);

	return true;
}

//loads the map grid data from the file associated with the map-obj
bool C_Map::Load(S_MapHeader *i_pstMapInfo)
{
	int i;

	//recreate grid
	for(i=0; i<m_iHeight; i++) {
		delete[] m_pGrid[i];
	}
	delete[] m_pGrid;
	if(i_pstMapInfo) {
		m_stHeader = *i_pstMapInfo;
		m_iWidth  = m_stHeader.iWidth;
		m_iHeight = m_stHeader.iHeight;
		m_iLastX = m_iLastY = 0;
		SetMapWrapping(m_iWrappX, m_iWrappY);
	}
	m_pGrid = new char*[m_iHeight];
	for(i=0; i<m_iHeight; i++) {
		m_pGrid[i] = new char[m_iWidth];
	}

	//load (always normal file in editor, not resource.dat)
	char szMapname[MAX_PATH];
	sprintf(szMapname, "%s/%s", m_szGridpath, m_stHeader.szGridfile);
	C_Resource *pclResource = new C_Resource();

	bool bResult = pclResource->SetFilename(szMapname);
	for(i=0; i<m_iHeight; i++) { //read each row
		if(!bResult) break;
		bResult = pclResource->Read(m_pGrid[i], m_iWidth*sizeof(**m_pGrid));
	}
	delete pclResource;
	return bResult;
}

//draw a certain tile anywhere on the screen
void C_Map::DrawTile(int i_iX, int i_iY, char i_cVal)
{
	if(!m_pclTilesImg || i_cVal>=m_iNumTiles) return;
	m_pclGraph->Blt(m_pclTilesImg, &m_stTileRects[(int)i_cVal], i_iX, i_iY);
}

//draws one clipped tile
void C_Map::DrawTileClipped(int i_iTileX, int i_iTileY, char i_cVal)
{
	S_Rect stRect = m_stTileRects[(int)i_cVal];
	int iScreenX = i_iTileX*m_iTileSize - m_iLastX;
	int iScreenY = i_iTileY*m_iTileSize - m_iLastY;

	int iTempX = m_stDrawarea.width - (iScreenX + m_iTileSize);
	int iTempY = m_stDrawarea.height - (iScreenY + m_iTileSize);
	if(iScreenX<0) {
		stRect.x += -iScreenX;
		stRect.width += iScreenX;
		if(stRect.width<=0) return;
		iScreenX = 0;
	} else if(iTempX<0) {
		stRect.width = m_iTileSize + iTempX;
		if(stRect.width<=0) return;
	}
	if(iScreenY<0) {
		stRect.y += -iScreenY;
		stRect.height += iScreenY;
		if(stRect.height<=0) return;
		iScreenY = 0;
	} else if(iTempY<0) {
		stRect.height = m_iTileSize + iTempY;
		if(stRect.height<=0) return;
	}

	m_pclGraph->Blt(m_pclTilesImg, &stRect, iScreenX + m_stDrawarea.x, iScreenY + m_stDrawarea.y);
}
