#include "image_utils.h"
#include "targa.h"
#include "object.h"

#include <math.h>

#include "FineGrainedPathFinder.h"
#include <algorithm>
#include <stdexcept>

#include "PathUtils.h"

//could not make it compile under linux in any better way
#ifndef WIN32
#define min(a, b) (((a) < (b)) ? (a) : (b))
#define max(a, b) (((a) > (b)) ? (a) : (b))
#endif

FineGrainedPathFinder::FineGrainedPathFinder(C_Map& map, const C_Image* shipImg, S_Rect shipRectangles[], int numActualDirections, const PathFinderSettings& settings) : 
	_numActualDirections(numActualDirections), _finalState(NULL), _result(0,0), _map(map), _mapWithSafetyMargin(NULL) 
{
	C_ShipState::Initialize();

	map.GetMapSize(&_w, &_h);
	_tileSize = map.GetTileSize();

	if (!settings.SettingsValid())
	{
		throw std::runtime_error("Invalid pathfinder settings.");
	}
	_numSubtiles = settings.NumSubtilesPerAxis;
	_timeStepMs = settings.TimeStepMs;
	_useHeuristicDistance = settings.UseHeuristicDistance;
	_maxDistMs = settings.MaxDistanceToTarget;
	_numDirectionCategories = settings.NumDirectionCategories;
	_numDirectionsPerCategory = PathFinderSettings::DEFAULT_NUM_DIRECTION_CATEGORIES / _numDirectionCategories;
	_bufferZoneWidth = settings.BufferZoneWidth;
	_timeFlowsForward = settings.TimeFlowsForward;
	_simpleMoveChecking = settings.UseSimplifiedMoveValidation;

	if (_tileSize % _numSubtiles != 0)
	{
		throw std::runtime_error("Tile size not supported.");
	}

	if (PathFinderSettings::DEFAULT_NUM_DIRECTION_CATEGORIES % _numDirectionCategories != 0)
	{
		throw std::runtime_error("Num. directions not supported.");
	}

	_subtileSizeShift = PathUtils::Round(log((double)_tileSize / _numSubtiles) / log(2.0));
	if (pow(2.0, _subtileSizeShift) != _tileSize / _numSubtiles)
	{
		throw std::runtime_error("Tile size not supported.");
	}

	_N = _numSubtiles * _w / _tileSize;
	_M = _numSubtiles * _h / _tileSize;
		
	_mapPixels.resize(_M, CharVector(_N));
	_distanceToWall.resize(_M, DoubleVector(_N));

	_moveAllowed = new bool**[_M];
	for (int k = 0; k < _M; ++k)
	{
		_moveAllowed[k] = new bool*[_N];
		for (int l = 0; l < _N; ++l)
		{
			_moveAllowed[k][l] = new bool[numActualDirections];
		}
	}

	_allShipStates = new C_ShipState**[_M];
	for (int k = 0; k < _M; ++k)
	{
		_allShipStates[k] = new C_ShipState*[_N];
		for (int l = 0; l < _N; ++l)
		{
			_allShipStates[k][l] = NULL;
		}
	}

	/*_distanceHeuristicLowerBound = new double*[_M];
	for (int k = 0; k < _M; ++k)
	{
		_distanceHeuristicLowerBound[k] = new double[_N];
		memset(_distanceHeuristicLowerBound[k], 0, _N * sizeof(double));
	}*/

	_nodesToVisit = new ShipStateQueue*[_maxDistMs / _timeStepMs];
	memset(_nodesToVisit, 0, sizeof(ShipStateQueue*) * _maxDistMs / _timeStepMs);

	_shipImg = shipImg;
	_shipRectangles = shipRectangles;

	_wallDistThreshold = 10.0;
}

FineGrainedPathFinder::~FineGrainedPathFinder()
{
	for (int k = 0; k < _M; ++k)
	{
		for (int l = 0; l < _N; ++l)
		{
			delete[] _moveAllowed[k][l];
			_moveAllowed[k][l] = NULL;
		}
		delete[] _moveAllowed[k];
	}
	delete[] _moveAllowed;

	if (_mapWithSafetyMargin)
	{
		delete _mapWithSafetyMargin;
	}
	
	for (int k = 0; k < _maxDistMs / _timeStepMs; ++k)
	{
		if(_nodesToVisit[k])
		{
			delete _nodesToVisit[k];
		}
	}
	delete[] _nodesToVisit;

	for (int k = 0; k < _M; ++k)
	{
		for (int l = 0; l < _N; ++l)
		{
			if (_allShipStates[k][l])
			{
				delete[] _allShipStates[k][l];
			}
		}
		delete[] _allShipStates[k];
	}
	delete[] _allShipStates;
}

void FineGrainedPathFinder::initialize()
{
	_map.Partition(_numSubtiles, _tilePartitioning);
	contributeTiles();
	calculateDistancesToWall();
	createBufferZone();
	calculateFeasiblePositions(_shipImg, _shipRectangles, _numActualDirections);
}

inline bool operator==(const S_AIOutput& o1, const S_AIOutput& o2)
{
	return o1.bFire == o2.bFire && o1.bLeft == o2.bLeft && o1.bRight == o2.bRight && o1.bThrottle == o2.bThrottle;
}

double FineGrainedPathFinder::getDistanceBetweenPoints(const S_ShipState& source, POINT destination, ShipMoveVector& moves)
{
	ShipStateQueue* currentQ;
	double deltaTs = _timeStepMs / 1000.0;
	
	C_ShipState initialState(source);
	initialState.setDistanceToSource(0.0);

	_destination = destination;
	_source.x = (int)source.stPos.x;
	_source.y = (int)source.stPos.y;
	
	destination.x = getSubtileIndex(destination.x);
	destination.y = getSubtileIndex(destination.y);

	_moveAllowed[getSubtileIndex(source.stPos.y)][getSubtileIndex(source.stPos.x)][(int)(source.dPointing / OBJECT_HEADING_STEP)] = true;
	
	commitToQueue(&initialState, deltaTs);
		
	bool foundPath = false;
	_finalState = NULL;
	moves.clear();

	_terminate = false;
	
	for (int32_t k = 0; k < _maxDistMs / _timeStepMs; ++k)
	{
		currentQ = _nodesToVisit[k];
		
		while (!_terminate && !foundPath && currentQ && !currentQ->empty())
		{
			C_ShipState* currState = currentQ->front();
			currentQ->pop();

			if (!currState->isProcessingFinished())
			{
				//mark as finished - we cannot enhance the distance to this state anymore
				currState->markProcessingFinished();
				
				int currSubtileX = getSubtileIndex(currState->getX());
				int currSubtileY = getSubtileIndex(currState->getY());

				bool closeQuarters = _distanceToWall[currSubtileY][currSubtileX] < _wallDistThreshold;

				bool canDoLeft = (closeQuarters || currState->getLastRightRotationTimeElapsed() >= 3);
				bool canDoRight = (closeQuarters || currState->getLastLeftRotationTimeElapsed() >= 3);

				if (currSubtileX == destination.x
					&& currSubtileY == destination.y)
				{
					//we're done
					foundPath = true;
					_finalState = currState;
					break;
				}

				{
					//check move 1: thrust
					S_AIOutput theMove = {false, false, true, false};
					checkMove(currState, theMove, deltaTs);
				}
				
				if (canDoLeft)
				{
					//check move 2: thrust + left
					S_AIOutput theMove = {true, false, true, false};
					checkMove(currState, theMove, deltaTs);
				}

				if (canDoRight)
				{
					//check move 3: thrust + right
					S_AIOutput theMove = {false, true, true, false};
					checkMove(currState, theMove, deltaTs);
				}

				if (closeQuarters)
				{
					{
					//check move 0: null
						S_AIOutput theMove = {false, false, false, false};
						checkMove(currState, theMove, deltaTs);
					}

					if (canDoLeft)
					{
						//check move 4: left
						S_AIOutput theMove = {true, false, false, false};
						checkMove(currState, theMove, deltaTs);
					}

					if (canDoRight)
					{
						//check move 5: right
						S_AIOutput theMove = {false, true, false, false};
						checkMove(currState, theMove, deltaTs);
					}
				}
			}
		}

		delete _nodesToVisit[k];
		_nodesToVisit[k] = NULL;
	}

	if (foundPath && _finalState)
	{
		visualizePath(_finalState);
		
		C_TargaImg::Save((char*)"path.tga", _mapWithSafetyMargin);
		backtracePathMoves(_finalState, moves);
		return _finalState->getDistanceToSource();
	}

	C_TargaImg::Save((char*)"path.tga", _mapWithSafetyMargin);

	return _maxDistMs;
}

S_AIOutput FineGrainedPathFinder::getOptimalTrajectory(const S_ShipState& source, const ShipMoveVector& moves, double timeQuantum)
{
	S_AIOutput result = {false, false, false, false};

	return result;
}

bool operator==(const POINT& p1, const POINT& p2);

inline bool FineGrainedPathFinder::canMove(const C_ShipState* from, const C_ShipState* to) const
{
	POINT startPt = {getSubtileIndex(from->getX()), getSubtileIndex(from->getY())};
	POINT endPt = {getSubtileIndex(to->getX()), getSubtileIndex(to->getY())};

	int minx = min(startPt.x, endPt.x);
	int maxx = max(startPt.x, endPt.x);

	int miny = min(startPt.y, endPt.y);
	int maxy = max(startPt.y, endPt.y);

	if (minx < 0 || maxx < 0
		|| minx >= _N || maxx >= _N
		|| miny < 0 || maxy < 0
		|| miny >= _M || maxy >= _M)
	{
		return false;
	}

	int d1 = (int)(from->getDirection() / OBJECT_HEADING_STEP);
	int d2 = (int)(to->getDirection() / OBJECT_HEADING_STEP);

	if (_simpleMoveChecking)
	{
		if (!_moveAllowed[startPt.y][startPt.x][d1]
			|| !_moveAllowed[endPt.y][endPt.x][d2])
		{
			return false;
		}

		return true;
	}

	//TODO optimize this
	for (int x = minx; x <= maxx; ++x)
	{
		for (int y = miny; y <= maxy; ++y)
		{
			if (!_moveAllowed[y][x][d1]
				|| !_moveAllowed[y][x][d2])
			{
#ifdef VISUALIZE_INVALID_MOVES
				if (!_moveAllowed[y][x][d1])
				{
					S_Rect* pRect = &_shipRectangles[d1];
					int w = _tileSize / _numSubtiles;
					ShipState st(x * w, y*w);
					uint32 transparentColor = 0;

					if (st.getX() + pRect->width < _w && st.getY() + pRect->height < _h)
					{
						_mapWithSafetyMargin->Blt2(const_cast<C_Image*>(_shipImg), pRect, (int)st.getX(), (int)st.getY(), &transparentColor);
					}
				}
				if (!_moveAllowed[y][x][d2])
				{
					S_Rect* pRect = &_shipRectangles[d2];
					int w = _tileSize / _numSubtiles;
					ShipState st(x * w, y*w);
					uint32 transparentColor = 0;

					if (st.getX() + pRect->width < _w && st.getY() + pRect->height < _h)
					{
						_mapWithSafetyMargin->Blt2(const_cast<C_Image*>(_shipImg), pRect, (int)st.getX(), (int)st.getY(), &transparentColor);
					}
				}
#endif
				return false;
			}
		}
	}

	return true;
}

void FineGrainedPathFinder::contributeTiles()
{
	int w, h;
	
	w = _w / _tileSize;
	h = _h / _tileSize;

	for (int x = 0; x < w; ++x)
	{
		for (int y = 0; y < h; ++y)
		{
			int tileNum = _map.GetElementAbsolute(x, y);

			vector<bool> subtiles = _tilePartitioning[tileNum];

			for (int i = 0; i < _numSubtiles; ++i)
			{
				for (int j = 0; j < _numSubtiles; ++j)
				{
					int xx = x*_numSubtiles + i;
					int yy = y*_numSubtiles + j;

					_mapPixels[yy][xx] = (char)subtiles[j*_numSubtiles + i];
				}
			}
		}
	}
}

//approximate Euclidean distances to wall from each subtile
void FineGrainedPathFinder::calculateDistancesToWall()
{
	double wtDiagonal = sqrt(2.0) / 2 + sqrt(sqrt(2.0) - 1);
	double wtNonDiagonal = 1;
	PointVector neighbors;

	for (int x = 0; x < _N; ++x)
	{
		for (int y = 0; y < _M; ++y)
		{
			_distanceToWall[y][x] = 0;
		}
	}

	for (int y = 0; y < _M; ++y)
	{
		for (int x = 0; x < _N; ++x)
		{	
			if (_mapPixels[y][x])
			{
				POINT p = {x,y};

				getPointUpperNeighborhood(p, neighbors);
				for (PointVector::iterator n = neighbors.begin(); n != neighbors.end(); ++n)
				{	
					double consideredDist = _distanceToWall[n->y][n->x] + (PathUtils::IsUnitDistance(p, *n) ? wtNonDiagonal : wtDiagonal);
					if (_distanceToWall[y][x] == 0 || consideredDist < _distanceToWall[y][x])
					{
						_distanceToWall[y][x] = consideredDist;
					}
				}
			}
		}
	}

	for (int y = _M - 1; y >= 0; --y)
	{
		for (int x = _N - 1; x >= 0; --x)
		{
			POINT p = {x,y};

			getPointLowerNeighborhood(p, neighbors);
			for (PointVector::iterator n = neighbors.begin(); n != neighbors.end(); ++n)
			{
				double consideredDist = _distanceToWall[n->y][n->x] + (PathUtils::IsUnitDistance(p, *n) ? wtNonDiagonal : wtDiagonal);
				if (consideredDist < _distanceToWall[y][x])
				{
					_distanceToWall[y][x] = consideredDist;
				}
			}
		}
	}

#if 0
	vector<DoubleVector> realDists(_M, DoubleVector(_N));

	double maxError = 0;
	double percentDev = 0;

	for (int x = 0; x < _N; ++x)
	{
		for (int y = 0; y < _M; ++y)
		{
			realDists[y][x] = 0;
			
			if (_mapPixels[y][x])
			{
				for (int x2 = 0; x2 < _N; ++x2)
				{
					for (int y2 = 0; y2 < _M; ++y2)
					{
						if (!_mapPixels[y2][x2])
						{
							POINT p1 = {x,y}, p2 = {x2, y2};

							double eucDist = EuclideanDistance(p1, p2);

							if (realDists[y][x] == 0 || eucDist < realDists[y][x])
							{
								realDists[y][x] = eucDist;
							}
						}
					}
				}
			}

			double currError = fabs(realDists[y][x] - _distanceToWall[y][x]);
			if (currError > maxError)
			{
				maxError = currError;
				percentDev = currError / min(realDists[y][x], _distanceToWall[y][x]);
			}
		}
	}

	if (maxError > 0)
	{
		maxError *= 1.0;
	}
#endif
}

PointVector& FineGrainedPathFinder::getPointUpperNeighborhood(POINT p, PointVector& vec)
{
	POINT n;

	vec.clear();

	if (p.x > 0)
	{
		n.x = p.x - 1;
		n.y = p.y;
		vec.push_back(n);
		
		if (p.y > 0)
		{
			n.x = p.x - 1;
			n.y = p.y - 1;
			vec.push_back(n);
		}
	}
	
	if (p.y > 0)
	{
		n.x = p.x;
		n.y = p.y - 1;
		vec.push_back(n);

		if (p.x < _N - 1)
		{
			n.x = p.x + 1;
			n.y = p.y - 1;
			vec.push_back(n);
		}
	}

	return vec;
}

PointVector& FineGrainedPathFinder::getPointLowerNeighborhood(POINT p, PointVector& vec)
{
	POINT n;

	vec.clear();

	if (p.x < _N - 1)
	{
		n.x = p.x + 1;
		n.y = p.y;
		vec.push_back(n);
		
		if (p.y < _M - 1)
		{
			n.x = p.x + 1;
			n.y = p.y + 1;
			vec.push_back(n);
		}
	}
	
	if (p.y < _M - 1)
	{
		n.x = p.x;
		n.y = p.y + 1;
		vec.push_back(n);

		if (p.x > 0)
		{
			n.x = p.x - 1;
			n.y = p.y + 1;
			vec.push_back(n);
		}
	}

	return vec;
}

void FineGrainedPathFinder::backtracePathMoves(const C_ShipState* state, ShipMoveVector& moves)
{
	while (state)
	{
		S_ExtendedAIOutput output;
		output.State = (S_ShipState)*state;
		output.Command = state->getOutput();
		moves.push_back(output);

		state = state->getPredecessor();
	}

	std::reverse(moves.begin(), moves.end());
}

void FineGrainedPathFinder::visualizePath(const C_ShipState* state)
{
	uint32_t transparentColor = 0;

	while (state)
	{
		int realDir = (int)(state->getDirection() / OBJECT_HEADING_STEP);

		S_Rect* pRect = &_shipRectangles[realDir];

		if (state->getX() + pRect->width < _w && state->getY() + pRect->height < _h)
		{
			_mapWithSafetyMargin->Blt2(const_cast<C_Image*>(_shipImg), pRect, (int)state->getX(), (int)state->getY(), &transparentColor);
		}

		state = state->getPredecessor();
	}

	//C_TargaImg::Save("path.tga", _mapWithSafetyMargin);
}

void FineGrainedPathFinder::calculateFeasiblePositions(const C_Image* shipImg, S_Rect shipRectangles[], int numActualDirections)
{
	//C_Image theMapCopy(theMap);

	//TODO calculate this in a more precise manner (take into account positions at all extreme points of the subtile - affects safety of ship)
	for (int x = 0; x < _N; ++x)
	{
		for (int y = 0; y < _M; ++y)
		{
			int w = _tileSize / _numSubtiles;
			S_Rect mapRect = {x*w, y*w,
								shipRectangles[0].width,
								shipRectangles[0].height};
						
			for (int d = 0; d < numActualDirections; ++d)
			{
				bool isCollision = false;

				if (mapRect.x + mapRect.width - 1 >= _w
					|| mapRect.y + mapRect.height - 1 >= _h)
				{
					isCollision = true;
				}
				else
				{
					isCollision = ImageUtils::Compare(_mapWithSafetyMargin, shipImg, &mapRect, &shipRectangles[d]);
				}

				_moveAllowed[y][x][d] = !isCollision;
			}

			
			/*mapRect.width = tileSize / _numSubtiles;
			mapRect.height = tileSize / _numSubtiles;

			if (xSubtileNum < numExtendedTiles)
			{
				mapRect.width++;
			}
			if (ySubtileNum < numExtendedTiles)
			{
				mapRect.height++;
			}

			S_Color walkableColor = {100, 0, 0, (char)128};
			S_Color wallColor = {0, 100, 0, (char)128};

			if (isCollision)
			{
				theMapCopy.FillWithRGBColor(&mapRect, &wallColor);
			}
			else
			{
				theMapCopy.FillWithRGBColor(&mapRect, &walkableColor);
			}*/
		}
	}

#if 0
	for (int k = 0; k < numActualDirections; ++k)
	{
		C_Image theMapCopy(_mapWithSafetyMargin);

		for (int x = 0; x < _N; ++x)
		{
			for (int y = 0; y < _M; ++y)
			{
				int w = _tileSize / _numSubtiles;
				S_Rect mapRect = {w * x, w*y, w, w};
				
				S_Color walkableColor = {100, 0, 0, (char)128};
				S_Color wallColor = {0, 100, 0, (char)128};

				if (!_moveAllowed[y][x][k])
				{
					theMapCopy.FillWithRGBColor(&mapRect, &wallColor);
				}
				else
				{
					theMapCopy.FillWithRGBColor(&mapRect, &walkableColor);
				}
			}
		}

		char imgName[50];
		sprintf_s(imgName, "collisions_%d.tga", (k+1));
		C_TargaImg::Save(imgName, &theMapCopy);
	}
#endif

	//C_TargaImg::Save("collisions28.tga", &theMapCopy);
	//C_TargaImg::Save("fullMap.tga", theMap);

	//delete _mapWithSafetyMargin;
}

void FineGrainedPathFinder::createBufferZone()
{
	int pixelFmt;
		
	_map.GetTilesetImage()->GetInfo(NULL, NULL, &pixelFmt);

	_mapWithSafetyMargin = new C_Image(_w, _h, pixelFmt);

	S_Rect drawingRange = {0, 0, _w / _tileSize, _h / _tileSize};
	_map.DrawOn(_mapWithSafetyMargin, &drawingRange);

	vector<POINT> leftBorder, rightBorder, topBorder, bottomBorder;

	for (int y = 1; y < _h - 1; ++y)
	{
		for (int x = 1; x < _w - 1; ++x)
		{
			S_Color color = _mapWithSafetyMargin->GetPixelRGB(x, y);
			POINT p = {x, y};

			if (color.b != 0 || color.g != 0 || color.r != 0)
			{
				S_Color color2 = _mapWithSafetyMargin->GetPixelRGB(x + 1, y);

				if (color2.b == 0 && color2.g == 0 && color2.r == 0)
				{
					leftBorder.push_back(p);
				}

				color2 = _mapWithSafetyMargin->GetPixelRGB(x - 1, y);
				if (color2.b == 0 && color2.g == 0 && color2.r == 0)
				{
					rightBorder.push_back(p);
				}

				color2 = _mapWithSafetyMargin->GetPixelRGB(x, y + 1);
				if (color2.b == 0 && color2.g == 0 && color2.r == 0)
				{
					topBorder.push_back(p);
				}

				color2 = _mapWithSafetyMargin->GetPixelRGB(x, y - 1);
				if (color2.b == 0 && color2.g == 0 && color2.r == 0)
				{
					bottomBorder.push_back(p);
				}
			}
		}
	}

	S_Color borderColor = { 0x80, 0x80, 0x00, 0};

	for (vector<POINT>::iterator pt = leftBorder.begin(); pt != leftBorder.end(); ++pt)
	{
		for (int k = 1; k <= _bufferZoneWidth; ++k)
		{
			if (pt->x + k < _w)
			{
				S_Rect area = {pt->x + k, pt->y, 1, 1};
				_mapWithSafetyMargin->FillWithRGBColor(&area, &borderColor);
			}
		}
	}

	for (vector<POINT>::iterator pt = rightBorder.begin(); pt != rightBorder.end(); ++pt)
	{
		for (int k = 1; k <= _bufferZoneWidth; ++k)
		{
			if (pt->x - k  >= 0)
			{
				S_Rect area = {pt->x - k, pt->y, 1, 1};
				_mapWithSafetyMargin->FillWithRGBColor(&area, &borderColor);
			}
		}
	}

	for (vector<POINT>::iterator pt = topBorder.begin(); pt != topBorder.end(); ++pt)
	{
		for (int k = 1; k <= _bufferZoneWidth; ++k)
		{
			if (pt->y + k < _h)
			{
				S_Rect area = {pt->x, pt->y + k, 1, 1};
				_mapWithSafetyMargin->FillWithRGBColor(&area, &borderColor);
			}
		}
	}

	for (vector<POINT>::iterator pt = bottomBorder.begin(); pt != bottomBorder.end(); ++pt)
	{
		for (int k = 1; k <= _bufferZoneWidth; ++k)
		{
			if (pt->y - k >= 0)
			{
				S_Rect area = {pt->x, pt->y - k, 1, 1};
				_mapWithSafetyMargin->FillWithRGBColor(&area, &borderColor);
			}
		}
	}

	//C_TargaImg::Save("bufferedMap.tga", _mapWithSafetyMargin);
}

void FineGrainedPathFinder::absorbHeuristicDistanceBound(const FineGrainedPathFinder& other)
{
	int subtileScalingFactor = _numSubtiles / other._numSubtiles;

	for (int x = 0; x < _N; ++x)
	{
		for (int y = 0; y < _M; ++y)
		{
			_distanceHeuristicLowerBound[y][x] = -1;

#if 0
			int otherY = y / subtileScalingFactor;
			int otherX = x / subtileScalingFactor;

			if (other._allShipStates[otherY][otherX] != NULL)
			{
				for (int d = 0; d < other._numDirectionCategories; ++d)
				{
					if (other._allShipStates[otherY][otherX][d].getDistanceToSource() != ShipState::INFINITE_DISTANCE)
					{
						if (_distanceHeuristicLowerBound[y][x] == -1
							|| _distanceHeuristicLowerBound[y][x] > other._allShipStates[otherY][otherX][d].getDistanceToSource())
						{
							_distanceHeuristicLowerBound[y][x] = other._allShipStates[otherY][otherX][d].getDistanceToSource();
						}
					}
				}
			}
#endif

			if (_distanceHeuristicLowerBound[y][x] == -1)
			{
				POINT position = { x * _tileSize / _numSubtiles, y * _tileSize / _numSubtiles};
				_distanceHeuristicLowerBound[y][x] = PathUtils::EuclideanDistance(position, other._source) / 200.0;
			}

			//to make it more of a lower bound - it's not exact
			//_distanceHeuristicLowerBound[y][x] *= 0.30;
		}
	}
}
