#pragma once

#include <queue>
#include <vector>
#include "ShipState.h"
#include "tilemap.h"

#include "DataStructsAI.h"
#include "PathUtils.h"

#include "PathFinderSettings.h"

using std::queue;
using std::vector;

typedef vector<S_ExtendedAIOutput> ShipMoveVector;
typedef vector<POINT> PointVector;
typedef vector<bool> BoolVector;
typedef vector<char> CharVector;
typedef vector<double> DoubleVector;


class FineGrainedPathFinder
{
public:
	FineGrainedPathFinder(C_Map& map, const C_Image* shipImg, S_Rect shipRectangles[], int numActualDirections, 
		const PathFinderSettings& settings);
	~FineGrainedPathFinder();

	void initialize();

	double getDistanceBetweenPoints(const S_ShipState& source, POINT destination, ShipMoveVector& moves);
	S_AIOutput getOptimalTrajectory(const S_ShipState& source, const ShipMoveVector& moves, double timeQuantum);

	const C_ShipState* getFinalState() const { return _finalState; }

	void absorbHeuristicDistanceBound(const FineGrainedPathFinder& other);
	
	void stop() { _terminate = true; }

private:
	void clear();

	bool canMove(const C_ShipState* from, const C_ShipState* to) const;
	void contributeTiles();
	void calculateDistancesToWall();
	void calculateFeasiblePositions(const C_Image* shipImg, S_Rect shipRectangles[], int numActualDirections);
	void createBufferZone();

	int getSubtileIndex(double pixelCoord) const;
	
	C_ShipState* commitState(const C_ShipState& stateRef, bool &isNew);
	void commitToQueue(C_ShipState* state, double deltaT);
	bool checkMove(C_ShipState* currState, const S_AIOutput& theMove, double deltaT);

	PointVector& getPointUpperNeighborhood(POINT p, PointVector& vec);
	PointVector& getPointLowerNeighborhood(POINT p, PointVector& vec);

	void visualizePath(const C_ShipState* state);
	void backtracePathMoves(const C_ShipState* state, ShipMoveVector& moves);
	
	int _N, _M;
	int _tileSize;
	
	int _numSubtiles;
	int _timeStepMs;
	int _maxDistMs;
	int _bufferZoneWidth;
	int _numDirectionCategories;
	int _numDirectionsPerCategory;
	int _numActualDirections;
	
	bool _useHeuristicDistance;
	bool _timeFlowsForward;
	bool _simpleMoveChecking;
	
	POINT _destination;
	C_ShipState* _finalState;

	volatile bool _terminate;
	
	int _w,_h; //map width and height in pixels

	int _subtileSizeShift;

	POINT _source;
	C_ShipState _result;
	ShipStateQueue** _nodesToVisit;
	C_Map& _map;
	vector<CharVector> _mapPixels;
	vector<DoubleVector> _distanceToWall;
	vector<BoolVector> _tilePartitioning;
	bool*** _moveAllowed;
	C_ShipState*** _allShipStates;
	double** _distanceHeuristicLowerBound;

	double _wallDistThreshold;

	const C_Image* _shipImg;
	S_Rect* _shipRectangles;

	C_Image* _mapWithSafetyMargin;

	static const uint32_t MAX_VELOCITY = 700;
};


inline int FineGrainedPathFinder::getSubtileIndex(double pixelCoord) const
{
	return (int)ceil(pixelCoord) >> _subtileSizeShift;
}

inline void FineGrainedPathFinder::commitToQueue(C_ShipState* state, double deltaT)
{
	int queueIndex = (int)ceil((_useHeuristicDistance ? state->getTotalDistance() : state->getDistanceToSource()) / deltaT);

	if (_nodesToVisit[queueIndex] == NULL)
	{
		_nodesToVisit[queueIndex] = new ShipStateQueue();
	}

	_nodesToVisit[queueIndex]->push(state);
}

inline bool FineGrainedPathFinder::checkMove(C_ShipState* currState, const S_AIOutput& theMove, double deltaT)
{
	C_ShipState newState(*currState, theMove, deltaT, _timeFlowsForward);
	bool isNew;

	if (canMove(currState, &newState))
	{
		if (_useHeuristicDistance)
		{
			//newState.setHeuristicDistance(_distanceHeuristicLowerBound[getSubtileIndex(newState.getY())][getSubtileIndex(newState.getX())]);
			newState.setHeuristicDistance(PathUtils::EuclideanDistance(newState.getPosition(), _destination) / 400.0);
		}

		C_ShipState* existingState = commitState(newState, isNew);
		//if (isNew ||
		//	(!existingState->isProcessingFinished() && 
		//	((!_useHeuristicDistance && existingState->getDistanceToSource() > newState.getDistanceToSource())
		//	||
		//	(_useHeuristicDistance && existingState->getTotalDistance() > newState.getTotalDistance())))
		//	)
		if (!existingState->isProcessingFinished())
		{
			*existingState = newState;
			commitToQueue(existingState, deltaT);

			return true;
		}
	}

	return false;
}

inline C_ShipState* FineGrainedPathFinder::commitState(const C_ShipState& stateRef, bool &isNew)
{
	int subtileIndexX = getSubtileIndex(stateRef.getX());
	int subtileIndexY = getSubtileIndex(stateRef.getY());

	isNew = false;

	if (_allShipStates[subtileIndexY][subtileIndexX] == NULL)
	{
		_allShipStates[subtileIndexY][subtileIndexX] = new C_ShipState[_numDirectionCategories];
	}

	int actualDirection = (int)stateRef.getDirection() / OBJECT_HEADING_STEP;
	int directionCategory = actualDirection / _numDirectionsPerCategory;

	C_ShipState* theState = &_allShipStates[subtileIndexY][subtileIndexX][directionCategory];

	if (theState->getDistanceToSource() == C_ShipState::INFINITE_DISTANCE)
	{
		isNew = true;
	}

	if (isNew)
	{
		*theState = stateRef;
	}

	return theState;
}
