// C_ShipState

#ifndef SHIPSTATE_H
#define SHIPSTATE_H

#include <queue>
#include <vector>
#include <stdexcept>
#include <math.h>
#include "common.h"

#include "DataStructsAI.h"
#include "environment_defaults.h"

using std::queue;
using std::vector;

class C_ShipState
{
public:
	C_ShipState(double x, double y) : _x(x), _y(y), _vx(0), _vy(0), _direction(270.0), 
		_distance(INFINITE_DISTANCE), _predecessor(NULL), _processingFinished(false),
		_lastRotationLeft((int)INFINITE_DISTANCE), _lastRotationRight((int)INFINITE_DISTANCE), _heuristicDistance(0) {}

	C_ShipState(const S_ShipState& base) : _x(base.stPos.x), _y(base.stPos.y), _vx(base.stVel.x), _vy(base.stVel.y), 
		_direction(base.dPointing),  _distance(INFINITE_DISTANCE), _predecessor(NULL), _processingFinished(false),
		_lastRotationLeft((int)INFINITE_DISTANCE), _lastRotationRight((int)INFINITE_DISTANCE), _heuristicDistance(0) {}

	C_ShipState() : _distance(INFINITE_DISTANCE), _processingFinished(false), _heuristicDistance(0) {}

	C_ShipState(const C_ShipState& base, const S_AIOutput& move, double deltaT, bool timeForward);

	double getDistanceToSource() const { return _distance; }
	void setDistanceToSource(double distance) { _distance = distance; }

	double getX() const { return _x; }
	double getY() const { return _y; }

	POINT getPosition() const { POINT p = {int(_x), int(_y)}; return p; }

	double getVelocityX() const { return _vx; }
	double getVelocityY() const { return _vy; }

	void setVelocityX(double vx) { _vx = vx; }
	void setVelocityY(double vy) { _vy = vy; }

	double getDirection() const { return _direction; }
	void setDirection(double dir) { _direction = dir; }

	void setPredecessor(C_ShipState* predecessor) { _predecessor = predecessor; }
	const C_ShipState* getPredecessor() const { return _predecessor; }

	const S_AIOutput& getOutput() const { return _output; }

	void setHeuristicDistance(double dist) { _heuristicDistance = dist; }

	bool isProcessingFinished() const { return _processingFinished; }
	void markProcessingFinished() { _processingFinished = true; }

	int getLastLeftRotationTimeElapsed() const { return _lastRotationLeft; }
	int getLastRightRotationTimeElapsed() const { return _lastRotationRight; }

	double getTotalDistance() const { return _distance + _heuristicDistance; }

	operator S_ShipState() const;
		
	static const double INFINITE_DISTANCE;

	static void Initialize();

private:
	
	double _x, _y;
	double _vx, _vy;
	double _direction;
	double _distance;
	const C_ShipState* _predecessor;
	
	bool _processingFinished;
	int _lastRotationLeft;
	int _lastRotationRight;

	double _heuristicDistance;

	S_AIOutput _output; //output submitted at previous location in order to get here
	double _deltaT;

	static double SinTable[360 / OBJECT_HEADING_STEP];
	static double CosTable[360 / OBJECT_HEADING_STEP];

	bool statesEqual(const C_ShipState& other) const;

	friend class FineGrainedPathFinder;
};

inline C_ShipState::C_ShipState(const C_ShipState& base, const S_AIOutput& move, double deltaT, bool timeForward) : 
	_distance(base._distance + deltaT), _predecessor(&base), _processingFinished(false), _heuristicDistance(0)
{
	_direction = base._direction;
	if (move.bLeft && !move.bRight)
	{
		_direction -= C_Environment::GetShipSteerSpeed() * deltaT;
	}
	else if (!move.bLeft && move.bRight)
	{
		_direction += C_Environment::GetShipSteerSpeed() * deltaT;
	}

	if (_direction < 0 || _direction >= 360)
	{
		double div = floor(_direction / 360.0);
		_direction -= div * 360;
	}

	if (_direction < 0 || _direction >= 360)
	{
		throw new std::range_error("invalid heading in ShipState!");
	}

	double tmpDir = _direction;

	if (!timeForward)
	{
		tmpDir -= 180.0;
		if (tmpDir < 0.0)
		{
			tmpDir += 360.0;
		}
	}

	int realDir = (int)(tmpDir / OBJECT_HEADING_STEP);
	double acceleration = move.bThrottle ? C_Environment::GetShipThrust() : 0.0;

	double ax = CosTable[realDir] * acceleration;
	double ay = SinTable[realDir] * acceleration;

	//ax += 0;
	ay += (timeForward ? C_Environment::GetShipGravityBaseY() : -C_Environment::GetShipGravityBaseY()); //TODO handle cargo
	
	//compute the next state (time flows forward)
	ax -= base._vx*C_Environment::GetShipResistanceX();
	ay -= base._vy*C_Environment::GetShipResistanceY();

	double axDeltaT = ax*deltaT;
	double axDeltaTsq = axDeltaT*deltaT;

	double ayDeltaT = ay*deltaT;
	double ayDeltaTsq = ayDeltaT*deltaT;

	//calc position
	_x = base._x + (base._vx * deltaT) + (0.5*axDeltaTsq);
	_y = base._y + (base._vy * deltaT) + (0.5*ayDeltaTsq);

	//calc vel
	_vx = base._vx + axDeltaT;
	_vy = base._vy + ayDeltaT;
	
	_output = move;
	_deltaT = deltaT;

	_lastRotationLeft = move.bLeft ? 0 : base._lastRotationLeft + 1;
	_lastRotationRight = move.bRight ? 0 : base._lastRotationRight + 1;
}

inline C_ShipState::operator S_ShipState() const
{
	S_ShipState state;
	state.dPointing = _direction;
	state.stPos.x = _x;
	state.stPos.y = _y;
	state.stVel.x = _vx;
	state.stVel.y = _vy;

	return state;
}

typedef queue<C_ShipState*> ShipStateQueue;
typedef vector<C_ShipState> ShipStateVector;

#endif
