
//main startup function and
//class to tie everything in the game together.
// - init
// - game state machine
//   - runs the gui
//   - network connect and server messaging
//   - setup and starts a game
//   - runs a game (or replay)
//   - handle hiscores after a game is over
// - deinit

#include "socket_2.h"
#include "netmsg.h"
#include "netmsgserver.h"

#include "resource_gf.h"

#include "replay.h"
#include "viewport.h"
#include "glgraph.h" //includes os specific
#include "soundfx.h"
#include "input.h"

#include "fileresource.h"
#include "timer.h"
#include "gameview2.h"
#include "global.h"
#include "settings.h"

#include "showhiscore.h"
#include "logger.h"

#include "global_hiscore.h"

//#define SAVE_MOVIE    //F2 toggles
#define SAVE_MOVIE_FPS 30
#define SAVE_SCREENS  //F3 saves one frame
#include "moviemaker.h"

#define TIMESLOT 4.0 //ms

#ifdef SAVE_MOVIE
//4:3 in when saving a movie
#define DEFAULT_WINDOW_W 800
#define DEFAULT_WINDOW_H 600
#else
#define DEFAULT_WINDOW_W 864
#define DEFAULT_WINDOW_H 632
#endif

#define PROGRAM_VERSION_MAJOR 1  //->2.0
#define PROGRAM_VERSION_MINOR 86

#define NUM_LIFES 5

//ip ports are specified in the config file (or default in global.cpp)

//fwd decl
#ifdef WIN32
LRESULT CALLBACK WindowProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam);
#endif

////////////////////////////////////////////////////////////////////////////////
class C_GameHandler {
public:
	C_GameHandler();
	~C_GameHandler();
	bool Init();
	void Update();

	void SetSize(int w, int h);

	char *GetErrorMsg() {return m_szErrorMsg;};

	C_Control            *m_pclControl;
private:
	bool                 m_bInitDone;
	char                 m_szErrorMsg[1024];
	//network
	C_Socket             *m_pclServSocket;
	C_NetMsgServer       *m_pclNetMsgServer;
	C_NetMsg             *m_pclNetMsg;

	//dialog
	C_GameView           *m_pclGVDlg;

	//graph
	C_GraphWrapperGL     *m_pclGraph;
	C_Image              *m_pclBKImg;

	//sound
	void StartMusic(bool i_bSetVolumeOnly = false);
	C_Sound              *m_pclSound;
	C_Timer              *m_pclTimer3;
	bool                 m_bMusicPlaying;
	void                 *m_pMusicChn; //need not be freed

	//gameplay
	C_Replay             *m_pclCurReplay;
	C_Replay             *m_pclLastReplay;
	C_Viewport           *m_pclViewport;
	C_ShowHiscore        *m_pclHiscore;
	C_Timer              *m_pclTimer;
	C_Timer              *m_pclTimer2;
	C_Timer              *m_pclTimer4;
	double               m_dAccTime;
	int                  m_iGameState;
	bool                 m_bEscPressed;
	bool                 m_bRestartPressed;
	int                  m_iScore;

	C_MemberAchievements *m_pclMemberAch;
	int                  m_iLoginState; //to get a draw before blocking login check

	//gameplay info
	char                 m_szPlayerName[20];
	int                  m_iPlayerId;
	int                  m_iGameId;
	bool                 m_bMaster;
	int                  m_iLevelType;
	bool                 m_bIsReplay;
	int                  m_aiAIIndex[8];
	int                  m_iNumAI;
	int GetPlayerScore(int i_iIndex);

	//linux: faulty joystick workaround
	int                  m_iJoystickSampleCounter;
	int                  m_iJoystickSampleX;

	//save movie
	void SaveScreens();
	bool                 m_bSaveMovie;
	bool                 m_bSaveScreen;
	C_MovieMaker         *m_pclMovieMaker;

	void FreeGameObjects();
	void FreeServerNet();

	bool FileCheck();
	bool InitInstance();

	void HandleServerMsg();
	int ProcessDialog();
	void SendOptions();
	void StartGame();
	void RedrawScreen();

	void UpdateKeypress();
	void UpdateMouse();

	void StartReplay();

#ifdef WIN32
	HANDLE m_hServer;
#else
	int m_hServer; //pid
#endif
};
////////////////////////////////////////////////////////////////////////////////

C_GameHandler::C_GameHandler()
{
	m_bInitDone        = false;
	//network
	m_pclServSocket    = NULL;
	m_pclNetMsgServer  = NULL;
	m_pclNetMsg        = NULL;
	//dialog
	m_pclGVDlg         = NULL;
	//graph
	m_pclGraph         = NULL;
	m_pclBKImg         = NULL;
	//sound
	m_pclSound         = NULL;
	m_pclTimer3        = new C_Timer();
	m_bMusicPlaying    = true;
	m_pMusicChn        = NULL;
	//gameplay
	m_pclControl       = NULL;
	m_pclViewport      = NULL;
	m_iGameState       = 0;
	m_bEscPressed      = false;
	m_bRestartPressed  = false;
	m_pclTimer         = new C_Timer();
	m_pclTimer2        = new C_Timer();
	m_pclTimer4        = new C_Timer();
	m_pclHiscore       = NULL;
	m_pclCurReplay     = new C_Replay();
	m_pclLastReplay    = NULL;
	m_bIsReplay        = false;
	m_iScore           = 0;

	m_pclMemberAch     = new C_MemberAchievements();
	m_iLoginState      = 0;
	//gameplay info
	m_iGameId          = 0;
	m_bMaster          = false;
	//other
	m_bSaveMovie       = false;
	m_bSaveScreen      = false;
	m_pclMovieMaker    = NULL;
	m_szErrorMsg[0]    = 0;

	m_hServer          = 0;
}

C_GameHandler::~C_GameHandler()
{
	FreeGameObjects();

	delete m_pclMovieMaker;	m_pclMovieMaker = NULL;

	//dialog
	delete m_pclGVDlg; m_pclGVDlg = NULL;
	//graph
	delete m_pclGraph; m_pclGraph = NULL;
	//sound
	delete m_pclTimer3; m_pclTimer3 = NULL;
	delete m_pclSound; m_pclSound = NULL;

	//other
	delete m_pclControl; m_pclControl = NULL;
	delete m_pclHiscore; m_pclHiscore = NULL;
	delete m_pclTimer;   m_pclTimer = NULL;
	delete m_pclTimer2;  m_pclTimer2 = NULL;
	delete m_pclTimer4;  m_pclTimer4 = NULL;
	delete m_pclCurReplay;  m_pclCurReplay = NULL;
	delete m_pclLastReplay; m_pclLastReplay = NULL;
	delete m_pclMemberAch; m_pclMemberAch = NULL;

	//network
	FreeServerNet();

#ifdef WIN32
	TerminateProcess(m_hServer, 0);
#else
	kill(m_hServer, SIGKILL);
#endif
}

void C_GameHandler::FreeGameObjects()
{
	delete m_pclBKImg;    m_pclBKImg = NULL;
	delete m_pclViewport; m_pclViewport = NULL;
}

void C_GameHandler::FreeServerNet()
{
	delete m_pclNetMsg;       m_pclNetMsg = NULL;
	delete m_pclNetMsgServer; m_pclNetMsgServer = NULL;
	delete m_pclServSocket;   m_pclServSocket = NULL;
}

bool C_GameHandler::FileCheck()
{
	FILE *pFile = fopen(C_Global::szResourceFile, "rb");
	if(!pFile) {
		sprintf(m_szErrorMsg, "Missing file: %s", C_Global::szResourceFile);
		return false;
	}
	fclose(pFile);

	return true;
}

bool C_GameHandler::InitInstance()
{
	bool bFullscreen = C_Settings::iFullscreen!=0;

	//opengl (and windows)
	int iTest;
#ifdef WIN32
	m_pclGraph = new C_GraphWrapperGL(WindowProc, LoadIcon(GetModuleHandle(NULL), (LPCTSTR)IDI_ICON1));
#else
	m_pclGraph = new C_GraphWrapperGL();
#endif
	iTest = m_pclGraph->SetMode(DEFAULT_WINDOW_W, DEFAULT_WINDOW_H, bFullscreen, true);
	if(!iTest) {
		sprintf(m_szErrorMsg, "Error initializing OpenGL\n%s", m_pclGraph->GetErrorMsg());
		return false;
	}
	char szTitle[64];
	sprintf(szTitle, "GalaxyForces v%d.%02d", PROGRAM_VERSION_MAJOR, PROGRAM_VERSION_MINOR);
	m_pclGraph->SetWindowTitle(szTitle);

#ifdef WIN32
	m_pclControl = new C_Control(m_pclGraph->GetWindow(), C_Settings::iPadDeviceNr, false);
#else
	m_pclControl = new C_Control(C_Settings::iPadDeviceNr);
#endif
	m_pclControl->SetRate(60, 400);
	if(C_Settings::iPadEnabled) {
		m_pclControl->SetJoystickKeyMapping(&C_Settings::stPadMapping);
	}

	char szServerApp[MAX_PATH];
#if defined WIN32
	sprintf(szServerApp, "%s/server.exe", C_Global::szExePath); //windows
#elif defined __APPLE__
	sprintf(szServerApp, "%s/Server", C_Global::szExePath); //mac
#else
	sprintf(szServerApp, "%s/server", C_Global::szExePath); //linux
#endif

#ifdef WIN32
	STARTUPINFO stStartUp;
	PROCESS_INFORMATION stProcess;
	memset(&stStartUp, 0, sizeof(stStartUp));
	stStartUp.cb = sizeof(stStartUp);
	stStartUp.wShowWindow = SW_HIDE;
	stStartUp.dwFlags = STARTF_USESHOWWINDOW;
	CreateProcess(szServerApp, "server.exe -silent", NULL, NULL, false, NORMAL_PRIORITY_CLASS | CREATE_NO_WINDOW, NULL, NULL, &stStartUp, &stProcess);
	m_hServer = stProcess.hProcess;
#else
	char szLine[MAX_PATH];
	sprintf(szLine, "%s -silent", szServerApp);
	FILE *hPipe = (FILE *)popen(szLine, "r");
	if(hPipe) {
		if(fgets(szLine, sizeof(szLine), hPipe)) {
			sscanf(szLine, "%d", &m_hServer);
		}
		pclose(hPipe);
	}
#endif

	//create dialogs
	m_pclGVDlg = new C_GameView(m_pclGraph, m_pclMemberAch);

	return true;
}

void C_GameHandler::StartMusic(bool i_bSetVolumeOnly)
{
	float fVol = 0.0f;
	if(m_bMusicPlaying) fVol = (float)((m_iGameState>2 ? 0.25 : 0.5)*C_Settings::dMasterVolume);
	if(m_pclSound) {
		if(i_bSetVolumeOnly) m_pclSound->SetVolume(m_pMusicChn, fVol);
		else m_pMusicChn = m_pclSound->Play(10, fVol, 0.0f, true, 34500);
	}
}

bool C_GameHandler::Init()
{
	m_bInitDone = false;

	bool bTest = FileCheck();
	if(!bTest) goto out;

	bTest = InitInstance();
	if(!bTest) goto out;

	m_pclSound = new C_Sound();
	if(m_pclSound->Init()) {
		m_pclSound->SetMasterVolume((float)(0.75*C_Settings::dMasterVolume));
		m_pclSound->LoadWav((char*)"Galaxy tune - Omar soriano.ogg", C_Global::szResourceFile, 10);
		StartMusic();

		m_pclSound->LoadWav((char*)"bulletfire.wav",   C_Global::szResourceFile, 0); //replace this sound
		m_pclSound->LoadWav((char*)"explosion.wav",    C_Global::szResourceFile, 1);
		m_pclSound->LoadWav((char*)"engine.wav",       C_Global::szResourceFile, 2);
		m_pclSound->LoadWav((char*)"landonzone.wav",   C_Global::szResourceFile, 3);
		m_pclSound->LoadWav((char*)"bullethit.wav",    C_Global::szResourceFile, 4);
		m_pclSound->LoadWav((char*)"cp.wav",           C_Global::szResourceFile, 5);
		m_pclSound->LoadWav((char*)"cp_end.wav",       C_Global::szResourceFile, 6);
		m_pclSound->LoadWav((char*)"bounceonzone.wav", C_Global::szResourceFile, 7);
		m_pclSound->LoadWav((char*)"loadcargo.wav",    C_Global::szResourceFile, 8);
		m_pclSound->LoadWav((char*)"unloadcargo.wav",  C_Global::szResourceFile, 9);
	}

	m_pclHiscore = new C_ShowHiscore();
	{
		char szMessage[128];
		char szExternalIP[16];
		int iVersionMajor, iVersionMinor;
		bool bPortOk;
		if(m_pclMemberAch->GetInfo(szExternalIP, &iVersionMajor, &iVersionMinor, &bPortOk)) {
			m_pclGVDlg->SetExternalIP(szExternalIP);
			if(iVersionMajor > PROGRAM_VERSION_MAJOR || (iVersionMajor==PROGRAM_VERSION_MAJOR && iVersionMinor>PROGRAM_VERSION_MINOR)) {
				sprintf(szMessage, "A newer game version is available (v%d.%02d)", iVersionMajor, iVersionMinor);
				m_pclGVDlg->DisplayShortMessage(szMessage);
			}
			//cannot be tested with the sourceforge web server
			/*if(!bPortOk) {
				sprintf(szMessage, "Port %d is not ok. You cannot play with others.", C_Settings::iServerPort);
				m_pclGVDlg->DisplayShortMessage(szMessage);
			}*/
		}
	}
	m_bInitDone = true;
out:
	return m_bInitDone;
}

void C_GameHandler::SetSize(int w, int h)
{
	m_pclGraph->SetSize(w, h);
	RedrawScreen();
	RedrawScreen();
}

void C_GameHandler::HandleServerMsg()
{
	if(m_pclNetMsgServer) {
		S_NetMsg1 stMsg1;
		S_NetMsg2 stMsg2;
		char szChatString[2048];
		S_HiScore stList;
		int iResult;
		do {
			iResult = m_pclNetMsgServer->GetMsg(&stMsg1, &stMsg2, &stList, szChatString, sizeof(szChatString));
			if(iResult == 1) {
				switch(stMsg2.iType) {
					case N_MSG2_ADDGAME:
						m_pclGVDlg->AddGame(stMsg2.szGameName, stMsg2.iGameStatus != 0);
						break;
					case N_MSG2_REMOVEGAME:
						m_pclGVDlg->RemoveGame(stMsg2.szGameName);
						break;
					case N_MSG2_GAMESTATUS:
						m_pclGVDlg->SetGameStatus(stMsg2.szGameName, stMsg2.iGameStatus != 0);
						break;
					case N_MSG2_JOIN_NEW_R:
						m_iPlayerId = stMsg2.iPlayerId;
						m_iGameId = stMsg2.iGameId;
						m_pclNetMsgServer->m_iPlayerId = m_iPlayerId;
						m_pclNetMsgServer->m_iGameId = m_iGameId;
						if(m_iPlayerId>=0) m_pclGVDlg->SetJoinNewOK(m_iPlayerId);
						//else error

						if(m_bMaster) SendOptions(); //give server the default options

						RedrawScreen();
						break;
					case N_MSG2_ADDPLAYER:
						m_pclGVDlg->AddPlayer(stMsg2.szPlayerName, stMsg2.iPlayerId);
						break;
					case N_MSG2_OPTIONS:
						{
							m_iLevelType = stMsg2.iLevelType;
							bool bNoShooting = (stMsg2.iOptions & OPT_NOSHOOTING)!=0;
							bool bFlyThrough = (stMsg2.iOptions & OPT_FLYTHROUGH)!=0;
							bool bNoBulletHit = (stMsg2.iOptions & OPT_NOBULLETHIT)!=0;
							m_pclGVDlg->SetGameOpt(stMsg2.szLevelName, stMsg2.iNumLaps, bNoShooting, bFlyThrough, bNoBulletHit, m_iLevelType, stMsg2.iNumAI);
						}
						break;
					case N_MSG2_DISCONNECT:
						if(m_pclViewport) m_pclViewport->RemoveEnemy(stMsg2.iPlayerId);
						if(m_pclGVDlg) m_pclGVDlg->RemovePlayer(stMsg2.iPlayerId);
						if(stMsg2.iPlayerId==m_iPlayerId) {
							FreeGameObjects();
							FreeServerNet();
							m_iGameState = 0;
							if(m_pclGVDlg) m_pclGVDlg->DisplayShortMessage((char*)"You were kicked!");
						}
						break;
					case N_MSG2_START_R:
						StartGame();
						break;
					case N_MSG2_NUMKILLS:
						if(m_pclViewport) m_pclViewport->SetKillsOrLapsLeft(stMsg2.iNumKills);
						break;
					case N_MSG2_CARGO_R:
						if(m_pclViewport) m_pclViewport->HandleCargoReply(&stMsg2);
						break;
					case N_MSG2_CARGORESPAWN:
						if(m_pclViewport) m_pclViewport->HandleCargoRespawn(&stMsg2);
						break;
					case N_MSG2_LIFE_R:
						if(m_pclViewport) m_pclViewport->HandleLifeReply(&stMsg2);
						break;
					case N_MSG2_GAMEOVER:
						m_iGameState = 6;
						break;
					case N_MSG2_SCORE:
						if(!m_bIsReplay) {
							bool bIsAI = false;
							for(int i=0; i<m_iNumAI; i++) {
								if(stMsg2.iPlayerId==m_aiAIIndex[i]) {
									m_pclGVDlg->SetPlayerScore(i, stMsg2.iScore, true);
									bIsAI = true;
									break;
								}
							}
							if(!bIsAI) m_pclGVDlg->SetPlayerScore(stMsg2.iPlayerId, stMsg2.iScore, false);
						}
						break;
					case N_MSG2_NEWMASTER:
						m_bMaster = true;
						m_pclGVDlg->NewMaster();
						if(m_pclViewport) m_pclViewport->NewMaster();
						break;
				}
			} else if(iResult == 2) { //HISCORE
				m_pclHiscore->SetHiscoreList(&stList);

				if(m_iGameState <3) RedrawScreen();
			} else if(iResult == 3) { //CHAT
				//always add to dialog
				m_pclGVDlg->AddChatString(szChatString);
				//add to viewport if in game
				if(m_pclViewport && m_iGameState>=5) {
					m_pclViewport->AddChatString(szChatString);
				}
			} else if(iResult == 4) { //PLAYER->PLAYER via server
				m_pclNetMsg->AddMsg1(&stMsg1);
			}
		} while(iResult);
	}
}

int C_GameHandler::GetPlayerScore(int i_iIndex) //valid for normal player and AI
{
	int i, iScore = BAD_SCORE;
	bool bAI = false;
	for(i=0; i<m_iNumAI; i++) {
		if(m_aiAIIndex[i]==i_iIndex) {
			bAI = true;
			break;
		}
	}
	if(i_iIndex==m_iPlayerId || (bAI && m_bMaster)) {
		if(m_pclViewport->IsRace()) {
			iScore = m_pclViewport->GetRaceTime(i_iIndex);
		} else {
			m_pclViewport->GetScore(i_iIndex, &iScore);
		}
	}
	return iScore;
}

void C_GameHandler::Update()
{
	S_NetMsg2 stMsg2;

	m_pclControl->Update();
	UpdateKeypress();
	UpdateMouse();

	if(m_bRestartPressed) {
		if(/*m_iGameState > 4*/m_iGameState==5 && m_pclGVDlg->NumPlayers(false) == 1) {
			FreeGameObjects();
			stMsg2.iType = N_MSG2_START;
			m_pclNetMsgServer->SendMsg2(&stMsg2); //causes StartGame() and server restart level
			m_pclNetMsgServer->SetReplayMode(false);

			m_iGameState = 3;
		}
		m_bRestartPressed = false;
	}

	//handle escape pressed
	if(m_bEscPressed) {
		if(m_iGameState <4) { //in dialog => quit
			//send disconnect msg
			stMsg2.iType = N_MSG2_DISCONNECT;
			stMsg2.iPlayerId = (int16_t)m_iPlayerId;
			if(m_pclNetMsgServer) m_pclNetMsgServer->SendMsg2(&stMsg2);
#ifdef WIN32
			PostMessage(m_pclGraph->GetWindow(), WM_CLOSE, 0, 0);
#else
			exit(0); //should be done with a proper shutdown...
#endif
		}
		if(m_iGameState > 4) { //in game/replay => goto menu
			FreeGameObjects();
			//send leave msg
			if(!m_bIsReplay) {
				stMsg2.iType = N_MSG2_LEAVE;
				stMsg2.iPlayerId = (int16_t)m_iPlayerId;
				if(m_pclNetMsgServer) m_pclNetMsgServer->SendMsg2(&stMsg2);
				m_bIsReplay = false;
			}
			m_pclNetMsgServer->SetReplayMode(false);

			m_iGameState = 1;
			m_pclGVDlg->SetConnected(1, true);
		}
		m_bEscPressed = false;
	}

	//see if server connection succeded
	int iConnected = m_pclGVDlg->GetConnected();
	if(iConnected==-1) { //waiting
		if(m_pclServSocket && m_pclServSocket->WaitForWriteEvent(0) && m_pclServSocket->ConnectedToIP()) {
			m_pclNetMsgServer = new C_NetMsgServer(m_pclServSocket);
			m_pclNetMsgServer->SetReplayObject(m_pclCurReplay);
			m_pclGVDlg->SetConnected(1);
		} else if(m_pclTimer2->PeekInterval()>5000) {
			m_pclGVDlg->SetConnected(0);
		}
	} else {
		m_pclTimer2->Reset();
		if(iConnected==1 && (!m_pclServSocket || !m_pclServSocket->GetSocketActive())) {
			if(m_iGameState > 4) { //in game/replay => goto menu
				FreeGameObjects();
			}
			m_iGameState = 0;
			m_pclGVDlg->SetConnected(0);
		}
	}

	//update sound every 40 ms
	if(m_pclSound) {
		if(m_pclTimer3->PeekInterval()>40) {
			m_pclSound->Update();
			m_pclTimer3->Reset();
		}
	}

	//handle msg from server (or replay)
	if(m_iGameState!=5 && m_iGameState!=7) HandleServerMsg();

	//game state machine
	double dTime = m_pclTimer4->GetInterval();
	switch(m_iGameState) {
		case 0: //pre init
			m_pclGVDlg->SetDlg1State();
			m_iGameState++;
			break;
		case 1: //init game
			StartMusic(true);
			delete m_pclBKImg;
			m_pclBKImg = new C_Image((char*)"menu_bk.jpg", C_Global::szResourceFile);
			RedrawScreen();
			m_pclGVDlg->Show(true);
			m_iGameState++;
			m_pclTimer->Reset();
			break;
		case 2: //connect
			m_pclGVDlg->UpdateInput();
			ProcessDialog(); //gamestate is changed from this method
			RedrawScreen(); //draw and make the new frame visible
			mssleep(10); //limit cpu usage

			//update fade out (removes click/pop in looping sounds that are stopped)
			//this is normally done in m_pclViewport->Update() but is needed at this
			// state to have the sound stopped properly between games
			C_ShipPlayer::UpdateFadeOut(dTime);

			break;
		case 3: //gamestart init
			StartMusic(true);
			delete m_pclBKImg;
			m_pclBKImg = new C_Image((char*)"game_bk.jpg", C_Global::szResourceFile);
			RedrawScreen();
			RedrawScreen(); //to clear buffer 2
			if(!m_bIsReplay) m_pclCurReplay->Reset();
			m_pclGVDlg->ActivateSaveReplay(false);
			m_iGameState++;
			m_iJoystickSampleCounter = m_iJoystickSampleX = 0;
			m_pclTimer->Reset();
			break;
		case 4: //wait for 2 seconds
			if(m_pclTimer->PeekInterval() >= 2000) {
				m_iGameState++;
			}
			mssleep(10); //limit cpu usage

			//linux: faulty joystick workaround
			//use this time to check if left is constantly "pressed" on the joystick
			//in which case we disable the joystick
			//this is because in some linux other devices than a joystick may be present,
			//and in its idle state it sends -1 on the x axis, and there is no other way
			//to detect that its a bad joystick
			{
				S_JoystickState stState;
				m_pclControl->GetJoystickStateDigital(&stState);
				m_iJoystickSampleX += stState.iX;
				m_iJoystickSampleCounter++;
				if(m_iGameState==5) {
					//check that left has been almost constantly pressed
					if(m_iJoystickSampleX < (-m_iJoystickSampleCounter+10)) {
						//disable joystick
						m_pclControl->SetJoystickKeyMapping(NULL);
					}
				}
			}
			break;
		case 5: //play game
			m_dAccTime += dTime;
			while(m_dAccTime>TIMESLOT) {
				m_dAccTime -= TIMESLOT;
				HandleServerMsg();
				/*int iResult =*/ m_pclViewport->Update(TIMESLOT);
				m_pclCurReplay->IncTimeSlot();
			}
			m_pclViewport->Draw();
			m_pclGraph->Flip(); //make the new frame visible
			break;
		case 6: //ending, send score
			m_pclViewport->SetGameFinished();

			//send score from player and AI
			stMsg2.iType = N_MSG2_SCORE;
			for(int i=0; i<8; i++) {
				stMsg2.iScore = GetPlayerScore(i);
				if(stMsg2.iScore>BAD_SCORE) {
					stMsg2.iPlayerId = (int16_t)i;
					if(m_pclNetMsgServer) m_pclNetMsgServer->SendMsg2(&stMsg2);
				}
			}
			m_iScore = GetPlayerScore(m_iPlayerId);

			m_pclTimer->Reset();
			m_iGameState++;
			//no break
		case 7: //wait 6+ sec while playing (enough time to receive all other players score)
			if(m_pclTimer->PeekInterval() >= 6500) {
				m_iGameState++;
			}

			//score is recevied from server N_MSG2_SCORE messages above

			//play game, ignore results
			m_dAccTime += dTime;
			while(m_dAccTime>TIMESLOT) {
				m_dAccTime -= TIMESLOT;
				HandleServerMsg();
				m_pclViewport->Update(TIMESLOT);
				m_pclCurReplay->IncTimeSlot();
			}
			m_pclViewport->Draw();
			m_pclGraph->Flip(); //make the new frame visible
			break;
		case 8: //state 1 -> get, present score
			if(!m_bIsReplay) {
				m_pclGVDlg->ActivateSaveReplay(true);
				delete m_pclLastReplay;
				m_pclLastReplay = new C_Replay(m_pclCurReplay);
				char szLevel[40];
				if(m_pclGVDlg->GetHiscoreMapName(szLevel)) {
					if(m_pclViewport->MapType()!=MAP_MISSION_COOP || m_bMaster) {
						if (m_pclMemberAch->CheckHiscore(szLevel, m_iScore)) {
							m_pclMemberAch->SendHiscore(szLevel, m_iScore, m_pclLastReplay);
							m_pclGVDlg->UpdateAchievements();
						}
					}
				}
			}
			m_pclNetMsgServer->SetReplayMode(false);
			delete m_pclViewport; m_pclViewport = NULL;
			m_iGameState = 1;
			break;
	}

#if ((defined SAVE_SCREENS) || (defined SAVE_MOVIE))
	SaveScreens();
#endif
}

void C_GameHandler::SendOptions()
{
	S_NetMsg2 stMsg2;

	stMsg2.iType = N_MSG2_OPTIONS;
	bool bNoShooting, bFlyThrough, bNoBulletHit;
	int iNumAI;
	m_pclGVDlg->GetGameOpt(stMsg2.szLevelName, &stMsg2.iNumLaps, &bNoShooting, &bFlyThrough, &bNoBulletHit, &m_iLevelType, &iNumAI);
	stMsg2.iOptions = 0;
	if(bNoShooting)  stMsg2.iOptions |= OPT_NOSHOOTING;
	if(bFlyThrough)  stMsg2.iOptions |= OPT_FLYTHROUGH;
	if(bNoBulletHit) stMsg2.iOptions |= OPT_NOBULLETHIT;
	stMsg2.iLevelType = m_iLevelType;
	stMsg2.iNumAI = (uint16_t)iNumAI;
	m_pclNetMsgServer->SendMsg2(&stMsg2);
}

int C_GameHandler::ProcessDialog()
{
	S_NetMsg2 stMsg2;

	if(m_iLoginState) {
		m_iLoginState = 0;
		char szUser[MAXPLAYERNAMELENGTH], szPass[MAXPASSWORDLENGTH];
		m_pclGVDlg->GetUserPass(szUser, szPass);
		if(m_pclMemberAch->LoginCheck(szUser, szPass)) { //blocking
			m_pclGVDlg->SetLogedInState(1);
			m_pclMemberAch->FetchLevelInfoList(); //blocking
		} else m_pclGVDlg->SetLogedInState(0);
	}

	int iPollMsg = m_pclGVDlg->Poll();
	switch(iPollMsg) {
		case D0_MLOGIN:
			m_iLoginState = 1; //to be done next loop
			break;
		case D0_MCREATE:
			C_Global::OpenBrowser("http://www.galaxy-forces.com/member_register.php");
			break;
		case D1_MWEB:
			C_Global::OpenBrowser("http://www.galaxy-forces.com/index.php#hiscore");
			break;
		case D2_LEAVE:
			stMsg2.iType = N_MSG2_LEAVE;
			stMsg2.iPlayerId = (int16_t)m_iPlayerId;
			m_pclNetMsgServer->SendMsg2(&stMsg2);
			break;
		case D1_CONNECT:
			FreeServerNet(); //if exists
			m_pclServSocket = new C_Socket();
			m_pclNetMsg = new C_NetMsg(-1, 0);
			char szConnect[32];
			m_pclGVDlg->GetConnectString(szConnect);
			m_pclServSocket->Connect(szConnect, C_Settings::iServerPort, 0);
			break;
		case D1_NEW:
			m_bMaster = true;
			stMsg2.iType = N_MSG2_NEW;
			m_pclGVDlg->GetPlayerName(stMsg2.szPlayerName);
			m_pclGVDlg->GetGameName(stMsg2.szGameName);
			m_pclNetMsgServer->SendMsg2(&stMsg2);
			break;
		case D1_JOIN:
			m_bMaster = false;
			stMsg2.iType = N_MSG2_JOIN;
			m_pclGVDlg->GetPlayerName(stMsg2.szPlayerName);
			m_pclGVDlg->GetGameName(stMsg2.szGameName);
			m_pclNetMsgServer->SendMsg2(&stMsg2);
			break;
		case D2_CHATSEND:
			{
				char szChatString[2048];
				stMsg2.iType = N_MSG2_CHAT;
				m_pclGVDlg->GetChatString(szChatString);
				m_pclNetMsgServer->SendChatMsg(szChatString);
			}
			break;
		case D2_START:
			stMsg2.iType = N_MSG2_START;
			m_pclNetMsgServer->SendMsg2(&stMsg2);
			break;
		case D2_OPTCHANGE:
			if(m_bMaster) SendOptions();
			break;
		case D2_KICK:
			stMsg2.iType = N_MSG2_DISCONNECT;
			stMsg2.iGameId   = (int16_t)m_iGameId;
			stMsg2.iPlayerId = (int16_t)m_pclGVDlg->GetPlayerToKick();
			m_pclNetMsgServer->SendMsg2(&stMsg2);
			break;

		case D2_SAVEREPLAY:
			{
				S_ReplayHeader stReplayInfo;
				char szFilename[MAX_PATH];
				char szDate[24], szTime[24];

				C_Logger::GetTimeString(szTime, false, false);
				C_Logger::GetDateString(szDate, false);
				m_pclLastReplay->GetGameInfo(&stReplayInfo);
				char *pDot = strrchr(stReplayInfo.szMapFile, '.');
				if(pDot) *pDot = 0; //stop before ".des"

				sprintf(szFilename, "%s/replays/%s_%s_%s.rpl",
					C_Global::szUserPath, szDate, szTime, stReplayInfo.szMapFile);
#ifdef WIN32
				//workaround for avast antivirus (140120-0) triggered ->
				//vary the length of the string data to make the problem go away if present
				printf("11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111\n");
				printf("11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111211111111111111\n");
				printf("11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111311111111111111\n");
				printf("11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111411111111111111\n");
#endif

				m_pclLastReplay->Save(szFilename);
			}
			break;

		case D3_BEGIN:
			{
				C_Hiscores *pclHiscores = m_pclMemberAch->FetchHiscores();
				if(pclHiscores) {
					int i, iSize = (int)pclHiscores->size();
					for(i=0; i<iSize; i++) {
						S_HiscoreElement *pstEl = &(*pclHiscores)[i];
						m_pclGVDlg->AddExternalReplay(pstEl->szLevel, pstEl->szName, pstEl->iPlace, pstEl->iScoreMs);
					}
					delete pclHiscores;
				}
			}
			break;
		case D3_REPLAY:
			{
				char szFilename[MAX_PATH];
				char szName[32];
				bool bPlay = false;
				int iPlace = m_pclGVDlg->GetReplayFile(szFilename, szName);
				if(iPlace>=0) {
					//webserver replay
					C_Replay *pclReplay = m_pclMemberAch->FetchReplay(szFilename, szName);
					if(pclReplay) {
						delete m_pclCurReplay;
						m_pclCurReplay = pclReplay;
						bPlay = true;
					}
				} else if(strcmp(szFilename, "Last game")!=0) {
					//local file
					bPlay = m_pclCurReplay->Load(szFilename);
				} else {
					//last game
					delete m_pclCurReplay;
					m_pclCurReplay = new C_Replay(m_pclLastReplay);
					m_pclCurReplay->ResetBeforePlay();
					bPlay = true;
				}
				if(bPlay) StartReplay();
			}
			break;
		case -1: //exitbutton
			m_bEscPressed = true;
			break;
	}
	return 1;
}

void C_GameHandler::StartReplay()
{
	int i;
	S_ReplayHeader stGameInfo;

	m_pclNetMsg->SetServerComm(m_pclNetMsgServer);
	m_pclNetMsgServer->SetReplayObject(m_pclCurReplay);
	m_pclNetMsgServer->SetReplayMode(true);

	m_pclCurReplay->GetGameInfo(&stGameInfo);
	m_iLevelType = stGameInfo.iLevelType;

	delete m_pclViewport; //if it existed
	m_pclViewport = new C_Viewport(stGameInfo.szMapFile, -1, (char*)"", false, m_pclNetMsg, m_pclNetMsgServer);

	m_pclViewport->SetShootingMode(!stGameInfo.bNoShooting);
	m_pclViewport->SetPlayerPlayerCollision(!stGameInfo.bFlyThrough, !stGameInfo.bNoBulletsHit);

	for(i=0; i<8; i++) {
		if(stGameInfo.szPlayerName[i][0]!=0) { //all players are treated as enemies in replay
			m_pclViewport->AddEnemy(i, stGameInfo.szPlayerName[i]);
		}
	}
	m_iNumAI = 0;
	m_iPlayerId = 0;

	m_pclViewport->SetNumLifes(m_pclViewport->IsRace() ? -1 : NUM_LIFES);

	m_iGameState++;
	m_pclGVDlg->Show(false);

	m_pclTimer->Reset();
	m_dAccTime = 0.0;

	m_bIsReplay = true;
}

void C_GameHandler::StartGame()
{
	int i, iNumLaps;
	char szPlayerName[20];
	char szFilename[100];

	m_pclNetMsg->SetServerComm(m_pclNetMsgServer);
	m_pclServSocket->SetNoDelay();

	bool bNoShooting, bFlyThrough, bNoBulletsHit;
	m_pclGVDlg->GetGameOpt(szFilename, &iNumLaps, &bNoShooting, &bFlyThrough, &bNoBulletsHit, &m_iLevelType, &m_iNumAI);
	m_pclGVDlg->GetPlayerName(m_iPlayerId, m_szPlayerName);

	delete m_pclViewport; //if it existed
	m_pclViewport = new C_Viewport(szFilename, m_iPlayerId, m_szPlayerName, m_bMaster, m_pclNetMsg, m_pclNetMsgServer);

	m_pclViewport->SetShootingMode(!bNoShooting);
	m_pclViewport->SetPlayerPlayerCollision(!bFlyThrough, !bNoBulletsHit);

	S_ReplayHeader stGameInfo;
	memset(&stGameInfo, 0, sizeof(stGameInfo));
	stGameInfo.iVersion = 1;
	stGameInfo.iLevelType = m_iLevelType;
	for(i=0; i<8; i++) {
		m_pclGVDlg->GetPlayerName(i, szPlayerName);
		if(szPlayerName[0]!=0 && i!=m_iPlayerId) { //player exists, and is not you
			m_pclViewport->AddEnemy(i, szPlayerName);
		}
		if(szPlayerName[0]!=0) {
			stGameInfo.iNumPlayers++;
			strcpy(stGameInfo.szPlayerName[i], szPlayerName);
		}
	}
	for(i=0; i<m_iNumAI; i++) {
		m_aiAIIndex[i] = m_pclGVDlg->GetIndexForAI(i);
		sprintf(szPlayerName, "AI %d", i);
		if(m_bMaster) {
			m_pclViewport->AddAI(m_aiAIIndex[i], szPlayerName);
		} else {
			m_pclViewport->AddEnemy(m_aiAIIndex[i], szPlayerName, true);
		}
		strcpy(stGameInfo.szPlayerName[m_aiAIIndex[i]], szPlayerName);
	}
	m_pclViewport->SetNumLifes(m_pclViewport->IsRace() ? -1 : NUM_LIFES);

	stGameInfo.bNoBulletsHit = bNoShooting;
	stGameInfo.bFlyThrough = bFlyThrough;
	stGameInfo.bNoShooting = bNoBulletsHit;
	strcpy(stGameInfo.szMapFile, szFilename);
	m_pclCurReplay->SetGameInfo(&stGameInfo);

	//m_iGameState++;
	m_iGameState = 3;
	m_pclGVDlg->Show(false);

	m_pclTimer->Reset();
	m_dAccTime = 0.0;

	m_bIsReplay = false;
}

void C_GameHandler::RedrawScreen()
{
	if(!m_bInitDone || !m_pclBKImg) return;
	
	static S_FColor stBlack = {0.0f, 0.0f, 0.0f, 1.0f};

	//find out bk rects
	int w = m_pclGraph->GetModeWidth();
	int h = m_pclGraph->GetModeHeight();
	int w2, h2;
	m_pclBKImg->GetInfo(&w2, &h2, NULL);
	S_Rect stSR = {0,0,w2,h2};
	S_Rect stDR = {0,0,w,h};
	if(w2>w) {
		stSR.width = w;
		stSR.x = (w2-w)/2;
	} else if(w>w2) {
		stDR.width = w2;
		stDR.x = (w-w2)/2;
	}
	if(h2>h) {
		stSR.height = h;
		stSR.y = (h2-h)/2;
	} else if(h>h2) {
		stDR.height = h2;
		stDR.y = (h-h2)/2;
	}

	m_pclGraph->SetColor(&stBlack);
	m_pclGraph->Rect(NULL);
	m_pclGraph->Blt(m_pclBKImg, &stSR, &stDR, false);
	if(m_iGameState<3) {
		if(m_pclHiscore) m_pclHiscore->Draw(w, h);
		if(m_pclGVDlg) m_pclGVDlg->Draw();
	}
	m_pclGraph->Flip();
}

void C_GameHandler::SaveScreens()
{
	//screen size must be constant during save
	static bool bSaveSound = false;
	static bool bLastSave = false;
	if(m_bSaveMovie || m_bSaveScreen) {
		if(!m_pclMovieMaker) {
			m_pclMovieMaker = new C_MovieMaker(C_Global::szUserPath, m_pclGraph, m_bSaveMovie ? m_pclSound : NULL);
			if(m_bSaveMovie) {
				bSaveSound = true;
				//after sound is restarted, start playing music
				StartMusic();
			} else {
				bSaveSound = false;
			}
		}
		bLastSave = true;
		m_bSaveScreen = false;

		if(m_bSaveMovie) {
			//capture screen in movie
			m_pclMovieMaker->NextFrame();

			//try to limit framerate to ~SAVE_MOVIE_FPS
			m_pclMovieMaker->FrameRateLimit(SAVE_MOVIE_FPS);
		} else {
			//capture screen
			m_pclMovieMaker->SingleFrame();
		}
	} else if(bLastSave) {
		if(m_pclMovieMaker) {
			delete m_pclMovieMaker;
			m_pclMovieMaker = NULL;
			if(bSaveSound) {
				//after sound is restarted, start playing music
				StartMusic();
			}
		}
		bLastSave = false;
	}
}

void C_GameHandler::UpdateKeypress()
{
	//ESC - indicate esc to Update()
	if(m_pclControl->IsPressedOnce(DIK_ESCAPE)) m_bEscPressed = true;
#ifdef SAVE_MOVIE
	//F2 - toggle save movie on/off
	if(m_pclControl->IsPressedOnce(DIK_F2)) m_bSaveMovie = !m_bSaveMovie;
#endif
#ifdef SAVE_SCREENS
	//F3 - save one screen shot
	if(m_pclControl->IsPressedOnce(DIK_F3)) m_bSaveScreen = true;
#endif

	bool bEnteringTextMode = false;
	if(m_pclViewport) bEnteringTextMode = m_pclViewport->InEnteringTextMode();
	if(m_pclGVDlg && !bEnteringTextMode) {
		bEnteringTextMode = m_pclGVDlg->IsAnyEditFocused();
	}
	
	//restart game if in single player
	if(!bEnteringTextMode && m_pclControl->IsPressedOnce(DIK_7)) m_bRestartPressed = true;

	//M - toggle music on/off
	if(!bEnteringTextMode && m_pclSound && m_pclControl->IsPressedOnce(DIK_M)) {
		m_bMusicPlaying = !m_bMusicPlaying;
		StartMusic(true);
	}

	//N - toggle sound on/off
	if(!bEnteringTextMode && m_pclSound && m_pclControl->IsPressedOnce(DIK_N)) {
		m_pclSound->ToggleSoundsOnOff();
	}
}

void C_GameHandler::UpdateMouse()
{
	int iX, iY, iExtraVal;
	if(m_pclControl) {
		int iEvent = m_pclControl->GetMouseEvent(&iX, &iY, &iExtraVal);
		if(m_pclGVDlg) {
			switch(iEvent) {
				case EVT_MOUSE_MOVE:       m_pclGVDlg->MouseMove(iX, iY); break;
				case EVT_MOUSE_LEFT_DOWN:  m_pclGVDlg->MouseDown(iX, iY); break;
				case EVT_MOUSE_LEFT_UP:    m_pclGVDlg->MouseUp(iX, iY); break;
				case EVT_MOUSE_WHEEL_UP:   m_pclGVDlg->MouseWheel(iExtraVal); break;
				case EVT_MOUSE_WHEEL_DOWN: m_pclGVDlg->MouseWheel(iExtraVal); break;
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////

C_GameHandler *g_pclGameHandler = NULL;

#ifdef WIN32

//message-handler
LRESULT CALLBACK WindowProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam)
{
	switch(iMessage) {
		case WM_MOUSEMOVE:
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MOUSEWHEEL:
			if(g_pclGameHandler->m_pclControl) g_pclGameHandler->m_pclControl->XLateMouseEvent(iMessage, wParam, lParam);
			break;

		case WM_SIZE:
			if(g_pclGameHandler) g_pclGameHandler->SetSize(LOWORD(lParam), HIWORD(lParam));
			break;

		case WM_DESTROY:
			PostQuitMessage(0);
			break;
		case WM_ERASEBKGND:
			return 1; //never erase, but always say that we do
		default:
			return DefWindowProc(hWnd, iMessage, wParam, lParam);
	}
	return 0;
}

//main
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	bool bTest; //result var
	
	C_Global::Init((char*)"resource.dat", (char*)".GalaxyV2");
	C_Settings::Init((char*)"config.txt"); //read global settings (or init default if file not found)

	char szReplayPath[MAX_PATH];
	sprintf(szReplayPath, "%s/replays", C_Global::szUserPath);
	C_Resource::CreateDir(szReplayPath);

	g_pclGameHandler = new C_GameHandler();
	bTest = g_pclGameHandler->Init();
	if(!bTest) {
		MessageBox(NULL, g_pclGameHandler->GetErrorMsg(), "Error", MB_OK);
		goto out;
	}

	//messageloop
	MSG stMsg;
	while(1) {
		if(PeekMessage(&stMsg, NULL, 0, 0, PM_NOREMOVE)) {
			if(!GetMessage(&stMsg, NULL, 0, 0)) goto out;

			TranslateMessage(&stMsg);
			DispatchMessage(&stMsg);
		} else {
			g_pclGameHandler->Update();
		}
	}
out:
	delete g_pclGameHandler;
#ifdef _DEBUG
	_CrtDumpMemoryLeaks();
#endif

	return 0;
}
#else

#ifdef __APPLE__
bool HandleEvents() //returns true on exit
{
	static int iLastW = 0, iLastH = 0;
	int iW, iH;
	if(!FsPollDevice()) return true;
	FsGetWindowSize(iW, iH);
	if(iW!=iLastW || iH!=iLastH) {
		if(g_pclGameHandler) g_pclGameHandler->SetSize(iW, iH);
		iLastW = iW; iLastH = iH;
	}
	if(g_pclGameHandler && g_pclGameHandler->m_pclControl) g_pclGameHandler->m_pclControl->PollInput();
	
    return false;
}

void Redraw()
{
	int iW, iH;
	FsGetWindowSize(iW, iH);
	if(g_pclGameHandler) g_pclGameHandler->SetSize(iW, iH);
}
#else

bool HandleEvents() //returns true on exit
{
	XEvent stEvent;
	XEvent *pstEvent = NULL;

	C_GraphWrapperGL *pclGraph = C_GraphWrapperGL::GetGraphWrapperObject();
	Display *pclDisplay = pclGraph->GetDisplay();
	while(XPending(pclDisplay)) {
		XNextEvent(pclDisplay, &stEvent); //blocks until an event is available

		switch(stEvent.type) {
			case KeyPress:
			case KeyRelease:
				pstEvent = &stEvent;
				break;

			case MotionNotify:
			case ButtonPress:
			case ButtonRelease:
				pstEvent = &stEvent;
				break;

			case ConfigureNotify:
				if(g_pclGameHandler) g_pclGameHandler->SetSize(stEvent.xconfigure.width, stEvent.xconfigure.height);
				break;

			case ClientMessage:
				//window manager messages
				if((stEvent.xclient.format == 32) && ((unsigned int)stEvent.xclient.data.l[0] == pclGraph->m_Atom_DeleteWindow)) {
					return true;
				}
				break;
			case DestroyNotify: //apparantly does not happen (the above does the job)
				//window has been destroyed
				return true;
		}
		if(g_pclGameHandler && g_pclGameHandler->m_pclControl) g_pclGameHandler->m_pclControl->PollInput(pstEvent);
	}
	return false;
}

#endif

int main(int argc, char *argv[])
{
	bool bTest; //result var

	C_Global::Init((char*)"resource.dat", (char*)".GalaxyV2");
	C_Settings::Init((char*)"config.txt"); //read global settings (or init default if file not found)

	char szReplayPath[MAX_PATH];
	sprintf(szReplayPath, "%s/replays", C_Global::szUserPath);
	C_Resource::CreateDir(szReplayPath);

	g_pclGameHandler = new C_GameHandler();
	bTest = g_pclGameHandler->Init();
	if(!bTest) {
		printf("Error: %s\n", g_pclGameHandler->GetErrorMsg());
		goto out;
	}
#ifdef __APPLE__
	FsSetRedrawCallback(Redraw);
#endif

	//main loop
	while(1) {
		if(HandleEvents()) break;
		g_pclGameHandler->Update();
		mssleep(1);
	}

out:
	delete g_pclGameHandler;
	return 0;
}
#endif
