// C_GUI

#ifndef _GUI_H 
#define _GUI_H

#include "glgraph.h"
#include "font.h"
#include "image.h"
#include "input.h"
#include "timer.h"

#include <vector>

#define TYPE_BK        1
#define TYPE_BUTTON    2
#define TYPE_CHECK     3
#define TYPE_EDIT      4
#define TYPE_LIST      5
#define TYPE_STATIC    6

#define STATE_DISABLED 0
#define STATE_ENABLED  1
#define STATE_ACTIVE   2

#define COLOR_DISABLED STATE_DISABLED
#define COLOR_ENABLED  STATE_ENABLED
#define COLOR_ACTIVE   STATE_ACTIVE
#define COLOR_TEXT     3
#define COLOR_SELECT   4

#define MAX_COLUMNS    6
#define MAX_TEXT       128

struct S_GUIListEl
{
	int iID;

	char szText[MAX_COLUMNS][MAX_TEXT]; //short text for columns
	C_Image *pclIcon;                   //shown in column 0 if used
	int     iIconRect;
};
typedef std::vector<S_GUIListEl> C_GUIListElements;

struct S_GUIObject
{
	int iID;       //unique id

	int iType;     //one of the above
	int iState;    //one of the above. disabled shows in darker color and will not respond to input. active has focus.

	int bPressed; //valid for all, is currently pressed.
	int bChecked; //valid for TYPE_CHECK TYPE_BUTTON (sticky). is checked.
	int bClicked; //valid for TYPE_BUTTON TYPE_EDIT (when enter pressed).
	int bChanged; //valid for TYPE_EDIT, TYPE_LIST (selction)

	int bDeletePressed; //valid for TYPE_LIST (selction)

	S_Rect stRect; //valid for TYPE_BUTTON, TYPE_CHECK, TYPE_EDIT, TYPE_LIST. outer dim.

	char szText[MAX_TEXT]; //short text
	int iMaxText;

	int bOnlyNumbers; //valid for TYPE_EDIT
	int bOnlyEmail;   //valid for TYPE_EDIT (0-9,a-z,A-Z,@,.,-,_, space is also allowed to be able to use this for other than email)
	int bIsPassWord;  //valid for TYPE_EDIT
	int bSticky;      //valid for TYPE_BUTTON

	S_FColor stCol[3];
	S_FColor stTextCol;
	S_FColor stSelectCol;

	C_GUIListElements *pclListEl;
	int iListColumns; //number of columns in list
	int iListColumnW[MAX_COLUMNS];
	int iLastColumnWOriginal;
	int iListSelection; //index of selection
	int iListTop;       //index currently drawn at top of list
	int iListVBarStatus; //bits [isshown][buttonupactive][buttondownactive][buttonuppressed][buttondownpressed]
	int iBarPosY, iBarHeight; //saved scrollbar info
	int iListClickedY, iListClickedSelection; //saved scrollbar clicked info

	bool bListHasSelection;
};
typedef std::vector<S_GUIObject> C_GUIObjects;

class C_GUI
{
public:
	C_GUI();
	~C_GUI();

	//input and output
	bool LoadTemplate(char *i_szFile, char *i_szResource, int i_iScrOffsX, int i_iScrOffsY);
	bool SetPos(int i_iScrOffsX, int i_iScrOffsY);

	void ListAddElement(int i_iListId, S_GUIListEl *i_pstListEl);
	void ListRemoveElementById(int i_iListId, int i_iListEl);
	void ListRemoveElement(int i_iListId, char *i_szListText0);
	void ListRemoveAll(int i_iListId);
	S_GUIListEl *ListGetElementById(int i_iListId, int i_iListEl);
	S_GUIListEl *ListGetElement(int i_iListId, char *i_szListText0);
	S_GUIListEl *ListGetSelectedElement(int i_iListId, int *o_iId = NULL);
	void ListSetSelectedElement(int i_iListId, char *i_szListText0);
	void ListSetHasSelection(int i_iListId, bool i_bHasSelection);
	bool ListWasDeletePressed(int i_iListId);

	S_GUIObject *GetObject(int i_iId);

	void SetText(int i_iId, char *i_szText);
	void GetText(int i_iId, char *o_szText);

	bool IsChecked(int i_iId);
	bool WasClicked(int i_iId);
	bool WasChanged(int i_iId); //TYPE_EDIT, TYPE_LIST (selction), TYPE_CHECK
	void SetFocused(int i_iId); //TYPE_EDIT, TYPE_LIST
	bool IsAnyEditFocused();

	void SetChecked(int i_iId, bool i_bChecked); //TYPE_CHECK, TYPE_BUTTON (sticky)
	void SetActive(int i_iId, bool i_bActive);
	bool IsActive(int i_iId);

	void SetColor(int i_iId, int i_iColId, S_FColor *i_pstColor);
	void SetColor(int i_iId, int i_iColId, float r, float g, float b, float a);

	int GetTextWidth(char *i_szText);
	C_Font *GetFont() {return s_pclFont;}; //to lend the font

	//event handlers
	void MouseMove(int x, int y);
	void MouseDown(int x, int y);
	void MouseUp(int x, int y);
	void MouseWheel(int way);
	void UpdateInput();

	void Draw();
private:
	int m_iScrOffsX, m_iScrOffsY;
	//default colors
	static S_FColor s_stStateCol[3];
	static S_FColor s_stSelectCol;
	static S_FColor s_stTextColor;

	//button
	static C_Image *s_pclButton[2]; //normal, pressed
	static S_Rect s_stButtonR[9];

	//check
	static C_Image *s_pclCheck[2];  //unchecked, checked
	static S_Rect s_stCheckR[9];

	//edit
	static C_Image *s_pclEdit; //normal
	static S_Rect s_stEditR[9];

	//list
	static C_Image *s_pclListVBar; //normal
	static S_Rect s_stListVBarR[6];

	void ListModSelection(int iWay, int i_iID = -1); //-1 for current active
	void ListDelSelection(int i_iID = -1); //-1 for current active

	int ListGetNumRowsUntil(int i_iListId, int i_iLast);
	int ListGetElRows(int i_iWidth, char *i_szText);
	char m_aszRows[10][MAX_TEXT];

	double m_dTime, m_dKeyTimer;

	//common
	static int       s_iObjCount;
	static C_Font    *s_pclFont;
	C_GraphWrapperGL *m_pclGraph;
	C_GUIObjects     *m_pclObj;

	int GetInput();
	void EnterChar(int c);
	C_Control *m_pclInput;

	C_Timer   *m_pclTimer;

	void DrawBorderedRect(S_Rect *i_pstDstRect, S_FColor *i_pstModColor, C_Image *i_pstSrcImage, S_Rect *i_pstSrcRects/*9*/, int i_iBorderSize);
};

#endif
