
// ImageUtils

//low level functions to compare rects and images.
// - pixel for pixel compare
// - rect compare
// - used in collission detection with map and objects

#include "image_utils.h"

namespace ImageUtils
{
	bool Compare(const C_Image *i_pImg1, const C_Image *i_pImg2, const S_Rect *i_pstRect1, const S_Rect *i_pstRect2, const int i_iPixelStep)
	{
		//compares two images with the same size and position for pixel overlapping
		//images and rects cannot be NULL
		if (!i_pstRect1 || !i_pstRect2 || !i_pImg1 || !i_pImg2)
		{
			return false;
		}

		if (i_pstRect1->width != i_pstRect2->width 
			|| i_pstRect1->height != i_pstRect2->height)
		{
			return false;
		}

		const uint8_t *pBuf1, *pBuf2;
		int iPixelSize;
		int iLineBytes1, iLineBytes2;
		int y, x;

		i_pImg1->GetInfo(NULL, NULL, &iPixelSize);
		i_pImg1->GetBufferMemory(&pBuf1);
		i_pImg1->GetLineSizeInfo(NULL, &iLineBytes1, NULL);

		i_pImg2->GetBufferMemory(&pBuf2);
		i_pImg2->GetLineSizeInfo(NULL, &iLineBytes2, NULL);
		iPixelSize &= 0x0f;

		pBuf1 += i_pstRect1->y*iLineBytes1 + i_pstRect1->x*iPixelSize;
		pBuf2 += i_pstRect2->y*iLineBytes2 + i_pstRect2->x*iPixelSize;

		//compare overlapping pixels, (detect pixelcollision)
		uint32_t iMask = 0xffffffff >> (32-8*iPixelSize);
		iPixelSize *= i_iPixelStep;
		iLineBytes1 *= i_iPixelStep;
		iLineBytes2 *= i_iPixelStep;
		for(y=0; y<i_pstRect1->height; y+=i_iPixelStep) {
			register const uint8_t *pBuf1Cpy = pBuf1;
			register const uint8_t *pBuf2Cpy = pBuf2;
			for(x=0; x<i_pstRect1->width; x+=i_iPixelStep) {
				if(((*((uint32_t*)pBuf1Cpy) & iMask)!=0) && ((*((uint32_t*)pBuf2Cpy) & iMask)!=0)) {
					return true;
				}
				pBuf1Cpy += iPixelSize;
				pBuf2Cpy += iPixelSize;
			}
			pBuf1 += iLineBytes1;
			pBuf2 += iLineBytes2;
		}
		return false;
	}

	bool Compare(const C_Image *i_pImg1, int i_iX1, int i_iY1, const C_Image *i_pImg2, int i_iX2, int i_iY2, const S_Rect *i_pstRect1, const S_Rect *i_pstRect2, const int i_iPixelStep)
	{
		//compares two images/rects of variable size and offsets for pixel overlapping
		//options:
		//image==NULL means that the rect is considered to be solid, that is
		//  if both images are NULL and the rectangles overlap the result is true
		//  if one image is NULL and one of the overlapping pixels of the other image is!=0 the result is true
		//rect==NULL means that the rect whole image is taken as the rect
		//x1,y1 and x2,y2 are the offsets of the rects in the world
		const uint8_t *pBuf1 = NULL, *pBuf2 = NULL;
		int iPixelSize = 0;
		int iLineBytes1 = 0, iLineBytes2 = 0;
		int y, x;

		if(!i_pImg1 && !i_pstRect1) return false;
		if(!i_pImg2 && !i_pstRect2) return false;

		int iW, iH;
		S_Rect stR1, stR2, stRCommon;

		//get image/rect info
		stR1.x = i_iX1; stR1.y = i_iY1;
		if(i_pImg1) {
			i_pImg1->GetInfo(NULL, NULL, &iPixelSize);
			i_pImg1->GetBufferMemory(&pBuf1);
			i_pImg1->GetLineSizeInfo(NULL, &iLineBytes1, NULL);
		}
		if(!i_pstRect1) {
			i_pImg1->GetInfo(&iW, &iH, NULL);
			stR1.width = iW; stR1.height = iH;
		} else {
			stR1.width = i_pstRect1->width;
			stR1.height = i_pstRect1->height;
		}
		stR2.x = i_iX2; stR2.y = i_iY2;
		if(i_pImg2) {
			i_pImg2->GetInfo(NULL, NULL, &iPixelSize);
			i_pImg2->GetBufferMemory(&pBuf2);
			i_pImg2->GetLineSizeInfo(NULL, &iLineBytes2, NULL);
		}
		if(!i_pstRect2) {
			i_pImg2->GetInfo(&iW, &iH, NULL);
			stR2.width = iW; stR2.height = iH;
		} else {
			stR2.width = i_pstRect2->width;
			stR2.height = i_pstRect2->height;
		}
		iPixelSize &= 0x0f;

		//stR1, stR2 filled now; make overlapping rect
		if(!IntersectRect2(&stRCommon, &stR1, &stR2)) return false; //no collision
		if(!i_pImg1 && !i_pImg2) return true;                       //rectangle collision

		//calc new stR1, stR2
		stR2.x = stR1.x = stRCommon.x;
		stR2.y = stR1.y = stRCommon.y;
		stR2.width = stR1.width = stRCommon.width;
		stR2.height = stR1.height = stRCommon.height;
		//if(stR2.width <= 0 || stR2.height <=0) return false; //no collision
		stR1.x -= i_iX1; stR1.y -= i_iY1;
		stR2.x -= i_iX2; stR2.y -= i_iY2;
		if(i_pstRect1) { stR1.x += i_pstRect1->x; stR1.y += i_pstRect1->y; }
		if(i_pstRect2) { stR2.x += i_pstRect2->x; stR2.y += i_pstRect2->y; }

		pBuf1 += stR1.y*iLineBytes1 + stR1.x*iPixelSize;
		pBuf2 += stR2.y*iLineBytes2 + stR2.x*iPixelSize;

		//compare overlapping pixels, (detect pixelcollision)
		uint32_t iMask = 0xffffffff >> (32-8*iPixelSize);
		iPixelSize *= i_iPixelStep;
		iLineBytes1 *= i_iPixelStep;
		iLineBytes2 *= i_iPixelStep;
		if(i_pImg1 && i_pImg2) {
			for(y=0; y<stR1.height; y+=i_iPixelStep) {
				register const uint8_t *pBuf1Cpy = pBuf1;
				register const uint8_t *pBuf2Cpy = pBuf2;
				for(x=0; x<stR1.width; x+=i_iPixelStep) {
					if(((*((uint32_t*)pBuf1Cpy) & iMask)!=0) && ((*((uint32_t*)pBuf2Cpy) & iMask)!=0)) {
						return true;
					}
					pBuf1Cpy += iPixelSize;
					pBuf2Cpy += iPixelSize;
				}
				pBuf1 += iLineBytes1;
				pBuf2 += iLineBytes2;
			}
		} else { //img1 or img2
			if(i_pImg2) {
				pBuf1 = pBuf2;
				iLineBytes1 = iLineBytes2;
			}
			for(y=0; y<stR1.height; y+=i_iPixelStep) {
				register const uint8_t *pBuf1Cpy = pBuf1;
				for(x=0; x<stR1.width; x+=i_iPixelStep) {
					if((*((uint32_t*)pBuf1Cpy) & iMask)!=0) {
						return true;
					}
					pBuf1Cpy += iPixelSize;
				}
				pBuf1 += iLineBytes1;
			}
		}
		return false;
	}

	////////////////////////////////////////////////////////////////////////////////
	//code based on http://www.stephencalenderblog.com/?p=293
	bool RectContainsPoint(S_Rect *i_pstR, int x, int y)
	{
		if((i_pstR->x <= x && i_pstR->y <= y)
			&& (x <= (i_pstR->x + i_pstR->width) && y <= (i_pstR->y + i_pstR->height)))
		{
			return true;
		}
		return false;
	}

	bool RectContainsRect(S_Rect *i_pstR1, S_Rect *i_pstR2)
	{
		if((i_pstR1->x <= i_pstR2->x && i_pstR1->y <= i_pstR2->y)
			&& ((i_pstR2->x + i_pstR2->width) <= (i_pstR1->x + i_pstR1->width))
			&& ((i_pstR2->y + i_pstR2->height) <= (i_pstR1->y + i_pstR1->height)))
		{
			return true;
		}
		return false;
	}

	bool IntersectRect2(S_Rect *o_pstROverlap, S_Rect *i_pstR1, S_Rect *i_pstR2)
	{
		if(RectContainsRect(i_pstR1, i_pstR2)) {
			*o_pstROverlap = *i_pstR2;
			return true;
		}
		if(RectContainsRect(i_pstR2, i_pstR1)) {
			*o_pstROverlap = *i_pstR1;
			return true;
		}

		if(RectContainsPoint(i_pstR2, i_pstR1->x,i_pstR1->y)) {
			o_pstROverlap->x = i_pstR1->x;
			o_pstROverlap->y = i_pstR1->y;
		}
		else if(RectContainsPoint(i_pstR1, i_pstR2->x, i_pstR2->y)) {
			o_pstROverlap->x = i_pstR2->x;
			o_pstROverlap->y = i_pstR2->y;
		}
		else if((i_pstR1->x <= i_pstR2->x && i_pstR2->x <= (i_pstR1->x + i_pstR1->width))
			&&(i_pstR2->y <= i_pstR1->y && i_pstR1->y <= (i_pstR2->y + i_pstR2->height)))
		{
			o_pstROverlap->x = i_pstR2->x;
			o_pstROverlap->y = i_pstR1->y;
		}
		else if((i_pstR1->y <= i_pstR2->y && i_pstR2->y <= (i_pstR1->y + i_pstR1->height))
			&&(i_pstR2->x <= i_pstR1->x && i_pstR1->x <= (i_pstR2->x + i_pstR2->width)))
		{
			o_pstROverlap->x = i_pstR1->x;
			o_pstROverlap->y = i_pstR2->y;
		}
		else {
			return false; //intersection does not exist
		}

		if(RectContainsPoint(i_pstR2, i_pstR1->x + i_pstR1->width, i_pstR1->y + i_pstR1->height)) {
			o_pstROverlap->width  = (i_pstR1->x + i_pstR1->width)  - o_pstROverlap->x;
			o_pstROverlap->height = (i_pstR1->y + i_pstR1->height) - o_pstROverlap->y;
		}
		else if(RectContainsPoint(i_pstR1, i_pstR2->x + i_pstR2->width, i_pstR2->y + i_pstR2->height)) {
			o_pstROverlap->width  = (i_pstR2->x + i_pstR2->width)  - o_pstROverlap->x;
			o_pstROverlap->height = (i_pstR2->y + i_pstR2->height) - o_pstROverlap->y;
		}
		else if((i_pstR1->x <= (i_pstR2->x + i_pstR2->width) && (i_pstR2->x + i_pstR2->width) <= (i_pstR1->x + i_pstR1->width))
			&&(i_pstR2->y <= (i_pstR1->y + i_pstR1->height) && (i_pstR1->y + i_pstR1->height) <= (i_pstR2->y + i_pstR2->height)))
		{
			o_pstROverlap->width  = (i_pstR2->x + i_pstR2->width)  - o_pstROverlap->x;
			o_pstROverlap->height = (i_pstR1->y + i_pstR1->height) - o_pstROverlap->y;
		}
		else {
			o_pstROverlap->width  = (i_pstR1->x + i_pstR1->width)  - o_pstROverlap->x;
			o_pstROverlap->height = (i_pstR2->y + i_pstR2->height) - o_pstROverlap->y;
		}
		if(o_pstROverlap->width<=0)  o_pstROverlap->width = 1;
		if(o_pstROverlap->height<=0) o_pstROverlap->height = 1;
		return true;
	}

}
