// C_LandingZone, C_CheckPoint, C_ComputerEnemy, C_Door

//classes for map objects other than ship players.
// - landing zone
//   - drawing
//   - handle cargo load unload (via server)
//   - handle ship landing
// - check point
//   - drawing
//   - check for ship player at CP
// - door element
//   - drawing
//   - opening/closing (via server)
// - computer enemy
//   - drawing
//   - movement (via server)

#include "mapobjects.h"
#include "math.h"
#include "PathUtils.h"
#include "image_utils.h"
#include "global.h"

#include "soundfx.h"

#define POINTINRECT(xx, yy, rect) (((xx) >= (rect).x) && ((xx) <= (rect).x+(rect).width) && ((yy) >= (rect).y) && ((yy) <= (rect).y+(rect).height))

//constants
static int LAND_X1 = 9;  //when shipsize is 48*48...
static int LAND_X2 = 39; //^
static int LAND_Y  = 42; //^
static POINT LAND_BPOS[] = {{13,44},{10,42},{34,44}};

static int NUM_ANTENNARECTS = 5;
static S_Rect s_stLandingRect    = {0,44,96,4};
static S_Rect s_stHouseRect      = {70,23,24,21};
static S_Rect s_stAntennaRects[] = {{0,15,4,29},{4,15,4,29},{8,15,4,29},{12,15,4,29},{16,15,4,29}};
static S_Rect s_stCargoRects[]   = {{54,38,14,6},{54,32,14,6},{54,26,14,6},{54,20,14,6}};
static S_Rect s_stExtraLifeRect  = {20,25,14,19};

C_Image *C_LandingZone::s_pclLandingZoneImg = NULL;
int C_LandingZone::s_iObjCount = 0;

C_LandingZone::C_LandingZone(C_Map *i_pclMap, int i_iX, int i_iY, int i_iWidth,
	bool i_bHomeBase, bool i_bShowAntenna, bool i_bShowHouse,
	std::vector<int> *i_pclCargoList, bool i_bExtraLife, C_NetMsgServer *i_pclNetMsgServer)
	: C_Object()
{
	m_dPosX = i_iX;
	m_dPosY = i_iY;
	m_dVelX = m_dVelY = m_dAcceleration = 0.0;
	m_dDirection = 270.0;
	m_iZoneSize = i_iWidth;

	m_pclMap = i_pclMap;

	if(!s_pclLandingZoneImg) {
		s_pclLandingZoneImg = new C_Image((char*)"zone.tga", C_Global::szResourceFile);
		C_GraphWrapperGL::GetGraphWrapperObject()->LoadTextureAbs(s_pclLandingZoneImg);
	}
	s_pclLandingZoneImg->GetInfo(&m_iWidth, &m_iHeight, NULL);
	s_iObjCount++;
	m_pclImage = s_pclLandingZoneImg;
	m_pstImageRect = NULL; //s_stLandingRect cannot be used since landingzone size is now variable

	m_bHomeBase     = i_bHomeBase;
	m_bShowAntenna  = i_bShowAntenna;
	m_bShowHouse    = i_bShowHouse;
	m_pclCargoList  = i_pclCargoList;
	m_dAntennaFrame = 0.0;
	m_bExtraLife    = i_bExtraLife;

	for(int i=0; i<8; i++) {
		m_bPlayerIsLanded[i] = false;
	}

	m_pclNetMsgServer = i_pclNetMsgServer;
}

C_LandingZone::~C_LandingZone()
{
	//last object alive?
	if(--s_iObjCount == 0) {
		delete s_pclLandingZoneImg;
		s_pclLandingZoneImg = NULL;
	}

	delete m_pclCargoList; m_pclCargoList = NULL;
}

int C_LandingZone::GetZoneSize()
{
	return m_iZoneSize;
}

int C_LandingZone::GetTotalCargo()
{
	int iCargo = 0;
	if(m_pclCargoList!=NULL) {
		int i, iSize = (int)m_pclCargoList->size();
		for(i=0; i<iSize; i++) {
			iCargo += (*m_pclCargoList)[i];
		}
	}
	return iCargo;
}

void C_LandingZone::UpdateFuel(C_ShipPlayer *i_pclPlayer, double i_dTime)
{
	//refuel if homebase
	if(!m_bHomeBase) return;
	S_Rect stRect = {(int)m_dPosX, (int)m_dPosY, 32*m_iZoneSize, s_stLandingRect.height+20};
	if(i_pclPlayer->m_iRealDir==27 && POINTINRECT(i_pclPlayer->m_dPosX+LAND_BPOS[1].x, i_pclPlayer->m_dPosY+LAND_BPOS[1].y, stRect))
	{
		i_pclPlayer->AddFuel((i_dTime*0.01)*1.0); //refuel 10 times faster than fuelburn
	}
}

void C_LandingZone::UpdateLanding(C_Player *i_pclPlayer, double i_dTime)
{
	m_dAntennaFrame += i_dTime*0.0046; //4.6 fps anim
	if(m_dAntennaFrame>=NUM_ANTENNARECTS) m_dAntennaFrame = 0.0; //is m_dAntennaFrame-='length' really, but then it can happen that it'll be more than 'length' when too much time have passed since last time (frame)...

	if(!i_pclPlayer) return;
	int iPlayerId = i_pclPlayer->GetOwnerID();
	if(!i_pclPlayer->IsAlive()) {
		m_bPlayerIsLanded[iPlayerId] = false;
		return;
	}

	int iDir = i_pclPlayer->m_iRealDir - 26;
	S_Rect stRect = {(int)m_dPosX, (int)m_dPosY, 32*m_iZoneSize, s_stLandingRect.height+20};
	//^check against rectangle that is thicker than the landingzone to detect a landing even if the ship is a little below the bottom of the zone.

	if((i_pclPlayer->m_iRealDir<=28) && (i_pclPlayer->m_iRealDir>=26)
		&& POINTINRECT(i_pclPlayer->m_dPosX+LAND_BPOS[iDir].x, i_pclPlayer->m_dPosY+LAND_BPOS[iDir].y, stRect))
	{
		//^ok with only rectangle collission detection for this object (no OCheckCollision)

		if(!m_bPlayerIsLanded[iPlayerId]) {
			//land, bounce or crashland?
			if((i_pclPlayer->m_dPosX+LAND_X1>m_dPosX) && (i_pclPlayer->m_dPosX+LAND_X2<m_dPosX+32*m_iZoneSize)) {
				if((abs((int)i_pclPlayer->m_dVelX)<32) && (i_pclPlayer->m_dVelY>=0) && (i_pclPlayer->m_dVelY<40)) { //land
					i_pclPlayer->LandOnZone(((int)m_dPosY)-LAND_Y);
					m_bPlayerIsLanded[iPlayerId] = true;
					m_dTimeLanded[iPlayerId] = 0.0;
				} else if((abs((int)i_pclPlayer->m_dVelX)<72) && i_pclPlayer->m_dVelY>=0 && i_pclPlayer->m_dVelY<102) { //bounce
					i_pclPlayer->BounceOnZone(m_dPosY-LAND_Y);
				} else i_pclPlayer->Kill(KILLEDBY_SELF, 0); //to high velocity, ship explodes
			} else i_pclPlayer->Kill(KILLEDBY_SELF, 0); //outside landingzone, ship explodes
		}
	} else {
		m_bPlayerIsLanded[iPlayerId] = false;
	}

	if(m_bPlayerIsLanded[iPlayerId]) {
		//do action depending on type of LZ and depending on
		//current cargo on ship, all depending on m_dTimeLanded
		m_dTimeLanded[iPlayerId] += i_dTime;

		//first: can only do actions so fast
		if(m_dTimeLanded[iPlayerId]>800) {
			m_dTimeLanded[iPlayerId] = 0.0;

			//first take any extra life
			if(m_bExtraLife) {
				//send life request
				S_NetMsg2 stMsg2;
				stMsg2.iType = N_MSG2_LIFE;
				stMsg2.iZoneId = m_iOwnerID;
				m_pclNetMsgServer->SendMsg2WithId(&stMsg2);
			} else { //then act on cargo
				if(!m_bHomeBase) { //pick up cargo
					if(m_pclCargoList!=NULL && m_pclCargoList->size()>0) { //there are cargo left on zone
						int iPos = (int)m_pclCargoList->size()-1;
						if(i_pclPlayer->SpaceInHold((*m_pclCargoList)[iPos])) { //player has space in hold
							//send cargo request
							S_NetMsg2 stMsg2;
							stMsg2.iType = N_MSG2_CARGO;
							stMsg2.iZoneId = m_iOwnerID;
							stMsg2.iCargoNum = iPos;
							m_pclNetMsgServer->SendMsg2WithId(&stMsg2);
						}
					}
				} else { //unload cargo
					i_pclPlayer->UnloadCargo();
				}
			}
		} //else no action this time

		//refuel if homebase
		if(m_bHomeBase) i_pclPlayer->AddFuel((i_dTime*0.01)*1.0); //refuel 10 times faster than fuelburn
	}
}

void C_LandingZone::TakeExtraLife()
{
	m_bExtraLife = false;
}

int C_LandingZone::PopCargo()
{
	int iCargo = -1;
	int iPos = (int)m_pclCargoList->size()-1;
	if(iPos>=0) {
		iCargo = (*m_pclCargoList)[iPos];
		m_pclCargoList->pop_back();
	}

	return iCargo;
}

void C_LandingZone::PushCargo(int i_iWeight)
{
	m_pclCargoList->push_back(i_iWeight);
}

void C_LandingZone::Draw()
{
	int i, x, y;
	m_stScreenArea = *(m_pclMap->GetScreenRect());
	m_pclMap->GetPos(&x, &y);
	S_Rect stRect = s_stLandingRect;
	stRect.width = 32;
	ODraw(m_pclImage, &stRect, x, y); //left part
	stRect.x += 32;
	for(i=0; i<m_iZoneSize-2; i++) {
		ODraw(m_pclImage, &stRect, x-32-i*32, y); //centre parts
	}
	stRect.x += 32;
	ODraw(m_pclImage, &stRect, x-(m_iZoneSize-1)*32, y); //right part
	int iAdjustX = -(m_iZoneSize-3)*32;

	if(m_bShowHouse) ODraw(m_pclImage, &s_stHouseRect, x-70+iAdjustX, y+21);
	if(m_bShowAntenna) ODraw(m_pclImage, &s_stAntennaRects[(int)m_dAntennaFrame], x, y+29);
	if(m_pclCargoList!=NULL) {
		int iSize = (int)m_pclCargoList->size();
		int iBaseX = m_bShowHouse ? -50+iAdjustX : -78+iAdjustX;
		for(i=0; i<iSize; i++) {
			int iCargoType = 3-((*m_pclCargoList)[i]-1)/5; //1..5=small container, 6..10, 11..15, 16..20+ = large container
			if(iCargoType<0) iCargoType=0;
			//if(iCargoType>3) iCargoType=3;
			ODraw(m_pclImage, &s_stCargoRects[iCargoType], x+iBaseX+((i/3)*14), y+6+((i%3)*6));
		}
	}
	if(m_bExtraLife) {
		int iBaseX = m_bShowAntenna ? -8 : -4;
		ODraw(m_pclImage, &s_stExtraLifeRect, x+iBaseX, y+s_stExtraLifeRect.height);
	}
}

////////////////////////////////////////////////////////////////////////////////
int C_CheckPoint::s_iObjCount = 0;
C_Font *C_CheckPoint::s_pclFont = NULL;
C_Image *C_CheckPoint::s_pclCpBkImg = NULL;

C_CheckPoint::C_CheckPoint(C_Map *i_pclMap, S_CheckPointInfo *i_pstCPInfo, int i_iCPNum)
	: C_Object()
{
	m_iPosX1 = i_pstCPInfo->iX1;
	m_iPosY1 = i_pstCPInfo->iY1;
	m_iPosX2 = i_pstCPInfo->iX2;
	m_iPosY2 = i_pstCPInfo->iY2;
	m_iCPNum = i_iCPNum;
	sprintf(m_szString, "%d", i_iCPNum);

	m_pclMap = i_pclMap;
	m_dBlinkTimer = 0;
	m_bBlinkState = false;

	if(!s_pclFont) {
		s_pclFont = new C_Font(_RGBA8888);
		s_pclFont->Load((char*)"arial14", C_Global::szResourceFile);
		S_FColor stColor = S_FCOLORMAKE(1.0, 1.0, 1.0, 1.0);
		s_pclFont->SetColor(stColor);

		s_pclCpBkImg = new C_Image((char*)"cpbk.tga", C_Global::szResourceFile);
	}

	s_iObjCount++;
	m_pclImage = s_pclCpBkImg;
}

C_CheckPoint::~C_CheckPoint()
{
	//last object alive?
	if(--s_iObjCount == 0) {
		delete s_pclFont; s_pclFont = NULL;
		delete s_pclCpBkImg; s_pclCpBkImg = NULL;
	}
}

void C_CheckPoint::GetCheckPointInfo(S_CheckPointInfo *o_pstCP)
{
	int iSizeX, iSizeY;
	GetSize(&iSizeX, &iSizeY);
	iSizeX/=2; iSizeY/=2;

	o_pstCP->iX1 = m_iPosX1+iSizeX;
	o_pstCP->iY1 = m_iPosY1+iSizeY;
	o_pstCP->iX2 = m_iPosX2+iSizeX;
	o_pstCP->iY2 = m_iPosY2+iSizeY;
}

//true   player is in range of checkpoint
//false  player is not in range of checkpoint
#define CP_RANGE 11
bool C_CheckPoint::PlayerAtCP(C_ShipPlayer *i_pclPlayer)
{
	int iX, iY;
	int iSizeX, iSizeY;
	i_pclPlayer->GetCenterPos(&iX, &iY);
	GetSize(&iSizeX, &iSizeY);
	iSizeX/=2; iSizeY/=2;

	//calculate shortest distance from player to checkpoint line
	VECTOR p = PathUtils::S_VECTORMAKE2((double)iX, (double)iY);
	VECTOR l0 = PathUtils::S_VECTORMAKE2((double)m_iPosX1+iSizeX, (double)m_iPosY1+iSizeY);
	VECTOR l1 = PathUtils::S_VECTORMAKE2((double)m_iPosX2+iSizeX, (double)m_iPosY2+iSizeY);
	double dDist = PathUtils::PointDistanceToLineSeg(p, l0, l1);

	if(dDist < CP_RANGE) {
		return true;
	}
	return false;
}

void C_CheckPoint::SetBlinkState(bool i_bBlinkState)
{
	if(i_bBlinkState && !m_bBlinkState) m_dBlinkTimer = 0;
	m_bBlinkState = i_bBlinkState;
}

void C_CheckPoint::Update(double i_dTime)
{
	m_dBlinkTimer += i_dTime;
}

void C_CheckPoint::Draw()
{
	int iX, iY, iW;
	m_stScreenArea = *(m_pclMap->GetScreenRect());
	S_Rect *pstRect = m_pclMap->GetScreenRect();
	m_pclMap->GetPos(&iX, &iY);
	s_pclCpBkImg->GetInfo(&iW, NULL, NULL);

	m_dPosX = m_iPosX1;
	m_dPosY = m_iPosY1;
	ODraw(s_pclCpBkImg, NULL, iX, iY);
	m_dPosX = m_iPosX2;
	m_dPosY = m_iPosY2;
	ODraw(s_pclCpBkImg, NULL, iX, iY);

	if(!m_bBlinkState || (m_bBlinkState && ((int)m_dBlinkTimer % 450 < 300))) {
		int iWidth;
		s_pclFont->GetTextSize(m_szString, &iWidth);
		int iPosX = pstRect->x+ (m_iPosX1 - iX);
		int iPosY = pstRect->y+ (m_iPosY1 - iY);
		s_pclFont->Print(iPosX+((iW-iWidth)/2),iPosY+1, m_szString);
		iPosX = pstRect->x+ (m_iPosX2 - iX);
		iPosY = pstRect->y+ (m_iPosY2 - iY);
		s_pclFont->Print(iPosX+((iW-iWidth)/2),iPosY+1, m_szString);
	}
}

////////////////////////////////////////////////////////////////////////////////
static S_Rect s_stVBaseRect[] = {{0,0,52,36},{0,36,52,36}};
static S_Rect s_stVDoorRect[] = {{23,0,6,6},{23,6,6,24},{23,30,6,6}};
static S_Rect s_stHBaseRect[] = {{52,20,36,52},{88,20,36,52}};
static S_Rect s_stHDoorRect[] = {{52,43,6,6},{58,43,24,6},{82,43,6,6}};
static S_Rect s_stButtonRect[] = {{52,0,23,19},{75,0,23,19}};

C_Image *C_Door::s_pclDoorImg = NULL;
int C_Door::s_iObjCount = 0;

C_Door::C_Door(C_Map *i_pclMap, int i_iX, int i_iY, S_DoorInfo *i_pstInfo, int i_iID, bool i_bMaster, C_NetMsg *i_pclNetMsg)
	: C_Object()
{
	m_iOwnerID = i_iID;
	m_dPosX = i_iX;
	m_dPosY = i_iY;
	m_dVelX = m_dVelY = m_dAcceleration = 0.0;
	m_bMaster = i_bMaster;
	m_stInfo = *i_pstInfo;
	m_dDirection = m_stInfo.bHorizontal ? 270.0 : 0.0;

	m_pclMap = i_pclMap;

	if(!s_pclDoorImg) {
		s_pclDoorImg = new C_Image((char*)"door.tga", C_Global::szResourceFile);
		C_GraphWrapperGL::GetGraphWrapperObject()->LoadTextureAbs(s_pclDoorImg);
	}
	s_iObjCount++;
	m_pclImage = s_pclDoorImg;
	m_pstImageRect = NULL;

	m_bIsOpening = false;
	m_dOpenPos = 0.0;
	m_dStateTime = 0.0;

	m_pclNetMsg = i_pclNetMsg;
}

C_Door::~C_Door()
{
	//last object alive?
	if(--s_iObjCount == 0) {
		delete s_pclDoorImg;
		s_pclDoorImg = NULL;
	}
}

bool C_Door::CheckCollission(C_Object *i_pclObject, bool i_bIsBullet)
{
	int x, y, i;
	m_stScreenArea = *(m_pclMap->GetScreenRect());
	m_pclMap->GetPos(&x, &y);
	bool bResult = true;

	double xx = m_dPosX;
	double yy = m_dPosY;
	if(m_dDirection==270.0) {
		m_pstImageRect = &s_stVBaseRect[0];
		if(OCheckCollision(i_pclObject)) goto out;
		m_dPosX = xx; m_dPosY = yy+m_stInfo.iSize+36;
		m_pstImageRect = &s_stVBaseRect[1];
		if(OCheckCollision(i_pclObject)) goto out;

		int iSize = (m_stInfo.iSize-(int)m_dOpenPos)/2;
		S_Rect stRect = {0,0,6,iSize};

		m_dPosX = xx+23; m_dPosY = yy+36;
		if(ImageUtils::Compare(NULL, (int)m_dPosX, (int)m_dPosY, i_pclObject->m_pclImage,
			(int)i_pclObject->m_dPosX, (int)i_pclObject->m_dPosY, &stRect, i_pclObject->m_pstImageRect)) goto out;
		m_dPosX = xx+23; m_dPosY = yy+m_stInfo.iSize+36-iSize;
		if(ImageUtils::Compare(NULL, (int)m_dPosX, (int)m_dPosY, i_pclObject->m_pclImage,
			(int)i_pclObject->m_dPosX, (int)i_pclObject->m_dPosY, &stRect, i_pclObject->m_pstImageRect)) goto out;
	} else {
		m_pstImageRect = &s_stHBaseRect[0];
		if(OCheckCollision(i_pclObject)) goto out;
		m_dPosX = xx+m_stInfo.iSize+36; m_dPosY = yy;
		m_pstImageRect = &s_stHBaseRect[1];
		if(OCheckCollision(i_pclObject)) goto out;

		int iSize = (m_stInfo.iSize-(int)m_dOpenPos)/2;
		S_Rect stRect = {0,0,iSize,6};

		m_dPosX = xx+36; m_dPosY = yy+23;
		if(ImageUtils::Compare(NULL, (int)m_dPosX, (int)m_dPosY, i_pclObject->m_pclImage,
			(int)i_pclObject->m_dPosX, (int)i_pclObject->m_dPosY, &stRect, i_pclObject->m_pstImageRect)) goto out;
		m_dPosX = xx+m_stInfo.iSize+36-iSize; m_dPosY = yy+23;
		if(ImageUtils::Compare(NULL, (int)m_dPosX, (int)m_dPosY, i_pclObject->m_pclImage,
			(int)i_pclObject->m_dPosX, (int)i_pclObject->m_dPosY, &stRect, i_pclObject->m_pstImageRect)) goto out;
	}
	m_pstImageRect = &s_stButtonRect[0];
	for(i=0; i<m_stInfo.iNumButtons; i++) {
		m_dPosX = m_stInfo.stButtonPos[i].x;
		m_dPosY = m_stInfo.stButtonPos[i].y;
		if(OCheckCollision(i_pclObject)) {
			if(i_bIsBullet && m_bMaster) {
				m_dStateTime = 0;
				m_bIsOpening = !m_bIsOpening; //toggle
				SendDoorAction(m_bIsOpening);
			}
			goto out;
		}
	}
	bResult = false;

out:
	m_dPosX = xx;
	m_dPosY = yy;
	return bResult;
}

void C_Door::Update(double i_dTime)
{
	bool bSendDoorAction = false;
	if(!m_bMaster) {
		S_NetMsg1 stMsg1;
		while(m_pclNetMsg->GetMsg1(m_iOwnerID, &stMsg1)) { //try to recieve info from network
			if(stMsg1.iMsg & N_MSG1_KEYCHANGE) {
				m_dStateTime = 0;
				m_bIsOpening = stMsg1.iDoorAction!=0;
			}
		}
	}

	if(m_bIsOpening) {
		m_dOpenPos += i_dTime*(m_stInfo.iOpeningSpeed*0.001);
		if(m_dOpenPos>=m_stInfo.iSize) {
			m_dOpenPos = m_stInfo.iSize;
			m_dStateTime += i_dTime;
			if(m_bMaster && m_dStateTime>=m_stInfo.iOpenedForTime) {
				m_dStateTime = 0;
				m_bIsOpening = false;
				bSendDoorAction = true; //maybe remove if it shall not be possible to toggle open/close
			}
		}
	} else {
		m_dOpenPos -= i_dTime*(m_stInfo.iClosingSpeed*0.001);
		if(m_dOpenPos<=0) {
			m_dOpenPos = 0;
			m_dStateTime += i_dTime;
			if(m_bMaster && m_dStateTime>=m_stInfo.iClosedForTime) {
				if(m_stInfo.iNumButtons==0) {
					m_dStateTime = 0;
					m_bIsOpening = true; //interval opening
					bSendDoorAction = true;
				}
			}
		}
	}

	if(bSendDoorAction) {
		SendDoorAction(m_bIsOpening);
	}
}

void C_Door::SendDoorAction(bool i_bOpen)
{
	S_NetMsg1 stMsg1;
	stMsg1.iMsg = N_MSG1_KEYCHANGE; //use this message (door has no other messages than the door action)
	stMsg1.iSenderId = (int8_t)m_iOwnerID;
	stMsg1.iDoorAction = i_bOpen ? 1 : 0;
	m_pclNetMsg->SendMsg1(&stMsg1);
}

void C_Door::Draw()
{
	int x, y, i;
	m_stScreenArea = *(m_pclMap->GetScreenRect());
	m_pclMap->GetPos(&x, &y);

	double xx = m_dPosX;
	double yy = m_dPosY;
	S_Rect stRect;
	if(m_dDirection==270.0) {
		ODraw(m_pclImage, &s_stVBaseRect[0], x, y);
		m_dPosX = xx; m_dPosY = yy+m_stInfo.iSize+36;
		ODraw(m_pclImage, &s_stVBaseRect[1], x, y);

		stRect = s_stVDoorRect[1];
		int iSize = (m_stInfo.iSize-(int)m_dOpenPos)/2;
		int iPos = -6;
		while(iPos<iSize-6) {
			if(iPos+24>=iSize-6) stRect.height=(iSize-6)-iPos;
			m_dPosX = xx+23; m_dPosY = yy+36+iPos;
			ODraw(m_pclImage, &stRect, x, y);
			m_dPosX = xx+23; m_dPosY = yy+m_stInfo.iSize+36-iPos-stRect.height;
			ODraw(m_pclImage, &stRect, x, y);
			iPos+=stRect.height;
		}

		m_dPosX = xx+23; m_dPosY = yy+36+iPos;
		ODraw(m_pclImage, &s_stVDoorRect[2], x, y);
		m_dPosX = xx+23; m_dPosY = yy+m_stInfo.iSize+36-iPos-6;
		ODraw(m_pclImage, &s_stVDoorRect[0], x, y);
	} else {
		ODraw(m_pclImage, &s_stHBaseRect[0], x, y);
		m_dPosX = xx+m_stInfo.iSize+36; m_dPosY = yy;
		ODraw(m_pclImage, &s_stHBaseRect[1], x, y);

		stRect = s_stHDoorRect[1];
		int iSize = (m_stInfo.iSize-(int)m_dOpenPos)/2;
		int iPos = -6;
		while(iPos<iSize-6) {
			if(iPos+24>=iSize-6) stRect.width=(iSize-6)-iPos;
			m_dPosX = xx+36+iPos; m_dPosY = yy+23;
			ODraw(m_pclImage, &stRect, x, y);
			m_dPosX = xx+m_stInfo.iSize+36-iPos-stRect.width; m_dPosY = yy+23;
			ODraw(m_pclImage, &stRect, x, y);
			iPos+=stRect.width;
		}

		m_dPosX = xx+36+iPos; m_dPosY = yy+23;
		ODraw(m_pclImage, &s_stHDoorRect[2], x, y);
		m_dPosX = xx+m_stInfo.iSize+36-iPos-6; m_dPosY = yy+23;
		ODraw(m_pclImage, &s_stHDoorRect[0], x, y);
	}
	for(i=0; i<m_stInfo.iNumButtons; i++) {
		m_dPosX = m_stInfo.stButtonPos[i].x;
		m_dPosY = m_stInfo.stButtonPos[i].y;
		ODraw(m_pclImage, &s_stButtonRect[m_bIsOpening ? 1 : 0], x, y);
	}

	m_dPosX = xx;
	m_dPosY = yy;
}

////////////////////////////////////////////////////////////////////////////////
#define SENEMY_FIRERANGE 450

#define SENEMY_FI_TABLESIZE 30
int SENEMY_FIREINTERVALTABLE[SENEMY_FI_TABLESIZE] = {
	300,  300,  400,  500,  500,  700,  850,  900,  1000, 1100,
	1200, 1500, 1700, 1800, 2000, 2100, 2200, 2300, 2450, 2600,
	2800, 3000, 3500, 3800, 4000, 5000, 6000, 7000, 8000, 8000};

//enemy properties
int  SENEMY_HITSTOKILL[]        = {4, 5, -1, -1, 8, 4, 4, 6}; //-1=immortal
int  SENEMY_SCOREPOINTS[]       = {10, 20, 0, 0, 20, 15, 25, 25};
int  SENEMY_NUMBULLETS[]        = {1, 3, 1, 3, 1, 2, 5, 8};
int  SENEMY_BULLETSPEED[]       = {190, 160, 190, 160, 110, 90, 120, 140};
bool SENEMY_RANDOMBULLETANGLE[] = {false, false, false, false, true, false, false, false};
int  SENEMY_BULLETANGLE[][8]    = {
	{0,0,0,0,0,0,0,0}, {-50,0,50,0,0,0,0,0}, {0,0,0,0,0,0,0,0}, {-50,0,50,0,0,0,0,0},
	{0,0,0,0,0,0,0,0}, {0,0,0,0,0,0,0,0}, {-70,-35,0,35,70,0,0,0}, {-180,-135,-90,-45,0,45,90,135}};
int  SENEMY_NUMFRAMES[]         = {1, 1, 1, 1, 12, 1, 1, 12}; //valid for type 4 and 7 only

//unanimated enemies (rectangles are 4 directions (0, 90, 180, 270))
S_Rect SENEMY6_RECT[] = {{0,14,14,14},{14,0,14,14},{14,14,14,14},{0,0,14,14}};
POINT  SENEMY6_FIREPOINT[] = {{13,6},{6,13},{0,6},{6,0}};
//animated enemies, 1 direction (rectangles are each frame)
S_Rect SENEMY4_RECT[] = {
	{0,0,64,32},{64,0,64,32},{128,0,64,32},{192,0,64,32},
	{0,32,64,32},{64,32,64,32},{128,32,64,32},{192,32,64,32},
	{0,64,64,32},{64,64,64,32},{128,64,64,32},{192,64,64,32}};
POINT  SENEMY4_FIREPOINT = {32,16};
S_Rect SENEMY7_RECT[] = {
	{0,0,64,64},{64,0,64,64},{128,0,64,64},{192,0,64,64},
	{0,64,64,64},{64,64,64,64},{128,64,64,64},{192,64,64,64},
	{0,128,64,64},{64,128,64,64},{128,128,64,64},{192,128,64,64}};
POINT  SENEMY7_FIREPOINT = {32,32};
//enemies with animated fireing, 4 directions
S_Rect SENEMY0_RECT[] = {
	{0,16,16,12},{24+0,16,16,12}, {12,0,12,16},{24+12,0,12,16},
	{0,28,16,12},{24+0,28,16,12}, {0,0,12,16},{24+0,0,12,16}};
POINT  SENEMY0_FIREPOINT[] = { {16,5}, {5,16}, {0,5}, {5,0}};
S_Rect SENEMY2_RECT[] = {
	{0,16,16,12},{24+0,16,16,12}, {12,0,12,16},{24+12,0,12,16},
	{0,28,16,12},{24+0,28,16,12}, {0,0,12,16},{24+0,0,12,16}};
POINT  SENEMY2_FIREPOINT[] = { {16,5}, {5,16}, {0,5}, {5,0}};
S_Rect SENEMY1_RECT[] = {
	{14,14,14,24},{28+14,14,14,24}, {0,38,24,14},{28+0,38,24,14},
	{0,14,14,24},{28+0,14,14,24}, {0,0,24,14},{28+0,0,24,14}};
POINT  SENEMY1_FIREPOINT[] = {{14,12},{12,14},{0,12},{12,0}};
S_Rect SENEMY3_RECT[] = {
	{14,14,14,24},{28+14,14,14,24}, {0,38,24,14},{28+0,38,24,14},
	{0,14,14,24},{28+0,14,14,24}, {0,0,24,14},{28+0,0,24,14}};
POINT  SENEMY3_FIREPOINT[] = {{14,12},{12,14},{0,12},{12,0}};
S_Rect SENEMY5_RECT[] = { 
	{0,26,13,22},{0,48,13,22}, {0,13,22,13},{22,13,22,13},
	{13,26,13,22},{13,48,13,22}, {0,0,22,13},{22,0,22,13}};
POINT  SENEMY5_FIREPOINT[] = { {13,4},{13,18}, {4,13},{18,13}, {0,4},{0,18}, {4,0},{18,0}};

C_ObjPPList *C_ComputerEnemy::s_pclRangeList = NULL;

C_Image *C_ComputerEnemy::s_pclEnemy0Img = NULL;
C_Image *C_ComputerEnemy::s_pclEnemy1Img = NULL;
C_Image *C_ComputerEnemy::s_pclEnemy2Img = NULL;
C_Image *C_ComputerEnemy::s_pclEnemy3Img = NULL;
C_Image *C_ComputerEnemy::s_pclEnemy4Img = NULL;
C_Image *C_ComputerEnemy::s_pclEnemy5Img = NULL;
C_Image *C_ComputerEnemy::s_pclEnemy6Img = NULL;
C_Image *C_ComputerEnemy::s_pclEnemy7Img = NULL;
int C_ComputerEnemy::s_iObjCount = 0;

C_ComputerEnemy::C_ComputerEnemy(C_Map *i_pclMap, C_BulletList *i_pclBulletList, C_NetMsg *i_pclNetMsg, bool i_bMaster, S_EnemyInfo *i_pstEnemyInfo, int i_iID) : C_PlayerBase()
{
	m_iOwnerID = i_iID;
	m_bMaster = i_bMaster;
	m_pclMap = i_pclMap;
	m_pclBulletList = i_pclBulletList;
	m_pclNetMsg = i_pclNetMsg;

	m_iEnemyType    = i_pstEnemyInfo->iEnemyType;
	m_iFireInterval = i_pstEnemyInfo->iFireInterval;
	m_dDirection    = i_pstEnemyInfo->iAngle;
	m_iSpeed        = i_pstEnemyInfo->iSpeed;
	m_iNumWayPoints = i_pstEnemyInfo->iNumWayPoints;
	memcpy(&m_pstWayPointList[0], &i_pstEnemyInfo->astWayPoint[0], m_iNumWayPoints*2*sizeof(int));
	m_dPosX = m_pstWayPointList[0].x;
	m_dPosY = m_pstWayPointList[0].y;
	m_iCurWP = 1; //after 0
	m_dWayTime = 0.0;
	m_dVelX = m_dVelY = m_dAcceleration = 0.0;
	m_dFrameNr = 0.0;
	m_dFireAnimTime = 0.0;
	m_stMsg1.iMsg = 0; //no message is to be sent yet (if master)

	if(m_iFireInterval == -1) {
		m_dFireTime = SENEMY_FIREINTERVALTABLE[rand() % SENEMY_FI_TABLESIZE];
	} else m_dFireTime = m_iFireInterval;

	m_bFinished = false;
	m_iPathTimer = 0;
 
	if(!s_pclRangeList) s_pclRangeList = new C_ObjPPList();

	switch(m_iEnemyType) {
		case 0: 
			if(!s_pclEnemy0Img) s_pclEnemy0Img = new C_Image((char*)"enemy0.tga", C_Global::szResourceFile);
			m_pclImage = s_pclEnemy0Img;
			m_pstAnimRects = &SENEMY0_RECT[0];
			m_pstImageRect = &SENEMY0_RECT[i_pstEnemyInfo->iAngle/90];
			m_pstFirePoint = &SENEMY0_FIREPOINT[i_pstEnemyInfo->iAngle/90];
			break;
		case 1:
			if(!s_pclEnemy1Img) s_pclEnemy1Img = new C_Image((char*)"enemy1.tga", C_Global::szResourceFile);
			m_pclImage = s_pclEnemy1Img;
			m_pstAnimRects = &SENEMY1_RECT[0];
			m_pstImageRect = &SENEMY1_RECT[i_pstEnemyInfo->iAngle/90];
			m_pstFirePoint = &SENEMY1_FIREPOINT[i_pstEnemyInfo->iAngle/90];
			break;
		case 2:
			if(!s_pclEnemy2Img) s_pclEnemy2Img = new C_Image((char*)"enemy2.tga", C_Global::szResourceFile);
			m_pclImage = s_pclEnemy2Img;
			m_pstAnimRects = &SENEMY2_RECT[0];
			m_pstImageRect = &SENEMY2_RECT[i_pstEnemyInfo->iAngle/90];
			m_pstFirePoint = &SENEMY2_FIREPOINT[i_pstEnemyInfo->iAngle/90];
			break;
		case 3:
			if(!s_pclEnemy3Img) s_pclEnemy3Img = new C_Image((char*)"enemy3.tga", C_Global::szResourceFile);
			m_pclImage = s_pclEnemy3Img;
			m_pstAnimRects = &SENEMY3_RECT[0];
			m_pstImageRect = &SENEMY3_RECT[i_pstEnemyInfo->iAngle/90];
			m_pstFirePoint = &SENEMY3_FIREPOINT[i_pstEnemyInfo->iAngle/90];
			break;
		case 4:
			if(!s_pclEnemy4Img) s_pclEnemy4Img = new C_Image((char*)"enemy4.tga", C_Global::szResourceFile);
			m_pclImage = s_pclEnemy4Img;
			m_pstAnimRects = &SENEMY4_RECT[0];
			m_pstImageRect = &SENEMY4_RECT[0];
			m_pstFirePoint = &SENEMY4_FIREPOINT;
			break;
		case 5:
			if(!s_pclEnemy5Img) s_pclEnemy5Img = new C_Image((char*)"enemy5.tga", C_Global::szResourceFile);
			m_pclImage = s_pclEnemy5Img;
			m_pstAnimRects = &SENEMY5_RECT[0];
			m_pstImageRect = &SENEMY5_RECT[0];
			//m_pstFirePoint = &SENEMY5_FIREPOINT;
			break;
		case 6:
			if(!s_pclEnemy6Img) s_pclEnemy6Img = new C_Image((char*)"enemy6.tga", C_Global::szResourceFile);
			m_pclImage = s_pclEnemy6Img;
			m_pstImageRect = &SENEMY6_RECT[i_pstEnemyInfo->iAngle/90];
			m_pstFirePoint = &SENEMY6_FIREPOINT[i_pstEnemyInfo->iAngle/90];
			break;
		case 7:
			if(!s_pclEnemy7Img) s_pclEnemy7Img = new C_Image((char*)"enemy7.tga", C_Global::szResourceFile);
			m_pclImage = s_pclEnemy7Img;
			m_pstAnimRects = &SENEMY7_RECT[0];
			m_pstImageRect = &SENEMY7_RECT[0];
			m_pstFirePoint = &SENEMY7_FIREPOINT;
			break;
	}
	m_iHitsLeft = SENEMY_HITSTOKILL[m_iEnemyType];

	s_iObjCount++;
}

C_ComputerEnemy::~C_ComputerEnemy()
{
	//last object alive?
	if(--s_iObjCount == 0) {
		delete s_pclEnemy0Img; s_pclEnemy0Img = NULL;
		delete s_pclEnemy1Img; s_pclEnemy1Img = NULL;
		delete s_pclEnemy2Img; s_pclEnemy2Img = NULL;
		delete s_pclEnemy3Img; s_pclEnemy3Img = NULL;
		delete s_pclEnemy4Img; s_pclEnemy4Img = NULL;
		delete s_pclEnemy5Img; s_pclEnemy5Img = NULL;
		delete s_pclEnemy6Img; s_pclEnemy6Img = NULL;
		delete s_pclEnemy7Img; s_pclEnemy7Img = NULL;

		delete s_pclRangeList; s_pclRangeList = NULL;
	}
}

void C_ComputerEnemy::AddFireRangeObject(C_Object **i_ppclFireRangeObj)
{
	if(s_pclRangeList) s_pclRangeList->push_back(i_ppclFireRangeObj);
}

bool C_ComputerEnemy::InFireRange()
{
	if(!s_pclRangeList) return false;

	C_Object **ppclFireRangeObj;
	bool bResult = false;
	int i, iSize = (int)s_pclRangeList->size();
	double dDist;

	for(i=0; i<iSize; i++) {
		ppclFireRangeObj = (*s_pclRangeList)[i];
		if(*ppclFireRangeObj) {
			dDist = GetDistanceTo(*ppclFireRangeObj);

			if(dDist < SENEMY_FIRERANGE) {
				bResult = true;
				break;
			}
		}
	}
	return bResult;
}

int C_ComputerEnemy::GetScoreValue()
{
	return SENEMY_SCOREPOINTS[m_iEnemyType];
}

C_Bullet *C_ComputerEnemy::CreateBullet(int i_iIndex)
{
	S_BulletInfo stBulletInfo;
	double dAngle = m_dDirection+SENEMY_BULLETANGLE[m_iEnemyType][i_iIndex];
	double dTemp = dAngle*(PI/180);
	double dSin = sin(dTemp);
	double dCos = cos(dTemp);
	stBulletInfo.dPosX = m_dPosX + m_pstFirePoint->x;
	stBulletInfo.dPosY = m_dPosY + m_pstFirePoint->y;
	stBulletInfo.dVelX = dCos * SENEMY_BULLETSPEED[m_iEnemyType];
	stBulletInfo.dVelY = dSin * SENEMY_BULLETSPEED[m_iEnemyType];
	if(m_iEnemyType==7) {
		//this is the correct thing to do, but it looks better
		// for most enemies to skip this
		stBulletInfo.dVelX += m_dVelX;
		stBulletInfo.dVelY += m_dVelY;
	}
	stBulletInfo.dDirection = dAngle;

	return new C_Bullet(m_pclMap, &stBulletInfo, m_iOwnerID);
}

void C_ComputerEnemy::Draw()
{
	int iX, iY;
	m_pclMap->GetPos(&iX, &iY);
	m_stScreenArea = *(m_pclMap->GetScreenRect());

	if(m_bAlive) ODraw(m_pclImage, m_pstImageRect, iX, iY);
	else DrawExplosion(iX, iY);
}

void C_ComputerEnemy::SetMaster(bool i_bMaster)
{
	m_bMaster = i_bMaster;

	if(m_iNumWayPoints>1) {
		//figure out the current waypoint
		POINTD stPos = {m_dPosX, m_dPosY};
		for(int i=0; i<m_iNumWayPoints; i++) {
			int iNextWP = (m_iCurWP+1) % m_iNumWayPoints;
			VECTOR stP1 = {m_pstWayPointList[m_iCurWP].x, m_pstWayPointList[m_iCurWP].y};
			VECTOR stP2 = {m_pstWayPointList[iNextWP].x, m_pstWayPointList[iNextWP].y};
			m_iCurWP = iNextWP;
			double dDist = PathUtils::PointDistanceToLineSeg(stPos, stP1, stP2, NULL);
			if(dDist<10) break;
			//this may not be 100% but it's better than going through walls to reach checkpoint 0
		}
	}
}

void C_ComputerEnemy::Move(double i_dTime)
{
	if(m_bMaster && m_iNumWayPoints>1) { //move enemy
		m_dWayTime -= i_dTime;
		if(m_dWayTime <= 0) {
			double dDistX, dDistY, dDist;
			bool bSendMovement;

			dDistX = m_pstWayPointList[m_iCurWP].x - m_dPosX;
			dDistY = m_pstWayPointList[m_iCurWP].y - m_dPosY;
			bSendMovement = true;

			dDist = sqrt(dDistX*dDistX + dDistY*dDistY);
			m_dWayTime = dDist / m_iSpeed;
			if(m_dWayTime == 0.0) m_dWayTime = 0.8; //stand still for 0.8s if no destination!

			m_dVelX = dDistX / m_dWayTime;
			m_dVelY = dDistY / m_dWayTime;
			m_dWayTime *= 1000; //make it milliseconds
			m_iCurWP = (m_iCurWP+1) % m_iNumWayPoints;
			if(bSendMovement) m_stMsg1.iMsg |= N_MSG1_MOVEMENT;
		}
		OUpdate(i_dTime);
		SetCommonInfo(&m_stMsg1);
	} else if(m_iNumWayPoints > 1) {
		OUpdate(i_dTime);
	}
}

int C_ComputerEnemy::Update(double i_dTime)
{
	int i, iUId, iPlaySound = -1, iResult = -3;
	C_Bullet *pclBullet;
	bool bFireBullet = false;

	if(!m_bAlive) m_bFinished = UpdateExplosion(i_dTime);
	if(m_bFinished) return -2;

	if(m_bMaster && m_bAlive) {
		m_dFireTime -= i_dTime;
		int iBulletUID;
		if(m_dFireTime <= 0) { //fire bullet, send to network
			if(m_iFireInterval == -1) {
				m_dFireTime = SENEMY_FIREINTERVALTABLE[rand() % SENEMY_FI_TABLESIZE];
			} else m_dFireTime = m_iFireInterval;

			if(InFireRange()) {
				iPlaySound = 0; //bullet sound

				if(SENEMY_RANDOMBULLETANGLE[m_iEnemyType]) {
					m_dDirection = rand() % 360;
					m_stMsg1.iMsg |= N_MSG1_MOVEMENT;
					SetCommonInfo(&m_stMsg1);
				}
				m_stMsg1.iMsg |= N_MSG1_NEWBULLET;
				m_stMsg1.iNumBullets = (uint8_t)SENEMY_NUMBULLETS[m_iEnemyType]; //not used

				for(i=0; i<SENEMY_NUMBULLETS[m_iEnemyType]; i++) {
					if(m_iEnemyType==5) m_pstFirePoint = &SENEMY5_FIREPOINT[2*(int)(m_dDirection/90)+i];
					pclBullet = CreateBullet(i);
					s_iBulletCounter++;
					iUId = m_iOwnerID | (s_iBulletCounter << 8);
					pclBullet->SetUId(iUId);
					if(i==0) m_stMsg1.iBulletUniqeValBase = iUId; //additional bullets by this enemy will be calculated by receiver
					m_pclBulletList->Add(pclBullet);
				}
				bFireBullet = true;
			}
		}

		if(m_pclBulletList->CheckCollision(this, &iBulletUID, true)) {
			if(m_iHitsLeft!=-1) m_iHitsLeft--;
			iPlaySound = 4; //hit sound
			if(m_iHitsLeft <= 0 && m_iHitsLeft!=-1) {
				m_bAlive = false;
				m_bCanCollide = false;
				iPlaySound = 1; //explode sound
				iResult = iBulletUID & 255;
			}
			m_stMsg1.iMsg |= N_MSG1_KILLED;
			m_stMsg1.iCauseID = iBulletUID;
			m_stMsg1.iHitsLeft = (uint8_t)m_iHitsLeft;
		}
		Move(i_dTime);

		if(m_stMsg1.iMsg) { //send data
			m_stMsg1.iSenderId = (int8_t)m_iOwnerID;
			m_pclNetMsg->SendMsg1(&m_stMsg1);
			m_stMsg1.iMsg = 0;
		}
	} else if(!m_bMaster) {
		while(m_pclNetMsg->GetMsg1(m_iOwnerID, &m_stMsg1)) { //try to recieve info from network
			//movement for moving enemies
			if(m_stMsg1.iMsg & N_MSG1_MOVEMENT) GetCommonInfo(&m_stMsg1);

			//bullet(s) fired
			if(m_stMsg1.iMsg & N_MSG1_NEWBULLET) {
				iPlaySound = 0; //bullet sound
				int iBulletBase = m_stMsg1.iBulletUniqeValBase >> 8;
				for(i=0; i<SENEMY_NUMBULLETS[m_iEnemyType]; i++) {
					if(m_iEnemyType==5) m_pstFirePoint = &SENEMY5_FIREPOINT[2*(int)(m_dDirection/90)+i];
					pclBullet = CreateBullet(i);
					int iUId = (m_stMsg1.iBulletUniqeValBase & 0xff) | ((iBulletBase+i) << 8);
					pclBullet->SetUId(iUId);
					m_pclBulletList->Add(pclBullet);
				}
				bFireBullet = true;
			}

			//is hit
			if(m_stMsg1.iMsg & N_MSG1_KILLED) {
				m_pclBulletList->Remove(m_stMsg1.iCauseID);
				iPlaySound = 4; //hit sound
				m_iHitsLeft = m_stMsg1.iHitsLeft;

				if(m_iHitsLeft <= 0 && m_iHitsLeft!=-1) {
					m_bAlive = false;
					m_dVelX = m_dVelY = 0; //make explosion be still
					m_bCanCollide = false;
					iPlaySound = 1; //explode sound
					iResult = m_stMsg1.iCauseID & 255;
				}
			}
		}
		Move(i_dTime);
	}

	//animated?
	if(m_iEnemyType==4 || m_iEnemyType==7) {
		m_dFrameNr += i_dTime*0.021; //21 fps anim
		int iFrameNr = (int)m_dFrameNr;
		if(iFrameNr >= SENEMY_NUMFRAMES[m_iEnemyType]) {
			m_dFrameNr = 0.0; //restart anim
			iFrameNr = 0;
		}
		m_pstImageRect = &m_pstAnimRects[iFrameNr];
	}

	//animated fireing
	if(m_iEnemyType==5 || m_iEnemyType==0 || m_iEnemyType==2 || m_iEnemyType==1 || m_iEnemyType==3) {
		if(bFireBullet) m_dFireAnimTime = 250.0; //250 ms
		m_dFireAnimTime -= i_dTime;
		m_pstImageRect = &m_pstAnimRects[2*(int)(m_dDirection/90)];
		if(m_dFireAnimTime<=0) m_dFireAnimTime = 0;
		else m_pstImageRect++; //next rect
	}

	if(iPlaySound != -1) {
		CalcSoundVolPan(s_pclListner, &m_fVol, &m_fPan);
		if(C_Sound::GetSoundObject()) C_Sound::GetSoundObject()->Play(iPlaySound, m_fVol, m_fPan); //play sound
	}

	return iResult;
}
