// C_LandingZone, C_CheckPoint, C_ComputerEnemy, C_Door

#ifndef _MAP_OBJECTS_H
#define _MAP_OBJECTS_H

#include "netmsg.h"
#include "netmsgserver.h"

#include "common.h"

#include "player.h"
#include "object.h"
#include "bullet.h"
#include "tilemap.h"
#include "timer.h"
#include "glgraph.h"
#include "font.h"

#include <vector>

class C_LandingZone : public C_Object
{
public: 
	C_LandingZone(C_Map *i_pclMap, int i_iX, int i_iY, int i_iWidth,
		bool i_bHomeBase, bool i_bShowAntenna, bool i_bShowHouse,
		std::vector<int> *i_pclCargoList, bool i_bExtraLife, C_NetMsgServer *i_pclNetMsgServer);
	~C_LandingZone();

	int GetZoneSize();
	int GetTotalCargo();
	int PopCargo(); //-1 if empty
	void PushCargo(int i_iWeight);
	void TakeExtraLife();

	bool IsHomeBase() {return m_bHomeBase;};
	void UpdateLanding(C_Player *i_pclPlayer, double i_dTime);
	void UpdateFuel(C_ShipPlayer *i_pclPlayer, double i_dTime);
	void Draw();
private:
	static C_Image *s_pclLandingZoneImg;
	static int s_iObjCount;

	C_Map *m_pclMap;
	int m_iWidth, m_iHeight;
	int m_iZoneSize; //width in tiles

	//8 possible players (AI and normal)
	bool m_bPlayerIsLanded[8];
	double m_dTimeLanded[8];

	//other
	bool m_bHomeBase;
	bool m_bShowAntenna, m_bShowHouse;
	double m_dAntennaFrame;
	std::vector<int> *m_pclCargoList; //array of cargo weights (small = 1..5,6..10,11..15,16..20+ = huge)
	bool m_bExtraLife;

	C_NetMsgServer *m_pclNetMsgServer;
};
typedef std::vector<C_LandingZone*> C_LandingZoneVector;

////////////////////////////////////////////////////////////////////////////////
struct S_CheckPointInfo
{
	int iX1, iY1;
	int iX2, iY2;
};

class C_CheckPoint : public C_Object
{
public:
	C_CheckPoint(C_Map *i_pclMap, S_CheckPointInfo *i_pstCPInfo, int i_iCPNum);
	~C_CheckPoint();

	bool PlayerAtCP(C_ShipPlayer *i_pclPlayer);
	void SetBlinkState(bool i_bBlinkState);

	void Update(double i_dTime);
	void Draw();

	void GetCheckPointInfo(S_CheckPointInfo *o_pstCP);
private:
	int m_iPosX1, m_iPosY1;
	int m_iPosX2, m_iPosY2;
	int m_iCPNum;
	char m_szString[4];

	C_Map *m_pclMap;
	bool m_bBlinkState;
	double m_dBlinkTimer;

	static int s_iObjCount;
	static C_Font *s_pclFont;
	static C_Image *s_pclCpBkImg;
};
typedef std::vector<C_CheckPoint*> C_CheckPointVector;

////////////////////////////////////////////////////////////////////////////////
struct S_DoorInfo
{
	int iSize;
	bool bHorizontal;
	int iOpenedForTime;
	int iClosedForTime;
	int iOpeningSpeed; //pixels/sec
	int iClosingSpeed; //pixels/sec
	int iNumButtons; //0 -> interval opening
	POINT stButtonPos[3];
};

class C_Door : public C_Object
{
public: 
	C_Door(C_Map *i_pclMap, int i_iX, int i_iY, S_DoorInfo *i_pstInfo, int i_iID, bool i_bMaster, C_NetMsg *i_pclNetMsg);
	~C_Door();

	void SetMaster(bool i_bMaster) {m_bMaster = i_bMaster;};
	bool CheckCollission(C_Object *i_pclObject, bool i_bIsBullet);
	void Update(double i_dTime);
	void Draw();

private:
	void SendDoorAction(bool i_bOpen);

	static C_Image *s_pclDoorImg;
	static int s_iObjCount;

	bool m_bMaster;
	C_Map *m_pclMap;

	//animation
	bool m_bIsOpening;
	double m_dOpenPos;
	double m_dStateTime;

	//behavior
	S_DoorInfo m_stInfo;

	C_NetMsg *m_pclNetMsg;
};
typedef std::vector<C_Door*> C_DoorVector;

////////////////////////////////////////////////////////////////////////////////
#define MAX_WAYPOINTS 16
struct S_EnemyInfo
{
	int iEnemyType;
	int iAngle;
	int iFireInterval;
	int iSpeed;
	int iNumWayPoints;
	POINT astWayPoint[MAX_WAYPOINTS];
};

typedef std::vector<C_Object **> C_ObjPPList;

class C_ComputerEnemy : public C_PlayerBase
{
public: 
	C_ComputerEnemy(C_Map *i_pclMap, C_BulletList *i_pclBulletList, C_NetMsg *i_pclNetMsg, bool i_bMaster, S_EnemyInfo *i_pstEnemyInfo, int i_iID);
	~C_ComputerEnemy();
	static void AddFireRangeObject(C_Object **i_ppclFireRangeObj);

	void SetMaster(bool i_bMaster);
	int Update(double i_dTime);
	void Hit(int i_iBulletUID);

	int GetScoreValue(); //returns the points this enemy is worth for the player

	void Draw();
protected:
	C_Bullet *CreateBullet(int i_iIndex);

	//common
	C_Map *m_pclMap;
	C_BulletList *m_pclBulletList;

	//network
	bool m_bMaster;
	C_NetMsg *m_pclNetMsg;
	S_NetMsg1 m_stMsg1;

	//movement
	void Move(double i_dTime);
	int m_iNumWayPoints, m_iCurWP, m_iSpeed;
	POINT m_pstWayPointList[MAX_WAYPOINTS];
	double m_dWayTime;

	//animation
	double m_dFrameNr;
	S_Rect *m_pstAnimRects;
	double m_dFireAnimTime;

	//enemy specific
	double m_dFireTime;
	POINT *m_pstFirePoint;
	int m_iFireInterval;
	int m_iEnemyType;
	int m_iHitsLeft;

	bool m_bFinished; //explosion ready

	//firerange
	bool InFireRange();
	static C_ObjPPList *s_pclRangeList;
private:
	static C_Image *s_pclEnemy0Img;
	static C_Image *s_pclEnemy1Img;
	static C_Image *s_pclEnemy2Img;
	static C_Image *s_pclEnemy3Img;
	static C_Image *s_pclEnemy4Img;
	static C_Image *s_pclEnemy5Img;
	static C_Image *s_pclEnemy6Img;
	static C_Image *s_pclEnemy7Img;

	static int s_iObjCount;

	int m_iPathTimer;
};
typedef std::vector<C_ComputerEnemy*> C_ComputerEnemyVector;

////////////////////////////////////////////////////////////////////////////////

struct S_ExtraObjects
{
	C_LandingZoneVector   *pclLandingZoneList;
	C_ComputerEnemyVector *pclComputerEnemyList;
	C_CheckPointVector    *pclCheckPointList;
	C_DoorVector          *pclDoorList;
	C_BulletList          *pclBulletList;
	C_NetMsg              *pclNetMsg;
	C_NetMsgServer        *pclNetMsgServer;
	bool                  bMaster;
};

#endif
