// C_NetMsgServer

//class to send and receive server messages using TCP.
//this class is used both by the game and the server.
// - send/receive
//  - object data packets (S_NetMsg2) (used when C_NetMsg is run in TCP mode)
//  - server data packets (S_NetMsg1)
//  - chat messgaes
//  - server hiscore list
// - handle replay creation/play
//  to easily create and play a replay a "hook in" for this is added to this class.
//  - during gameplay, all important messages are saved through the replay object
//  - during replay, messages are made available at the correct timestep through the replay object
//  - when the class is used from the server, replay functions are never used

//replay play
//0      0     always the case for the server
//0      1     forbidden
//1      x     always the case for the game

//0      0     don't add to replay, send on network, recv from network
//1      0     add to replay, send on network, recv from network
//1      1     don't add to replay, don't send on network, recv from replay


#include "netmsgserver.h"
#include "replay.h"

#pragma pack(1)
struct S_MinHeader
{
	uint16_t iSize;
	uint16_t iType;
};
#pragma pack()

C_NetMsgServer::C_NetMsgServer(C_Socket *i_pclSocket)
{
	m_pclSocket = i_pclSocket;
	m_pclReplay = NULL;
	m_bRunReplay = false;
}

C_NetMsgServer::~C_NetMsgServer()
{
}

void C_NetMsgServer::SetReplayMode(bool bRunReplay)
{
	m_bRunReplay = bRunReplay; //from replay or network
}

void C_NetMsgServer::AddReplayMsg1(S_NetMsg1 *i_pstMsg1)
{
	if(m_pclReplay && !m_bRunReplay) {
		//no type filtering for this message (all i_pstMsg1->iType)
		S_ReplayAction stAction;
		stAction.iMsg = 1;
		stAction.iId = i_pstMsg1->iSenderId;
		stAction.stMsg1 = *i_pstMsg1;
		m_pclReplay->Add(&stAction);
	}
}

void C_NetMsgServer::AddReplayMsg2(S_NetMsg2 *i_pstMsg2)
{
	if(m_pclReplay && !m_bRunReplay) {
		//filter on type
		int iType = i_pstMsg2->iType;
		if(iType==N_MSG2_GAMEOVER || iType==N_MSG2_SCORE
			|| iType==N_MSG2_NUMKILLS || iType==N_MSG2_NUMLAPS
			|| iType==N_MSG2_CARGO_R || iType==N_MSG2_CARGORESPAWN
			|| iType==N_MSG2_LIFE_R || iType==N_MSG2_NEXTLEVEL)
		{
			S_ReplayAction stAction;
			stAction.iMsg = 2;
			stAction.iId = (int8_t)i_pstMsg2->iPlayerId;
			stAction.stMsg2 = *i_pstMsg2;
			m_pclReplay->Add(&stAction);
		}
	}
}

void C_NetMsgServer::SendMsg1(S_NetMsg1 *i_pstMsg1)
{
	i_pstMsg1->iSize = sizeof(S_NetMsg1);

	if(!m_pclReplay || !m_bRunReplay) m_pclSocket->SendQueued((char*)i_pstMsg1, sizeof(S_NetMsg1));

	//replay
	AddReplayMsg1(i_pstMsg1);
}

void C_NetMsgServer::SendMsg2WithId(S_NetMsg2 *i_pstMsg2)
{
	i_pstMsg2->iGameId   = (int16_t)m_iGameId;
	i_pstMsg2->iPlayerId = (int16_t)m_iPlayerId;
	SendMsg2(i_pstMsg2);
}

void C_NetMsgServer::SendMsg2(S_NetMsg2 *i_pstMsg2)
{
	i_pstMsg2->iSize = sizeof(S_NetMsg2);

	if(!m_pclReplay || !m_bRunReplay) m_pclSocket->SendQueued((char*)i_pstMsg2, sizeof(S_NetMsg2));

	//replay
	if(i_pstMsg2->iType!=N_MSG2_CARGORESPAWN) { //filter outgoing messages
		AddReplayMsg2(i_pstMsg2);
	}
}

void C_NetMsgServer::SendChatMsg(char *i_szChatString)
{
	if(!m_bRunReplay || !m_pclReplay) {

		int iType = N_MSG2_CHAT;
		int iSize = (int)strlen(i_szChatString)+1;
		char *pcMem = new char[iSize+sizeof(S_MinHeader)];
	
		S_MinHeader stTemp = {(uint16_t)iSize, (uint16_t)iType};
	
		memcpy(pcMem,                     &stTemp,        sizeof(stTemp));
		memcpy(pcMem+sizeof(S_MinHeader), i_szChatString, iSize);
		m_pclSocket->SendQueued(pcMem, sizeof(S_MinHeader)+iSize);
		delete[] pcMem;
	}
}

void C_NetMsgServer::SendHiScoreMsg(S_HiScore *i_pstHiScore)
{
	if(!m_bRunReplay || !m_pclReplay) {

		int iType = N_MSG2_HISCORELIST;
		int iSize = sizeof(S_HiScore);
		char *pcMem = new char[sizeof(S_HiScore)+sizeof(S_MinHeader)];
	
		S_MinHeader stTemp = {(uint16_t)iSize, (uint16_t)iType};
	
		memcpy(pcMem,                     &stTemp,      sizeof(stTemp));
		memcpy(pcMem+sizeof(S_MinHeader), i_pstHiScore, sizeof(S_HiScore));
		m_pclSocket->SendQueued(pcMem, sizeof(S_MinHeader)+sizeof(S_HiScore));
		delete[] pcMem;
	}
}

int C_NetMsgServer::GetMsg(S_NetMsg1 *o_pstMsg1, S_NetMsg2 *o_pstMsg2, S_HiScore *o_pstScore, char *o_szChatString, int i_iMaxChatLen)
{
	int  iLength;
	bool bResult;
	S_MinHeader stTemp;

	//polling of socket
	m_pclSocket->WaitForReadEvent(0);
	m_pclSocket->WaitForWriteEvent(0);

	//replay
	if(m_pclReplay && m_bRunReplay) {
		S_ReplayAction stAction;
		if(m_pclReplay->Get(&stAction)) {
			if(stAction.iMsg == 1) {
				*o_pstMsg1 = stAction.stMsg1;
				return 4;
			}
			if(stAction.iMsg == 2) {
				*o_pstMsg2 = stAction.stMsg2;
				return 1;
			}
		}
		return 0;
	}

	iLength = sizeof(S_MinHeader);
	bResult = m_pclSocket->PeekRecvBuffer((char*)&stTemp, &iLength);
	if(bResult) {
		if(stTemp.iSize>12000) { //data size safety
			//data too large, ignore this header and hope for sync in the future
			m_pclSocket->GetRecvBuffer((char*)&stTemp, &iLength);
		} else {
			int iAvail = m_pclSocket->RecvBufferSize();
			if(iAvail < stTemp.iSize) return 0;
			int iSize = stTemp.iSize;

			switch(stTemp.iType) {
				case N_MSG1:
					if(iSize!=sizeof(S_NetMsg1)) break;

					bResult = m_pclSocket->GetRecvBuffer((char*)o_pstMsg1, &iSize);
					if(bResult) AddReplayMsg1(o_pstMsg1); //replay
					if(bResult) return 4;
					break;
				case N_MSG2_CHAT:
					m_pclSocket->GetRecvBuffer((char*)&stTemp, &iLength);

					if(iSize>i_iMaxChatLen) o_szChatString = NULL;

					bResult = m_pclSocket->GetRecvBuffer(o_szChatString, &iSize);
					if(!o_szChatString) {
						bResult = false;
						break;
					}
					o_szChatString[i_iMaxChatLen-1] = 0; //be sure to null terminate
					if(bResult) return 3;
					break;
				case N_MSG2_HISCORELIST:
					m_pclSocket->GetRecvBuffer((char*)&stTemp, &iLength);

					if(!o_pstScore) break;
					if(iSize!=sizeof(S_HiScore)) break;

					bResult = m_pclSocket->GetRecvBuffer((char*)o_pstScore, &iSize);
					if(bResult) return 2;
					break;
				default:
					if(iSize!=sizeof(S_NetMsg2)) break;

					bResult = m_pclSocket->GetRecvBuffer((char*)o_pstMsg2, &iSize);
					if(bResult) AddReplayMsg2(o_pstMsg2); //replay
					if(bResult) return 1;
					break;
			}
		}
	}

	return 0;
}
