// C_Object

//base class for all objects in the game.
//keeps direction, position, velocity, and acceleration, and methods to update them with time.
//also
// - a simple drawing metod (to draw an image at a given map position)
// sometimes objects need a more complex drawing method, but can still use this for basic drawing.
// - collission detection
// - helper functions for distance calculation

#include <math.h>
#include "object.h"

#include "image_utils.h"

double C_Object::PointDistance(int i_iX1, int i_iY1, int i_iX2, int i_iY2)
{
	int iDistX = i_iX1 - i_iX2;
	int iDistY = i_iY1 - i_iY2; 
	return sqrt((double)(iDistX*iDistX + iDistY*iDistY));
}

double C_Object::GetDistanceTo(C_Object *i_pclObj)
{
	double dCentreX1, dCentreY1, dCentreX2, dCentreY2;
	double dDistX, dDistY;
	int iSizeX, iSizeY;

	GetSize(&iSizeX, &iSizeY);
	dCentreX1 = m_dPosX+iSizeX*0.5;
	dCentreY1 = m_dPosY+iSizeY*0.5;

	i_pclObj->GetSize(&iSizeX, &iSizeY);
	dCentreX2 = i_pclObj->m_dPosX+iSizeX*0.5;
	dCentreY2 = i_pclObj->m_dPosY+iSizeY*0.5;

	dDistX = dCentreX1 - dCentreX2;
	dDistY = dCentreY1 - dCentreY2;

	return sqrt(dDistX*dDistX + dDistY*dDistY);
}

C_Object::C_Object()
{
	m_dTime = 0.0;
	m_iDirStep = OBJECT_HEADING_STEP;
	m_pclGraph = C_GraphWrapperGL::GetGraphWrapperObject();
	m_bCanCollide = true;
	m_pclImage = NULL;
	m_pstImageRect = NULL;

	m_iOwnerID = -1;

	m_dAcceleration = 0.0;
	m_dResistanceX = m_dResistanceY = 0.0;
	m_dGravityX = 0;
	m_dGravityY = 0;
}

void C_Object::GetPos(int *o_iX, int *o_iY)
{
	*o_iX = (int)m_dPosX;
	*o_iY = (int)m_dPosY;
}

void C_Object::SetPos(int i_iX, int i_iY)
{
	m_dPosX = i_iX;
	m_dPosY = i_iY;
}

bool C_Object::GetSize(int *o_iSizeW, int *o_iSizeH)
{
	if(!m_pclImage) *o_iSizeW = *o_iSizeH = 0;
	else if(m_pstImageRect) {
		*o_iSizeW = m_pstImageRect->width;
		*o_iSizeH = m_pstImageRect->height;
	} else m_pclImage->GetInfo(o_iSizeW, o_iSizeH, NULL);
	return true;
}

bool C_Object::GetCenterPos(int *o_iCenterX, int *o_iCenterY)
{
	int iSizeX, iSizeY;
	if(!GetSize(&iSizeX, &iSizeY)) return false;

	*o_iCenterX = (int)(m_dPosX + iSizeX*0.5);
	*o_iCenterY = (int)(m_dPosY + iSizeY*0.5);
	return true;
}

bool C_Object::GetCenterPos(POINT *o_pstCenter)
{
	int iSizeX, iSizeY;
	if(!GetSize(&iSizeX, &iSizeY)) return false;

	o_pstCenter->x = (int)(m_dPosX + iSizeX*0.5);
	o_pstCenter->y = (int)(m_dPosY + iSizeY*0.5);
	return true;
}

void C_Object::OUpdate(double i_dTime)
{
	m_dTime = (i_dTime/1000);

	//calc velocity
	m_iRealDir = (int)(m_dDirection/m_iDirStep);

	double ax = cos(m_iRealDir*(m_iDirStep*(PI/180))) * m_dAcceleration;
	double ay = sin(m_iRealDir*(m_iDirStep*(PI/180))) * m_dAcceleration;

	ax += m_dGravityX;
	ay += m_dGravityY;
	ax -= m_dVelX*m_dResistanceX;
	ay -= m_dVelY*m_dResistanceY;

	//calc position
	m_dPosX += (m_dVelX * m_dTime) + (0.5*ax*m_dTime*m_dTime); //??
	m_dPosY += (m_dVelY * m_dTime) + (0.5*ay*m_dTime*m_dTime); //??

	//calc vel
	m_dVelX += ax * m_dTime;
	m_dVelY += ay * m_dTime;
}

void C_Object::ODraw(C_Image *i_pclImg, S_Rect *i_pstRect, int i_iX, int i_iY)
{
	if(!i_pclImg) return;

	bool bDraw = true;
	int iPosX = (int)m_dPosX - i_iX;
	int iPosY = (int)m_dPosY - i_iY;
	int iWidth, iHeight;
	S_Rect stRect;

	if(i_pstRect) {
		stRect = *i_pstRect;
		iWidth = stRect.width;
		iHeight = stRect.height;
	} else {
		i_pclImg->GetInfo(&iWidth, &iHeight, NULL);
		stRect.x = stRect.y = 0;
		stRect.width = iWidth;
		stRect.height = iHeight;
	}

	//clipping x
	if(iPosX < 0) {
		stRect.x -= iPosX;
		stRect.width += iPosX;
		iPosX = 0;
	} else if(iPosX+iWidth > m_stScreenArea.width) {
		stRect.width -= (iPosX+iWidth) - m_stScreenArea.width;
	}
	if(stRect.width <= 0) bDraw = false;
	else { //clipping y
		if(iPosY < 0) {
			stRect.y -= iPosY;
			stRect.height += iPosY;
			iPosY = 0;
		} else if(iPosY+iHeight > m_stScreenArea.height) {
			stRect.height -= (iPosY+iHeight) - m_stScreenArea.height;
		}
		if(stRect.height <= 0) bDraw = false;
	}
	//draw?
	if(bDraw)
		m_pclGraph->Blt(i_pclImg, &stRect, iPosX+m_stScreenArea.x, iPosY+m_stScreenArea.y, true);
}

bool C_Object::OCheckCollision(C_Image *i_pclOtherImg, int i_iOtherX, int i_iOtherY)
{
	if(!m_bCanCollide || !m_pclImage) return false; //no collision

	return ImageUtils::Compare(m_pclImage, (int)m_dPosX, (int)m_dPosY, i_pclOtherImg, i_iOtherX, i_iOtherY, m_pstImageRect, NULL);
}

bool C_Object::OCheckCollision(C_Object *i_pclObject)
{
	if(!m_bCanCollide || !m_pclImage || !i_pclObject->m_bCanCollide || !i_pclObject->m_pclImage) return false; //no collision

	return ImageUtils::Compare(m_pclImage, (int)m_dPosX, (int)m_dPosY, i_pclObject->m_pclImage, (int)i_pclObject->m_dPosX, (int)i_pclObject->m_dPosY, m_pstImageRect, i_pclObject->m_pstImageRect);
}
