// C_Replay

//class to handle replay functionality
// - load from disk
// - load from memory (received from web server)
// - save to disk
// - save to memory (before sending to web server)
// - play (get messages at the correct timeslot)
// - create (add messages to a big list)

#include "replay.h"

//constructur
C_Replay::C_Replay()
{
	m_iCurGetPos = 0;
	m_iCurTimeSlot = 0;
	memset(&m_stHeader, 0, sizeof(m_stHeader));
	m_pclActions = new C_ActionVector();
}

//copy constructur
C_Replay::C_Replay(C_Replay *pclReplay)
{
	m_iCurTimeSlot = pclReplay->m_iCurTimeSlot;
	m_iCurGetPos = pclReplay->m_iCurGetPos;
	m_stHeader = pclReplay->m_stHeader;
	m_pclActions = new C_ActionVector(*pclReplay->m_pclActions);
}

//destructor
C_Replay::~C_Replay()
{
	delete m_pclActions;
}

//clears all but header
void C_Replay::Reset()
{
	m_iCurGetPos = 0;
	m_iCurTimeSlot = 0;
	m_pclActions->clear();
}

void C_Replay::ResetBeforePlay()
{
	m_iCurGetPos = 0;
	m_iCurTimeSlot = 0;
}

void C_Replay::IncTimeSlot()
{
	m_iCurTimeSlot++;
}

//set header
void C_Replay::SetGameInfo(S_ReplayHeader *i_pstHeader)
{
	m_stHeader = *i_pstHeader;
}

//get header
void C_Replay::GetGameInfo(S_ReplayHeader *o_pstHeader)
{
	*o_pstHeader = m_stHeader;
}

//add a new replay action last in the list
// at the current timeslot
bool C_Replay::Add(S_ReplayAction *i_pstAction)
{
	i_pstAction->iTimeSlot = m_iCurTimeSlot;
	m_pclActions->push_back(*i_pstAction);
	return true;
}

//get replay action (if any) for the current timeslot
bool C_Replay::Get(S_ReplayAction *o_pstAction)
{
	S_ReplayAction *pstAction;
	int iSize = (int)m_pclActions->size();
	if(m_iCurGetPos<iSize) {
		pstAction = &(*m_pclActions)[m_iCurGetPos];
		if(pstAction->iTimeSlot <= m_iCurTimeSlot) {
			//do it as fast as possible if less than current (never skip actions)
			m_iCurGetPos += 1;
			*o_pstAction = *pstAction;
			return true;
		}
	}
	return false;
}

//load from disk
bool C_Replay::Load(char *i_szFile)
{
	S_ReplayAction stAction;
	FILE *pFile = fopen(i_szFile, "rb");
	Reset();
	if(!pFile) return false;
	if(fread(&m_stHeader, sizeof(m_stHeader), 1, pFile)!=1) return false;
	while(fread(&stAction, sizeof(stAction), 1, pFile)==1) {
		m_pclActions->push_back(stAction);
	}
	fclose(pFile);
	return true;
}

//save to disk
bool C_Replay::Save(char *i_szFile)
{
	FILE *pFile = fopen(i_szFile, "wb");
	if(!pFile) return false;
	if(fwrite(&m_stHeader, sizeof(m_stHeader), 1, pFile)!=1) return false;
	int i, iSize = (int)m_pclActions->size();
	for(i=0; i<iSize; i++) {
		fwrite(&(*m_pclActions)[i], sizeof(S_ReplayAction), 1, pFile);
	}
	fclose(pFile);
	return true;
}

//load from memory after received from web-server
bool C_Replay::LoadFromMem(char *i_pMem, int i_iSize)
{
	bool bResult = true;
	S_ReplayAction stAction;
	Reset();
	if(!i_pMem || i_iSize<(int)sizeof(m_stHeader)) return false;
	memcpy(&m_stHeader, i_pMem, sizeof(m_stHeader));
	if(VerifyHeader()) {
		int iPos = sizeof(m_stHeader);
		while((int)(iPos+sizeof(stAction))<=i_iSize) {
			memcpy(&stAction, i_pMem+iPos, sizeof(stAction));
			if(!VerifyElement(&stAction)) {
				bResult = false;
				break;
			}
			m_pclActions->push_back(stAction);
			iPos += sizeof(stAction);
		}
	} else bResult = false;
	if(!bResult) Reset();
	return bResult;
}

//save to memory before sending to web-server
bool C_Replay::SaveToMem(char *o_pMem, int *io_iSize)
{
	if(!io_iSize) return false;
	int i, iSize = (int)m_pclActions->size();
	int iTotSize = sizeof(m_stHeader)+iSize*sizeof(S_ReplayAction);
	int iMaxSize = *io_iSize;
	*io_iSize = iTotSize;
	if(iTotSize<iMaxSize || !o_pMem) return false;
	memcpy(o_pMem, &m_stHeader, sizeof(m_stHeader));
	int iPos = sizeof(m_stHeader);
	for(i=0; i<iSize; i++) {
		memcpy(o_pMem+iPos, &(*m_pclActions)[i], sizeof(S_ReplayAction));
		iPos += sizeof(S_ReplayAction);
	}
	return true;
}

//private helper used for verification of replay from web-server (and disk)
bool C_Replay::VerifyString(char *i_szText, int i_iSize)
{
	int i;
	for(i=0; i<i_iSize; i++) {
		if(i_szText[i]==0) {
			return true;
		}
		if(i_szText[i]<32) return false;
	}
	return false;
}

//private helper used for verification of replay from web-server (and disk)
bool C_Replay::VerifyHeader()
{
	int i;
	if(m_stHeader.iVersion!=1) return false;
	if(m_stHeader.iNumPlayers<1 || m_stHeader.iNumPlayers>8) return false;
	if(m_stHeader.iLevelType<1 || m_stHeader.iLevelType>4) return false;

	if(!VerifyString(m_stHeader.szMapFile, sizeof(m_stHeader.szMapFile))) return false;
	for(i=0; i<8; i++) if(!VerifyString(m_stHeader.szPlayerName[i], MAXPLAYERNAMELENGTH)) return false;

	return true;
}

//private helper used for verification of replay from web-server (and disk)
bool C_Replay::VerifyElement(S_ReplayAction *i_pstAction)
{
	if(i_pstAction->iMsg!=1 && i_pstAction->iMsg!=2) return false;
	if(i_pstAction->iTimeSlot<0) return false;
	//if(i_pstAction->iId<-1) return false; //in N_MSG2_CARGORESPAWN message this was uninitialized (fixed now, but not in v1.67)

	if(i_pstAction->iMsg==1) {
		if(i_pstAction->stMsg1.iSize != sizeof(S_NetMsg1)) return false;
		if(i_pstAction->stMsg1.iType != N_MSG1) return false;
		//TODO further verify i_pstAction->stMsg1
	} else {
		int iType = i_pstAction->stMsg2.iType;
		if(iType!=N_MSG2_GAMEOVER && iType!=N_MSG2_SCORE
			&& iType!=N_MSG2_NUMKILLS && iType!=N_MSG2_NUMLAPS
			&& iType!=N_MSG2_CARGO_R && iType!=N_MSG2_CARGORESPAWN
			&& iType!=N_MSG2_LIFE_R && iType!=N_MSG2_NEXTLEVEL) return false;
		if(i_pstAction->stMsg2.iSize != sizeof(S_NetMsg2)) return false;
		//TODO further verify i_pstAction->stMsg2
	}
	return true;
}
