// C_Map

//class to handle a map
// - load from disk
// - draw
// - collission detection
// - methods useful for the new AI

#include "mapobjects.h"

#include "tilemap.h"
#include "fileresource.h"
#include "global.h"

#include <vector>

using std::vector;

typedef vector<int> IntVector;


//constructor to create a map from an existing file
C_Map::C_Map(char *i_szFilename, S_Rect *i_pstDrawarea, S_ExtraObjects *io_pclExtra, bool *o_bResult)
{
	m_pGrid = NULL;
	m_stTileRects = NULL;
	m_pclTilesImg = NULL;
	m_pclTilesImg_Sysmem = NULL;

	//allocate memory for map, and set defaults
	m_iLastX = m_iLastY = 0;
	m_stDrawarea.x = 0;
	m_stDrawarea.y = 0;
	m_stDrawarea.width = 0;
	m_stDrawarea.height = 0;
	if(i_pstDrawarea) memcpy(&m_stDrawarea, i_pstDrawarea, sizeof(m_stDrawarea));

	//load the mapfile
	m_iLevelType = MAP_MISSION; //assume
	m_iRaceLaps = 0;
	m_iNumDoors = 0;
	m_iMaxPlayers = m_iMinPlayers = 1;
	m_iWidth = 0;
	m_iHeight = 0;
	bool bResult = Load(i_szFilename, io_pclExtra);

	//save the area entered by the user if wrapparound mode should change later
	m_iDrawareaWidthSave = m_stDrawarea.width;
	m_iDrawareaHeightSave = m_stDrawarea.height;
	SetMapWrapping(false, false);

	//get graphwrapper interface
	m_pclGraph = C_GraphWrapperGL::GetGraphWrapperObject();

	if(o_bResult) *o_bResult = bResult;
	m_bMapLoaded = bResult;
}

//destructor
C_Map::~C_Map()
{
	FreeGrid();

	delete[] m_stTileRects;
	delete m_pclTilesImg;
	delete m_pclTilesImg_Sysmem;
}

//free the map grid, note that m_iHeight must be the height of the allocated grid
void C_Map::FreeGrid()
{
	int i;

	if(m_pGrid) {
		for(i=0; i<m_iHeight; i++) {
			delete[] m_pGrid[i];
		}
		delete[] m_pGrid;
		m_pGrid = NULL;
	}
}

//set the draw area
void C_Map::SetScreenRect(S_Rect *i_pstDrawarea)
{
	m_stDrawarea = *i_pstDrawarea;
	m_iDrawareaWidthSave = m_stDrawarea.width;
	m_iDrawareaHeightSave = m_stDrawarea.height;
	SetMapWrapping(m_iWrappX, m_iWrappY);
}

//loads a map (the header, map grid data and tileset)
bool C_Map::Load(char *i_szMapHeader, S_ExtraObjects *io_pclExtra)
{
	int i, j;
	bool bResult = false;

	//free old map if exists
	FreeGrid();
	m_iLevelType = MAP_MISSION;
	m_iRaceLaps = 0;
	m_iMinPlayers = 1;
	m_iMaxPlayers = 1;
	m_iNumDoors = 0;

	char szExternalFile[MAX_PATH];
	C_Resource *pclResource = new C_Resource();
	bResult = pclResource->SetFilename(i_szMapHeader, C_Global::szResourceFile);
	if(!bResult) { //not in resource, try external file
		sprintf(szExternalFile, "%s/external_maps/%s", C_Global::szDataPath, i_szMapHeader);
		bResult = pclResource->SetFilename(szExternalFile);
	}
	if(!bResult) {
		delete pclResource;
		return false;
	}
	int iSize = pclResource->GetSize();
	char *szFile = new char[iSize+1];
	bResult = pclResource->Read(szFile, iSize);
	szFile[iSize] = 0; //zero terminate
	delete pclResource;
	if(!bResult) {
		delete[] szFile;
		return false;
	}

	int iStartPosCounter = 0;
	int iEnemyCounter = 0;
	int iZoneCounter  = 0;
	int iLineIndex    = 0;
	char szLine[512], szCommand[32];
	char szTemp[64];
	double dGravityX = DEFAULT_SHIPGRAVITYBASEX, dGravityY = DEFAULT_SHIPGRAVITYBASEY;
	double dResistanceX = DEFAULT_SHIPRESISTANCEX, dResistanceY = DEFAULT_SHIPRESISTANCEY;
	double dSteeringSpeed = DEFAULT_SHIPSTEARSPEED, dThrust = DEFAULT_SHIPTHRUST;
	while(C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex)) {

		if(szCommand[0]!='*') continue;

		if(strcmp(szCommand, "*MAPTYPE") == 0) {
			sscanf(szLine, "%s %s %d", szCommand, szTemp, &m_iRaceLaps);
			if(strcmp(szTemp, "RACE") == 0) m_iLevelType = MAP_RACE;
			else if(strcmp(szTemp, "DOGFIGHT") == 0) {
				m_iLevelType = MAP_DOGFIGHT;
				if(m_iMinPlayers<2) m_iMinPlayers = 2;
			} else if(strcmp(szTemp, "MISSION_COOP") == 0) m_iLevelType = MAP_MISSION_COOP;
			//else if(strcmp(szTemp, "MISSION") == 0) m_iLevelType = MAP_MISSION;
		} else if(strcmp(szCommand, "*MAPSIZE") == 0) {
			sscanf(szLine, "%s %d %d", szCommand, &m_stHeader.iWidth, &m_stHeader.iHeight);
		} else if(strcmp(szCommand, "*MAPFILE") == 0) {
			sscanf(szLine, "%s %s", szCommand, m_stHeader.szGridfile);
		} else if(strcmp(szCommand, "*TILESIZE") == 0) {
			sscanf(szLine, "%s %d", szCommand, &m_stHeader.iTileSize);
		} else if(strcmp(szCommand, "*TILESET") == 0) {
			sscanf(szLine, "%s %s", szCommand, m_stHeader.szTileset);
		} else if(strcmp(szCommand, "*GRAVITY") == 0) {
			sscanf(szLine, "%s %lf %lf", szCommand, &dGravityX, &dGravityY);
		} else if(strcmp(szCommand, "*RESISTANCE") == 0) {
			sscanf(szLine, "%s %lf %lf", szCommand, &dResistanceX, &dResistanceY);
		} else if(strcmp(szCommand, "*PLAYERSTARTPOS") == 0) {
			if(iStartPosCounter<8) sscanf(szLine, "%s %d %d", szCommand, &m_stPlayerStartPos[iStartPosCounter].x, &m_stPlayerStartPos[iStartPosCounter].y);
			iStartPosCounter++;
		} else if(strcmp(szCommand, "*MAXPLAYERS") == 0) {
			sscanf(szLine, "%s %d", szCommand, &m_iMaxPlayers);
		} else if(strcmp(szCommand, "*MINPLAYERS") == 0) {
			sscanf(szLine, "%s %d", szCommand, &m_iMinPlayers);
		} else if(strcmp(szCommand, "*LANDINGZONE") == 0) {
			int x, y, w, iNumCargo = 0;
			std::vector<int> *pclCargoList = NULL;
			char szHomeBase[32], szExtra[32], szAntenna[32], szHouse[32];
			int iFields = sscanf(szLine, "%s %d %d %d %s %s %s %s %d",
				szCommand, &x, &y, &w, szHomeBase, szExtra, szAntenna, szHouse, &iNumCargo);
			bool bHomeBase  = (strcmp(szHomeBase, "HOMEBASE")  == 0);
			bool bExtraLife = (strcmp(szExtra,    "EXTRALIFE") == 0);
			bool bAntenna   = (strcmp(szAntenna,  "ANTENNA")   == 0);
			bool bHouse     = (strcmp(szHouse,    "WAREHOUSE") == 0);

			if(io_pclExtra) {
				if(iNumCargo>0) {
					int iCargos[12];
					iFields = sscanf(szLine, "%s %d %d %d %s %s %s %s %d %d %d %d %d %d %d %d %d %d %d %d %d",
						szCommand, &x, &y, &w, szHomeBase, szExtra, szAntenna, szHouse, &iNumCargo, &iCargos[0],
						&iCargos[1], &iCargos[2], &iCargos[3], &iCargos[4], &iCargos[5], &iCargos[6],
						&iCargos[7], &iCargos[8], &iCargos[9], &iCargos[10], &iCargos[11]);
					pclCargoList = new std::vector<int>;
					for(j=0; j<iNumCargo; j++)
						pclCargoList->push_back(iCargos[j]);
				}

				C_LandingZone *pclZone = new C_LandingZone(this, x, y, w, bHomeBase, bAntenna, bHouse,
					pclCargoList, bExtraLife, io_pclExtra->pclNetMsgServer);
				pclZone->SetOwnerID(iZoneCounter++);
				io_pclExtra->pclLandingZoneList->push_back(pclZone);
			}
		} else if(strcmp(szCommand, "*CHECKPOINTS") == 0) {
			int iNumCP = 0;
			sscanf(szLine, "%s %d", szCommand, &iNumCP);

			//read checkpoints
			S_CheckPointInfo stCPInfo;
			for(j=0; j<iNumCP; j++) {
				C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex);
				sscanf(szLine, "%s %d %d %d %d", szCommand, &stCPInfo.iX1, &stCPInfo.iY1, &stCPInfo.iX2, &stCPInfo.iY2);

				if(io_pclExtra) {
					C_CheckPoint *pclCP = new C_CheckPoint(this, &stCPInfo, j);
					io_pclExtra->pclCheckPointList->push_back(pclCP);
					if(j==0) pclCP->SetBlinkState(true); //first CP is the one to get first
				}
			}
		} else if(strcmp(szCommand, "*ENEMY") == 0) {
			S_EnemyInfo stParams;
			sscanf(szLine, "%s %d", szCommand, &stParams.iEnemyType);

			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*ANGLE"
			sscanf(szLine, "%s %d", szCommand, &stParams.iAngle);
			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*FIREINTERVAL"
			sscanf(szLine, "%s %d", szCommand, &stParams.iFireInterval);
			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*SPEED"
			sscanf(szLine, "%s %d", szCommand, &stParams.iSpeed);

			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*NUMWAYPOINTS"
			sscanf(szLine, "%s %d", szCommand, &stParams.iNumWayPoints);

			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*WAYPOINTSX"
			sscanf(szLine, "%s %d %d %d %d %d %d %d %d %d %d", szCommand, &stParams.astWayPoint[0].x, &stParams.astWayPoint[1].x,
				&stParams.astWayPoint[2].x, &stParams.astWayPoint[3].x, &stParams.astWayPoint[4].x, &stParams.astWayPoint[5].x,
				&stParams.astWayPoint[6].x, &stParams.astWayPoint[7].x, &stParams.astWayPoint[8].x, &stParams.astWayPoint[9].x);

			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*WAYPOINTSY"
			sscanf(szLine, "%s %d %d %d %d %d %d %d %d %d %d", szCommand, &stParams.astWayPoint[0].y, &stParams.astWayPoint[1].y,
				&stParams.astWayPoint[2].y, &stParams.astWayPoint[3].y, &stParams.astWayPoint[4].y, &stParams.astWayPoint[5].y,
				&stParams.astWayPoint[6].y, &stParams.astWayPoint[7].y, &stParams.astWayPoint[8].y, &stParams.astWayPoint[9].y);

			if(io_pclExtra) {
				//add enemy to list
				C_ComputerEnemy *pclComputerEnemy = new C_ComputerEnemy(this, io_pclExtra->pclBulletList, io_pclExtra->pclNetMsg,
					io_pclExtra->bMaster, &stParams, COMPUTER_ENEMY_BASEID+iEnemyCounter);
				io_pclExtra->pclComputerEnemyList->push_back(pclComputerEnemy);
			}
			iEnemyCounter++; //network id
		} else if(strcmp(szCommand, "*DOOR") == 0) {
			S_DoorInfo stParams;
			int x, y, iAngle;
			sscanf(szLine, "%s %d %d %d", szCommand, &x, &y, &stParams.iSize);

			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*ANGLE"
			sscanf(szLine, "%s %d", szCommand, &iAngle);
			stParams.bHorizontal = ((iAngle/90)%2)!=0;
			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*OPENTIME"
			sscanf(szLine, "%s %d", szCommand, &stParams.iOpenedForTime);
			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*CLOSETIME"
			sscanf(szLine, "%s %d", szCommand, &stParams.iClosedForTime);
			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*OPENSPEED"
			sscanf(szLine, "%s %d", szCommand, &stParams.iOpeningSpeed);
			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*CLOSESPEED"
			sscanf(szLine, "%s %d", szCommand, &stParams.iClosingSpeed);

			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*NUMBUTTONS"
			sscanf(szLine, "%s %d", szCommand, &stParams.iNumButtons);

			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*BUTTONSX"
			sscanf(szLine, "%s %d %d %d", szCommand, &stParams.stButtonPos[0].x, &stParams.stButtonPos[1].x, &stParams.stButtonPos[2].x);
			C_Global::GetNextLineAndCommand(szFile+iLineIndex, szLine, szCommand, &iLineIndex); //command "*BUTTONSY"
			sscanf(szLine, "%s %d %d %d", szCommand, &stParams.stButtonPos[0].y, &stParams.stButtonPos[1].y, &stParams.stButtonPos[2].y);

			//add door to list
			if(io_pclExtra) {
				C_Door *pclDoor = new C_Door(this, x, y, &stParams, DOOR_BASEID+m_iNumDoors, io_pclExtra->bMaster, io_pclExtra->pclNetMsg);
				io_pclExtra->pclDoorList->push_back(pclDoor);
			}
			m_iNumDoors++; //network id
		}

	}
	delete[] szFile;

	//header loaded
	//load map grid from file described in header
	m_iWidth    = m_stHeader.iWidth;
	m_iHeight   = m_stHeader.iHeight;
	m_iTileSize = m_stHeader.iTileSize;

	m_pGrid = new char*[m_iHeight]; //alloc rows
	for(i=0; i<m_iHeight; i++)     //alloc mem for each row
		m_pGrid[i] = new char[m_iWidth]; 

	pclResource = new C_Resource();
	bResult = pclResource->SetFilename(m_stHeader.szGridfile, C_Global::szResourceFile);
	if(!bResult) { //not in resource, try normal file
		sprintf(szExternalFile, "%s/external_maps/%s", C_Global::szDataPath, m_stHeader.szGridfile);
		bResult = pclResource->SetFilename(szExternalFile);
	}
	for(i=0; i<m_iHeight; i++) { //read each row
		if(!bResult) break;
		bResult = pclResource->Read(m_pGrid[i], m_iWidth*sizeof(**m_pGrid));
	}
	delete pclResource;
	if(!bResult) return false;

	C_Environment::Initialize(dGravityX, dGravityY, dResistanceX, dResistanceY, dSteeringSpeed, dThrust);

	//header and map grid loaded
	//load tileset
	return LoadTileSet(m_stHeader.szTileset);
}

//load tileset
bool C_Map::LoadTileSet(char *i_szTileset)
{
	int i, j;
	bool bResult = false;

	m_pclTilesImg = new C_Image(i_szTileset, C_Global::szResourceFile, &bResult);
	if(!bResult) { //not in resource, try normal file
		char szFileAndPath[MAX_PATH];
		sprintf(szFileAndPath, "%s/%s", C_Global::szDataPath, i_szTileset);
		delete m_pclTilesImg;
		m_pclTilesImg = new C_Image(szFileAndPath, NULL, &bResult);
		if(!bResult) {
			delete m_pclTilesImg; m_pclTilesImg = NULL;
			return false;
		}
	}

	m_pclTilesImg_Sysmem = new C_Image(m_pclTilesImg);

	//analyse tileset (get rectangles)
	int iWidth, iHeight, iNumX, iNumY, iTileNum;
	m_pclTilesImg->GetInfo(&iWidth, &iHeight, NULL);
	iNumX = iWidth / m_iTileSize;
	iNumY = iHeight / m_iTileSize;
	m_iNumTiles = iNumX * iNumY;
	m_stTileRects = new S_Rect[m_iNumTiles];

	iTileNum = 0;
	for(j=0; j<iNumY; j++) {
		for(i=0; i<iNumX; i++) {
			m_stTileRects[iTileNum].x = i * m_iTileSize;
			m_stTileRects[iTileNum].y = j * m_iTileSize;
			m_stTileRects[iTileNum].width = m_iTileSize;
			m_stTileRects[iTileNum].height = m_iTileSize;
			iTileNum++;
		}
	}

	return true;
}

//checks if the values goes outside the limits, and take actions
bool C_Map::WrapAroundTest(int *o_iX, int *o_iY, int i_iWidth, int i_iHeight)
{
	bool bResult = true;

	if(*o_iX>=i_iWidth) {
		if(m_iWrappX) *o_iX %= i_iWidth;
		else {
			*o_iX = i_iWidth;
			bResult = false;
		}
	}
	if(*o_iY>=i_iHeight) {
		if(m_iWrappY) *o_iY %= i_iHeight;
		else {
			*o_iY = i_iHeight;
			bResult = false;
		}
	}
	if(*o_iX<0) {
		if(m_iWrappX) *o_iX = i_iWidth + (*o_iX % i_iWidth);
		else {
			*o_iX = 0;
			bResult = false;
		}
	}
	if(*o_iY<0) {
		if(m_iWrappY) *o_iY = i_iHeight + (*o_iY % i_iHeight);
		else {
			*o_iY = 0;
			bResult = false;
		}
	}

	return bResult;
}

//set if the map shall stop or wrapp around when scrolling past the borders
void C_Map::SetMapWrapping(bool i_bHorizontal, bool i_bVertical)
{
	m_iWrappX = i_bHorizontal;
	m_iWrappY = i_bVertical;

	m_stDrawarea.width = m_iDrawareaWidthSave;
	m_stDrawarea.height = m_iDrawareaHeightSave;
	//if the show area is larger than the map, and wrapparound
	// shall not be used, we need to limit the showarea
	if(!m_iWrappX && m_iWidth*m_iTileSize< m_iDrawareaWidthSave) m_stDrawarea.width = m_iWidth*m_iTileSize;
	if(!m_iWrappY && m_iHeight*m_iTileSize< m_iDrawareaHeightSave) m_stDrawarea.height = m_iHeight*m_iTileSize;

	//calc maximum upper left coordinate we can scroll to
	// without getting outside the map (used if no wrapparound)
	m_iPixMaxX = m_stDrawarea.x + m_stDrawarea.width;
	m_iPixMaxY = m_stDrawarea.y + m_stDrawarea.height;
	m_iScrollMaxX = m_iWidth*m_iTileSize - m_stDrawarea.width;
	m_iScrollMaxY = m_iHeight*m_iTileSize - m_stDrawarea.height;
	if(m_iScrollMaxX < 0) m_iScrollMaxX = 0;
	if(m_iScrollMaxY < 0) m_iScrollMaxY = 0;
}

//checks if the given pos is within the map
bool C_Map::IsValidPos(int i_iMapTileX, int i_iMapTileY)
{
	return ((i_iMapTileX < m_iWidth) && (i_iMapTileY < m_iHeight) && (i_iMapTileX >= 0) && (i_iMapTileY >= 0));
}

//sets a mapelement to the given value
void C_Map::SetElementAbsolute(int i_iMapTileX, int i_iMapTileY, char i_cVal)
{
	if(!WrapAroundTest(&i_iMapTileX, &i_iMapTileY, m_iWidth, m_iHeight)) return;

	m_pGrid[i_iMapTileY][i_iMapTileX] = i_cVal;
}

//gets a mapelement from the given pos
int C_Map::GetElementAbsolute(int i_iMapTileX, int i_iMapTileY)
{
	if(!WrapAroundTest(&i_iMapTileX, &i_iMapTileY, m_iWidth, m_iHeight)) return -1;

	return m_pGrid[i_iMapTileY][i_iMapTileX];
}

//moves the showpoint of the map relative to the last
// position and draws the result
void C_Map::MoveRelative(int i_iX, int i_iY)
{
	MoveAbsolute(m_iLastX + i_iX, m_iLastY + i_iY);
}

//moves the showpoint of the map to a new position,
// given in world pixel coordinates
void C_Map::MoveAbsolute(int i_iX, int i_iY)
{
	int iCmpx = m_iWrappX ? m_iWidth*m_iTileSize  : m_iScrollMaxX;
	int iCmpy = m_iWrappY ? m_iHeight*m_iTileSize : m_iScrollMaxY;
	WrapAroundTest(&i_iX, &i_iY, iCmpx, iCmpy);

	m_iLastX = i_iX; m_iLastY = i_iY;
}

//returns the given player starting position
bool C_Map::GetPlayerStartingPos(int i_iPlayer, POINT *o_stPos)
{
	if(i_iPlayer>=0 && i_iPlayer<8 && o_stPos) {
		o_stPos->x = m_stPlayerStartPos[i_iPlayer].x;
		o_stPos->y = m_stPlayerStartPos[i_iPlayer].y;
		return true;
	}
	return false;
}

//draws all the currently visible tiles in the map on screen
void C_Map::Draw()
{
	if(!m_bMapLoaded) return;
	int iTempX, iTempY;
	int iBaseTileX, iBaseOffsetX;
	int iTileX, iTileY;
	int iOffsetX, iOffsetY;
	int iScreenX = m_stDrawarea.x, iScreenY = m_stDrawarea.y;
	S_Rect stRect;
	S_Rect *pstRect;

	iBaseTileX = iTileX = m_iLastX / m_iTileSize;
	iBaseOffsetX = iOffsetX = m_iLastX % m_iTileSize;
	iTileY = m_iLastY / m_iTileSize;
	iOffsetY = m_iLastY % m_iTileSize;
	do {
		do {
			pstRect = &m_stTileRects[(int)m_pGrid[iTileY][iTileX]];
			stRect.x = pstRect->x + iOffsetX;
			stRect.y = pstRect->y + iOffsetY;
			iTempX = m_iPixMaxX - iScreenX;
			iTempY = m_iPixMaxY - iScreenY;
			stRect.width = (m_iTileSize - iOffsetX)<iTempX ? (m_iTileSize - iOffsetX):iTempX;
			stRect.height = (m_iTileSize - iOffsetY)<iTempY ? (m_iTileSize - iOffsetY):iTempY;

			m_pclGraph->Blt(m_pclTilesImg, &stRect, iScreenX, iScreenY);

			iTileX++;
			if(m_iWrappX) iTileX %= m_iWidth;
			iScreenX += stRect.width;
			iOffsetX = 0;
		} while (iScreenX<m_iPixMaxX && iTileX<m_iWidth);
		iOffsetX = iBaseOffsetX;
		iOffsetY = 0;
		iTileX = iBaseTileX;
		iTileY++;
		if(m_iWrappY) iTileY %= m_iHeight;

		iScreenX = m_stDrawarea.x;
		iScreenY += stRect.height;
	} while (iScreenY<m_iPixMaxY && iTileY<m_iHeight);
}

//renders all tiles as a minitile to a new image
// an outline is drawn in white, space in black, and others are transparrent
C_Image *C_Map::GetMiniMap(int i_iMiniTileSize)
{
	if(!m_bMapLoaded) return NULL;
	int x, y;
	int iX = 0, iY = 0;
	int iMiniTileSize = i_iMiniTileSize; //number of pixels per tile

	S_Rect stRect = {iX,iY,iMiniTileSize,iMiniTileSize};

	C_Image *pclImage = new C_Image(m_iWidth*iMiniTileSize, m_iHeight*iMiniTileSize, C_Image::GetDefaultFormat());
	S_Color stColor  = {1,1,1,255};
	S_Color stColor2 = {165,165,165,255};
	pclImage->FillWithAbsoluteColor(NULL, 0);

	for(y=0; y<m_iHeight; y++) {
		for(x=0; x<m_iWidth; x++) {
			int iTile = 0;
			iTile = m_pGrid[y][x];
			if(iTile!=0) {
				iTile = 2; //assume. a solid tile on all current tilesets (2=transparrent)
				//if any neighbor is 0, this tile is a border (1)
				int k, l;
				for(k=x-1; k<=x+1; k++) {
					for(l=y-1; l<=y+1; l++) {
						if(k<m_iWidth && l<m_iHeight && k>=0 && l>=0 && m_pGrid[l][k]==0) {
							iTile = 1;
							break;
						}
					}
					if(iTile==1) break;
				}
			}

			stRect.x = iX; stRect.y = iY;
			if(iTile==0) {
				pclImage->FillWithRGBColor(&stRect, &stColor);
			} else if(iTile==1) {
				pclImage->FillWithRGBColor(&stRect, &stColor2);
			}

			iX += iMiniTileSize;
		}
		iX = 0;
		iY += iMiniTileSize;
	}

	return pclImage;
}

//tests if an object has collided with nonblack pixels in the map
bool C_Map::CheckCollision(C_Object *object)
{
	if(!m_bMapLoaded) return false;
	bool bResult = false;
	int iX, iY, iW, iH, iPixelFmt;
	object->GetPos(&iX, &iY);
	object->GetSize(&iW, &iH);
	m_pclTilesImg->GetInfo(NULL, NULL, &iPixelFmt);

	//check boundaries
	if(iX<0 || iY<0 || iX+iW>m_iWidth*m_iTileSize || iY+iH>m_iHeight*m_iTileSize) bResult = true;

	if(!bResult) { //check pixels

		C_Image *pclMapImg = new C_Image(iW, iH, iPixelFmt);
		pclMapImg->FillWithAbsoluteColor(NULL, 0);

		int iTileNum;
		int iTempX, iTempY;
		int iBaseTileX, iBaseOffsetX;
		int iTileX, iTileY;
		int iOffsetX, iOffsetY;
		int iScreenX = 0, iScreenY = 0;
		S_Rect stRect;
		S_Rect *pstRect;
		bool bAllTilesZero = true;

		//make map piece
		iBaseTileX = iTileX = iX / m_iTileSize;
		iBaseOffsetX = iOffsetX = iX % m_iTileSize;
		iTileY = iY / m_iTileSize;
		iOffsetY = iY % m_iTileSize;

		do {
			do {
				iTileNum = m_pGrid[iTileY][iTileX];
				if(iTileNum) bAllTilesZero = false;
				pstRect = &m_stTileRects[iTileNum];
				stRect.x = pstRect->x + iOffsetX;
				stRect.y = pstRect->y + iOffsetY;
				iTempX = iW - iScreenX;
				iTempY = iH - iScreenY;
				stRect.width = (m_iTileSize - iOffsetX)<iTempX ? (m_iTileSize - iOffsetX):iTempX;
				stRect.height = (m_iTileSize - iOffsetY)<iTempY ? (m_iTileSize - iOffsetY):iTempY;

				if(iTileNum) pclMapImg->Blt2(m_pclTilesImg_Sysmem, &stRect, iScreenX, iScreenY);

				iTileX++;
				if(m_iWrappX) iTileX %= m_iWidth;
				iScreenX += stRect.width;
				iOffsetX = 0;
			} while (iScreenX < iW);
			iOffsetX = iBaseOffsetX;
			iOffsetY = 0;
			iTileX = iBaseTileX;
			iTileY++;
			if(m_iWrappY) iTileY %= m_iHeight;

			iScreenX = 0;
			iScreenY += stRect.height;
		} while (iScreenY < iH);

		//compare, or skip if all tiles are zero = no collission can happen
		bResult = bAllTilesZero ? false : object->OCheckCollision(pclMapImg, iX, iY);

		delete pclMapImg;
	}

	return bResult;
}

//draw to an image
void C_Map::DrawOn(C_Image* i_pclDst, const S_Rect* i_pstRange)
{
	if(!i_pstRange || !i_pclDst) {
		return;
	}

	int iStartTileX = i_pstRange->x;
	int iStartTileY = i_pstRange->y;

	int iEndTileX = i_pstRange->x + i_pstRange->width - 1;
	int iEndTileY = i_pstRange->y + i_pstRange->height - 1;

	if((iStartTileX < 0 || iStartTileY < 0) ||
		(iStartTileX >= m_iWidth || iStartTileY >= m_iHeight) ||
		(iEndTileX < 0 || iEndTileY < 0) ||
		(iEndTileX >= m_iWidth || iEndTileY >= m_iHeight))
	{
		return;
	}

	for(int i = iStartTileX; i <= iEndTileX; ++i) {
		for(int j = iStartTileY; j <= iEndTileY; ++j) {
			int tileNum = m_pGrid[j][i];
			S_Rect* pRect = &m_stTileRects[tileNum];
			i_pclDst->Blt2(m_pclTilesImg_Sysmem, pRect, i * m_iTileSize, j * m_iTileSize);
		}
	}
}

//used in new AI (private used by Partition())
void C_Map::ProcessTile(C_Image* tileImg, int tileNumber, int tileSize,
	int xStart, int yStart, int nParts, vector<vector<bool> >& subtiles)
{
	int truncatedInterval = tileSize / nParts;
	int numExtendedIntervals = tileSize % nParts;

	IntVector thresholds;
	thresholds.push_back(0);

	vector<bool> thisTileTessalation;

	for (int k = 0; k < nParts; ++k)
	{
		int nextInterval = thresholds[k] + truncatedInterval;
		if (numExtendedIntervals > 0)
		{
			nextInterval++;
			--numExtendedIntervals;
		}

		thresholds.push_back(nextInterval);
	}

	for (int i = 0; i < nParts; ++i)
	{
		for (int j = 0; j < nParts; ++j)
		{
			bool isWall = false;

			for (int y = yStart + thresholds[i]; !isWall && y < yStart + thresholds[i+1]; ++y)
			{
				for (int x = xStart + thresholds[j]; !isWall && x < xStart + thresholds[j+1]; ++x)
				{
					S_Color color = tileImg->GetPixelRGB(x, y);

					if (color.b != 0 || color.g != 0 || color.r != 0)
					{
						isWall = true;
					}
				}
			}

#ifdef MARK_TILES
			if (isWall)
			{
				S_Rect subtileArea = {xStart + thresholds[j], yStart + thresholds[i], thresholds[j+1] - thresholds[j], thresholds[i+1] - thresholds[i]};
				S_Color whiteColor = { 0xff,0xff,0xff, 0xcc};	
				tileImg->FillWithRGBColor(&subtileArea, &whiteColor);
			}
#endif
			thisTileTessalation.push_back(!isWall);
		}
	}

	subtiles.push_back(thisTileTessalation);
}

//used in new AI
void C_Map::Partition(int numPartsPerAxis, vector<vector<bool> >& subtiles)
{
	C_Image* tileImg = m_pclTilesImg;

#ifdef MARK_TILES
	C_Image* tileCopy = new C_Image(tileImg);
#endif

	subtiles.clear();

	if (tileImg)
	{
		int w, h;
		tileImg->GetInfo(&w, &h, NULL);

		int tileSize = GetTileSize();
		w /= tileSize;
		h /= tileSize;

		for (int y = 0; y < h; ++y)
		{
			for (int x = 0; x < w; ++x)
			{
				ProcessTile(tileImg, y*w + x, tileSize, x*tileSize, y * tileSize, numPartsPerAxis, subtiles);
			}
		}

#ifdef MARK_TILES
		for (int y = 0; y < h * tileSize; ++y)
			for (int x = 0; x < w * tileSize; ++x)
			{
				S_Color color = tileCopy->GetPixelRGB(x, y);

				if (color.b != 0 || color.g != 0 || color.r != 0)
				{
					S_Rect area = {x, y, 1, 1};
					tileImg->FillWithRGBColor(&area, &color);
				}
			}

		delete tileCopy;
#endif
	}
}
