// C_Game

//class that keeps info for one game on the server
// and has functions for handling a game
// - keeps all players in a game
// - handle cargo requests
// - handle cargo respawn
// - handle extra lifes
// - check for game over
// - functions for sending/resending messages to all players in a game

#include "s_game.h"

C_Game::C_Game(int i_iGameIndex, char *i_szGameName, C_Player *i_pclPlayer, C_Logger *i_pclLog)
{
	m_iId = i_iGameIndex;

	m_pclPlayerList = new C_PlayerList();
	m_pclTimer = new C_Timer();

	//gameover info
	m_iNumLaps   = 0;
	m_iLapsCount = -1;
	m_iKillCount = -1;
	m_iLevelType = MAP_MISSION;
	m_iNumAI     = 0;

	m_bNoShooting  = false;
	m_bFlyThrough  = false;
	m_bNoBulletHit = false;
	m_bOptionsSet  = false;
	m_bPlaying     = false;
	m_bGameStatus  = false;
	m_bGameStatusChanged = false;
	m_szLevelName[0] = 0;

	strcpy(m_szName, i_szGameName);
	m_pclPlayerList->push_back(i_pclPlayer);

	//log
	m_pclLog = i_pclLog;
	m_pclLog->Print(LOG_INFO, "Game[%d] '%s' created by %s (%s)\n",
		m_iId, m_szName, i_pclPlayer->GetName(), i_pclPlayer->GetIPString());
}

C_Game::~C_Game()
{
	int i, iSize = (int)m_pclPlayerList->size();
	for(i=0; i<iSize; i++) {
		delete (*m_pclPlayerList)[i];
	}
	delete m_pclPlayerList;

	delete m_pclTimer;

	//log
	m_pclLog->Print(LOG_INFO, "Game[%d] '%s' destroyed\n", m_iId, m_szName);
}

bool C_Game::CompareGameName(char *i_szName)
{
	return (strcmp(m_szName, i_szName) == 0);
}

char* C_Game::GetGameName()
{
	return m_szName;
}

bool C_Game::GetStatus()
{
	return m_bGameStatus;
}

bool C_Game::StatusChanged()
{
	bool bResult = m_bGameStatusChanged;
	m_bGameStatusChanged = false;
	return bResult;
}

int C_Game::GetNumPlayers()
{
	return (int)m_pclPlayerList->size();
}

C_Player *C_Game::GetPlayer(int i_iIndex)
{
	int i, iSize = (int)m_pclPlayerList->size();
	C_Player *pclPlayer;

	for(i=0; i<iSize; i++) {
		pclPlayer = (*m_pclPlayerList)[i];
		if(pclPlayer->GetIndex() == i_iIndex) return pclPlayer;
	}
	return NULL;
}

int C_Game::CheckPreState()
{
	if(!m_bPlaying) {
		for(int i=0; i<8; i++) {
			if(!GetPlayer(i)) return i; //first free index
		}
	}
	return -1;
}

//new player enters existing game
void C_Game::JoinToGame(C_Player *i_pclPlayer)
{
	int i, iSize;
	C_NetMsgServer *pclNetMsg; 
	S_NetMsg2 stMsg2;
	memset(&stMsg2, 0, sizeof(stMsg2));

	//log
	m_pclLog->Print(LOG_INFO, "Game[%d] '%s', player %s (%s) added\n",
		m_iId, m_szName, i_pclPlayer->GetName(), i_pclPlayer->GetIPString());

	//to new player (join ok)
	pclNetMsg = i_pclPlayer->GetNetMsg();  
	stMsg2.iType = N_MSG2_JOIN_NEW_R;
	stMsg2.iPlayerId = i_pclPlayer->GetIndex();
	stMsg2.iGameId = m_iId;
	pclNetMsg->SendMsg2(&stMsg2);

	//send info about all other players to the new player
	stMsg2.iType = N_MSG2_ADDPLAYER;
	iSize = (int)m_pclPlayerList->size();
	for(i=0; i<iSize; i++) {
		C_Player *pclSrcPlayer = (*m_pclPlayerList)[i];
		stMsg2.iPlayerId = pclSrcPlayer->GetIndex();
		strcpy(stMsg2.szPlayerName, pclSrcPlayer->GetName());
		pclNetMsg->SendMsg2(&stMsg2);
	}

	//send the current game options to the new player
	SendGameOptions(i_pclPlayer);

	//add new player in the list
	m_pclPlayerList->push_back(i_pclPlayer);
	iSize++;

	//send to all players that a new player has arrived
	//stMsg2.iMsg = N_MSG2_ADDPLAYER;
	stMsg2.iPlayerId = i_pclPlayer->GetIndex();
	strcpy(stMsg2.szPlayerName, i_pclPlayer->GetName());
	for(i=0; i<iSize; i++) {
		C_Player *pclDestPlayer = (*m_pclPlayerList)[i];
		pclNetMsg = pclDestPlayer->GetNetMsg(); //to var
		pclNetMsg->SendMsg2(&stMsg2);
	}

	ChangeGameStatus();
}

void C_Game::HandleCargoRequest(S_NetMsg2 *i_pstMsg, int i_iPlayerIndex, C_Player *i_pclSender)
{
	int i, iSize;
	C_NetMsgServer *pclNetMsg;
	C_Player *pclSndPlayer;
	bool bAlreadyTaken = false;
	bool bAddToList = true;

	//i_pstMsg->iZoneId, i_pstMsg->iCargoNum
	iSize = (int)m_clCargo.size();
	for(i=0; i<iSize; i++) {
		if(m_clCargo[i].iZoneId == i_pstMsg->iZoneId) {
			bAddToList = false;
			if(m_clCargo[i].iCargoNum <= i_pstMsg->iCargoNum) {
				bAlreadyTaken = true;
			} else {
				m_clCargo[i].iCargoNum = i_pstMsg->iCargoNum;
			}
			break;
		}
	}
	if(bAddToList) {
		S_ZoneCargo stCargo = {i_pstMsg->iZoneId, i_pstMsg->iCargoNum, false};
		m_clCargo.push_back(stCargo);
	}

	i_pstMsg->iType = N_MSG2_CARGO_R;

	if(!bAlreadyTaken) {
		iSize = (int)m_pclPlayerList->size();
		for(i=0; i<iSize; i++) {
			pclSndPlayer = (*m_pclPlayerList)[i];
			pclNetMsg = pclSndPlayer->GetNetMsg();
			i_pstMsg->iCargoReqAccepted = 0;
			if(pclSndPlayer==i_pclSender) {
				if(pclSndPlayer->GetIndex()== i_iPlayerIndex) {
					i_pstMsg->iCargoReqAccepted = 1;
				}
			}
			pclNetMsg->SendMsg2(i_pstMsg);
		}
	} //else do nothing
}

void C_Game::HandleCargoRespawn(S_NetMsg2 *i_pstMsg, int i_iPlayerIndex)
{
	int i, iSize;
	C_NetMsgServer *pclNetMsg;
	C_Player *pclSndPlayer;

	//i_pstMsg->iZoneId, i_pstMsg->iCargoNum
	iSize = (int)m_clCargo.size();
	for(i=0; i<iSize; i++) { //since the cargo has been taken before, we will find it in this list
		if(m_clCargo[i].iZoneId == i_pstMsg->iZoneId) {
			m_clCargo[i].iCargoNum++;
		}
	}

	//resend msg to all players
	iSize = (int)m_pclPlayerList->size();
	for(i=0; i<iSize; i++) {
		pclSndPlayer = (*m_pclPlayerList)[i];
		pclNetMsg = pclSndPlayer->GetNetMsg();
		pclNetMsg->SendMsg2(i_pstMsg);
	}
}

void C_Game::HandleLifeRequest(S_NetMsg2 *i_pstMsg, int i_iPlayerIndex, C_Player *i_pclSender)
{
	int i, iSize;
	C_NetMsgServer *pclNetMsg;
	C_Player *pclSndPlayer;
	bool bAlreadyTaken = false;
	bool bAddToList = true;

	//i_pstMsg->iZoneId
	iSize = (int)m_clCargo.size();
	for(i=0; i<iSize; i++) {
		if(m_clCargo[i].iZoneId == i_pstMsg->iZoneId) {
			bAddToList = false;
			if(!m_clCargo[i].bLifeTaken) {
				m_clCargo[i].bLifeTaken = true;
			} else bAlreadyTaken = true;
			break;
		}
	}
	if(bAddToList) {
		S_ZoneCargo stCargo = {i_pstMsg->iZoneId, 1000, true};
		m_clCargo.push_back(stCargo);
	}

	i_pstMsg->iType = N_MSG2_LIFE_R;

	if(!bAlreadyTaken) {
		iSize = (int)m_pclPlayerList->size();
		for(i=0; i<iSize; i++) {
			pclSndPlayer = (*m_pclPlayerList)[i];
			pclNetMsg = pclSndPlayer->GetNetMsg();
			i_pstMsg->iLifeReqAccepted = 0;
			if(pclSndPlayer==i_pclSender) {
				if(pclSndPlayer->GetIndex()== i_iPlayerIndex) {
					i_pstMsg->iLifeReqAccepted = 1;
				}
			}
			pclNetMsg->SendMsg2(i_pstMsg);
		}
	} //else do nothing
}

void C_Game::HandleNextLevel()
{
	//previously missions were played in order,
	// but now levels are played one at a time

	//old code:
	/*m_iMissionIndex++;
	if(m_iMissionIndex<NUM_MISSIONS) {
		//send nextlevel msg to all in game
		S_NetMsg2 stMsg2;
		stMsg2.iType = N_MSG2_NEXTLEVEL;
		SendMsg(&stMsg2);
		//reset leveldata
		m_clCargo.clear();
	} else*/
	CheckGameOver(true); //force gameover
}

void C_Game::SendMsg(S_NetMsg2 *i_pstMsg)
{
	int i, iSize;
	C_NetMsgServer *pclNetMsg;
	C_Player *pclSndPlayer;
	iSize = (int)m_pclPlayerList->size();

	SaveImportantInfo(i_pstMsg);
	//resend msg to others
	for(i=0; i<iSize; i++) {
		pclSndPlayer = (*m_pclPlayerList)[i];
		pclNetMsg = pclSndPlayer->GetNetMsg();
		pclNetMsg->SendMsg2(i_pstMsg);
	}
}

void C_Game::SendMsg(S_NetMsg2 *i_pstMsg, C_Player *i_pclPlayer)
{
	int i, iSize;
	C_NetMsgServer *pclNetMsg;
	C_Player *pclSndPlayer;
	iSize = (int)m_pclPlayerList->size();

	SaveImportantInfo(i_pstMsg);
	//resend msg to others
	for(i=0; i<iSize; i++) {
		pclSndPlayer = (*m_pclPlayerList)[i];
		if(pclSndPlayer != i_pclPlayer) { //do not send to that player
			pclNetMsg = pclSndPlayer->GetNetMsg();
			pclNetMsg->SendMsg2(i_pstMsg);
		}
	}
}

void C_Game::SendMsg1(S_NetMsg1 *i_pstMsg, C_Player *i_pclPlayer)
{
	int i, iSize;
	C_NetMsgServer *pclNetMsg;
	C_Player *pclSndPlayer;
	iSize = (int)m_pclPlayerList->size();

	//resend msg to others
	for(i=0; i<iSize; i++) {
		pclSndPlayer = (*m_pclPlayerList)[i];
		if(pclSndPlayer != i_pclPlayer) { //do not send to that player
			pclNetMsg = pclSndPlayer->GetNetMsg();
			pclNetMsg->SendMsg1(i_pstMsg);
		}
	}
}

void C_Game::SendChat(char *i_szChatString)
{
	int i, iSize;
	C_NetMsgServer *pclNetMsg;
	C_Player *pclSndPlayer;
	iSize = (int)m_pclPlayerList->size();

	//resend msg to others
	for(i=0; i<iSize; i++) {
		pclSndPlayer = (*m_pclPlayerList)[i];
		pclNetMsg = pclSndPlayer->GetNetMsg();
		pclNetMsg->SendChatMsg(i_szChatString);
	}
}

void C_Game::SaveImportantInfo(S_NetMsg2 *i_pstMsg)
{
	if(i_pstMsg->iType == N_MSG2_OPTIONS) {
		m_iNumLaps     = i_pstMsg->iNumLaps;
		m_iLevelType   = i_pstMsg->iLevelType;
		m_bNoShooting  = (i_pstMsg->iOptions & OPT_NOSHOOTING)!=0;
		m_bFlyThrough  = (i_pstMsg->iOptions & OPT_FLYTHROUGH)!=0;
		m_bNoBulletHit = (i_pstMsg->iOptions & OPT_NOBULLETHIT)!=0;
		m_iNumAI       = i_pstMsg->iNumAI;
		strcpy(m_szLevelName, i_pstMsg->szLevelName);
		m_bOptionsSet = true;
	}
}

void C_Game::SendGameOptions(C_Player *i_pclPlayer)
{
	S_NetMsg2 i_stMsg2;

	if(m_bOptionsSet) {
		i_stMsg2.iType      = N_MSG2_OPTIONS;
		i_stMsg2.iNumLaps   = m_iNumLaps;
		i_stMsg2.iLevelType = m_iLevelType;
		i_stMsg2.iOptions   = 0;
		if(m_bNoShooting)  i_stMsg2.iOptions |= OPT_NOSHOOTING;
		if(m_bFlyThrough)  i_stMsg2.iOptions |= OPT_FLYTHROUGH;
		if(m_bNoBulletHit) i_stMsg2.iOptions |= OPT_NOBULLETHIT;
		i_stMsg2.iNumAI     = m_iNumAI;
		strcpy(i_stMsg2.szLevelName, m_szLevelName);
		(i_pclPlayer->GetNetMsg())->SendMsg2(&i_stMsg2);
	}
}

void C_Game::SendStart()
{
	int i, iSize = (int)m_pclPlayerList->size();
	struct S_NetMsg2 stMsg2;
	stMsg2.iType = N_MSG2_START_R;
	SendMsg(&stMsg2);

	m_bPlaying = true;
	ChangeGameStatus();

	//reset per level states
	m_iKillCount = -1;
	m_iLapsCount = -1;
	for(i=0; i<8; i++) {
		m_iKillCountAI[i] = 0;
		m_iLapsCountAI[i] = 0;
		m_aiFreeIndexForAI[i] = true;
		m_iLifesAI[i] = -1;
		if(m_iLevelType!=MAP_RACE) m_iLifesAI[i] = NUM_LIFES;
	}
	for(i=0; i<iSize; i++) {
		C_Player *pclPlayer = (*m_pclPlayerList)[i];
		pclPlayer->m_iNumLaps = 0;
		pclPlayer->m_iNumKills = 0;
		pclPlayer->m_iNumLifes = -1;
		m_aiFreeIndexForAI[pclPlayer->GetIndex()] = false;
		if(m_iLevelType!=MAP_RACE) pclPlayer->m_iNumLifes = NUM_LIFES;
	}
	//reset cargo
	m_clCargo.clear();

	//make the server send the initial value to players
	if(m_iLevelType!=MAP_RACE) AddKill(-1, 0);
	else AddLap(-1);

	//log
	m_pclLog->Print(LOG_INFO, "Game[%d] '%s' started, level %s, with %d players\n",
		m_iId, m_szName, m_szLevelName, iSize);

	//start timer
	m_pclTimer->Reset();
}

bool C_Game::CheckGameOver(bool i_bForceGameOver)
{
	bool bResult = i_bForceGameOver;
	if(m_iNumLaps) {
		if(m_iLapsCount >= m_iNumLaps) bResult = true;
	}
	if(m_iKillCount==0) bResult = true;

	if(bResult) {
		m_bPlaying = false;
		ChangeGameStatus();

		//send gameover msg to all in game
		S_NetMsg2 stMsg2;
		stMsg2.iPlayerId = -1;
		stMsg2.iType = N_MSG2_GAMEOVER;
		SendMsg(&stMsg2);

		//log
		m_pclLog->Print(LOG_INFO, "Game[%d] '%s' is over, level %s\n",
			m_iId, m_szName, m_szLevelName);
	}
	return bResult;
}

void C_Game::RemovePlayer(C_Player *i_pclPlayer)
{
	int i, iSize = (int)m_pclPlayerList->size();

	//log
	m_pclLog->Print(LOG_INFO, "Game[%d] '%s', player %s (%s) left\n",
		m_iId, m_szName, i_pclPlayer->GetName(), i_pclPlayer->GetIPString());

	for(i=0; i<iSize; i++) {
		C_Player *pclRemovePlayer = (*m_pclPlayerList)[i];
		if(pclRemovePlayer == i_pclPlayer) { //remove player
			bool bNewMaster = pclRemovePlayer->m_bIsMaster;
			m_pclPlayerList->erase(m_pclPlayerList->begin()+i);
			if(bNewMaster) {
				//transfer mastership to next player
				iSize--;
				for(i=0; i<iSize; i++) {
					C_Player *pclNewMaster = (*m_pclPlayerList)[i];
					if(pclNewMaster != NULL) {
						pclNewMaster->m_bIsMaster = true;
						//inform the new master
						S_NetMsg2 i_stMsg2;
						i_stMsg2.iType = N_MSG2_NEWMASTER;
						(pclNewMaster->GetNetMsg())->SendMsg2(&i_stMsg2);

						m_pclLog->Print(LOG_INFO, "Game[%d] '%s', new master is %s (%s)\n",
							m_iId, m_szName, pclNewMaster->GetName(), pclNewMaster->GetIPString());

						break;
					}
				}
			}
			break;
		}
	}

	//make the server send the current value to players
	if(m_bPlaying) {
		if(m_iLevelType!=MAP_RACE) AddKill(-1, 0);
		else AddLap(-1);
	}

	ChangeGameStatus();
}

void C_Game::ChangeGameStatus()
{
	bool bNewGameStatus = false;

	if((m_pclPlayerList->size() == 8) || m_bPlaying) bNewGameStatus = true;

	if(m_bGameStatus != bNewGameStatus) {
		m_bGameStatus = bNewGameStatus;
		m_bGameStatusChanged = true;
	}
}

void C_Game::AddKill(int i_iPlayerIndex, int i_iKilledBy)
{
	int i;
	C_Player *pclPlayer = NULL;

	i_iKilledBy &= 255; //only other players id
	if(i_iPlayerIndex>=0) {
		C_Player *pclKilledByPlayer = GetPlayer(i_iKilledBy);
		pclPlayer = GetPlayer(i_iPlayerIndex);
		if(pclKilledByPlayer) {
			pclKilledByPlayer->m_iNumKills++;
		} else if(i_iKilledBy<8) {
			m_iKillCountAI[i_iKilledBy]++;
		}
		if(pclPlayer) {
			if(pclPlayer->m_iNumLifes>0) pclPlayer->m_iNumLifes--;
		} else {
			if(m_iLifesAI[i_iPlayerIndex]>0) m_iLifesAI[i_iPlayerIndex]--;
		}
	}

	//check if current numkills should change
	int iNewKillCount = 0;
	int iNumLifes, iMostLives = 0;
	int iNumAI = m_iNumAI;
	for(i=0; i<8; i++) {
		iNumLifes = -1;
		pclPlayer = GetPlayer(i);
		if(pclPlayer) iNumLifes = pclPlayer->m_iNumLifes;
		else if(iNumAI>0 && m_aiFreeIndexForAI[i]) {
			iNumAI--;
			iNumLifes = m_iLifesAI[i];
		}
		if(iNumLifes>=0) iNewKillCount += iNumLifes;
		if(iNumLifes>iMostLives) {
			iMostLives = iNumLifes;
		}
	}
	if(m_iLevelType==MAP_DOGFIGHT) iNewKillCount -= iMostLives; //only dogfight

	if(iNewKillCount!=m_iKillCount) {
		m_iKillCount = iNewKillCount;

		//send change to all players in game
		S_NetMsg2 stMsg2;
		stMsg2.iType = N_MSG2_NUMKILLS;
		stMsg2.iPlayerId = -1;
		stMsg2.iNumKills = m_iKillCount;
		SendMsg(&stMsg2);
	}
}

void C_Game::AddLap(int i_iPlayerIndex)
{
	int i;
	C_Player *pclPlayer = NULL;
	if(i_iPlayerIndex>=0) {
		pclPlayer = GetPlayer(i_iPlayerIndex);
		if(pclPlayer) pclPlayer->m_iNumLaps++;
		else m_iLapsCountAI[i_iPlayerIndex]++;
	}

	//check if numlaps should change
	int iLaps, iLessLaps = 10000;
	int iNumAI = m_iNumAI;
	for(i=0; i<8; i++) {
		iLaps = 10000;
		pclPlayer = GetPlayer(i);
		if(pclPlayer) iLaps = pclPlayer->m_iNumLaps;
		else if(iNumAI>0 && m_aiFreeIndexForAI[i]) {
			iNumAI--;
			iLaps = m_iLapsCountAI[i];
		}
		if(iLaps < iLessLaps) iLessLaps = iLaps;
	}
	
	if(iLessLaps!=m_iLapsCount) {
		m_iLapsCount = iLessLaps;

		//send change to all players in game
		S_NetMsg2 stMsg2;
		stMsg2.iType = N_MSG2_NUMKILLS;
		stMsg2.iPlayerId = -1;
		stMsg2.iNumLaps = m_iNumLaps - m_iLapsCount;
		SendMsg(&stMsg2);
	}
}
