// C_Manager

//class that keeps info for all games and connected players on the server.
// all of the server logic is here

#include <string.h>
#include "s_manager.h"
#include "global.h"

C_Manager::C_Manager()
{
	m_pclPlayerList = new C_PlayerList();
	m_pclGameList   = new C_GameList();

	m_iGameIdCounter = 0;

	//log
	char szLogFile[MAX_PATH];
	sprintf(szLogFile, "%s/server_log.txt", C_Global::szUserPath);
	m_pclLog = new C_Logger();
	m_pclLog->SetLogLevel(LOG_ALL);
	m_pclLog->AssignFile(szLogFile);
#ifdef WIN32
	m_pclLog->SetOutLevel(0);
	//m_pclLog->SetOutLevel(LOG_TO_FILE);
#else
	m_pclLog->SetOutLevel(LOG_TO_STDOUT);
#endif
	m_pclLog->Print(LOG_INFO, "Server started\n");
}

C_Manager::~C_Manager()
{
	C_Player *pclPlayer;

	//free player list
	int i, j, iSize = (int)m_pclPlayerList->size();
	for(i=0; i<iSize; i++) {
		pclPlayer = (*m_pclPlayerList)[i];
		delete pclPlayer->GetSocket(); //socket is owned by manager
		delete pclPlayer;
	}
	delete m_pclPlayerList;

	//free game list
	iSize = (int)m_pclGameList->size();
	for(i=0; i<iSize; i++) {
		//socket is owned by manager
		C_Game *pclGame = (*m_pclGameList)[i];
		for(j=0; j<8; j++) {
			pclPlayer = pclGame->GetPlayer(j);
			if(pclPlayer) delete pclPlayer->GetSocket();
		}
		delete pclGame;
	}
	delete m_pclGameList;

	m_pclLog->Print(LOG_INFO, "Server closed\n");
	delete m_pclLog;
}


void C_Manager::Update()
{
	int i, j, iSize = (int)m_pclGameList->size();
	C_Player *pclPlayer;
	C_Socket *pclSocket;
	S_NetMsg2 stMsg2;

	//update games and check for dead sockets
	for(i=0; i<iSize; i++) {
		C_Game *pclGame = (*m_pclGameList)[i];
		if(pclGame->GetPlaying())
			if(pclGame->CheckGameOver())
				if(pclGame->StatusChanged())
					SendGameStatus(pclGame);

		//dead sockets?
		for(j=0; j<8; j++) {
			pclPlayer = pclGame->GetPlayer(j);
			if(pclPlayer) {
				pclSocket = pclPlayer->GetSocket();
				if(!pclSocket->GetSocketActive()) { //remove player?
					pclGame->RemovePlayer(pclPlayer);
					stMsg2.iType = N_MSG2_DISCONNECT;
					stMsg2.iPlayerId = j;
					pclGame->SendMsg(&stMsg2);
					delete pclPlayer;
					delete pclSocket; //socket is owned by manager
					if(pclGame->GetNumPlayers() == 0) SendRemoveGame(pclGame); //no players in game
					else if(pclGame->StatusChanged()) SendGameStatus(pclGame); //game not full
					break; //to prevent from keep checking empty game (NULL), if game is removed
				}
			}
		}
	}

	//check for dead sockets in player list
	iSize = (int)m_pclPlayerList->size();
	for(i=0; i<iSize; i++) {
		pclPlayer = (*m_pclPlayerList)[i];
		pclSocket = pclPlayer->GetSocket();
		if(!pclSocket->GetSocketActive()) { //remove player?
			//log
			char szText[256];
			char *szName = pclPlayer->GetName();
			if(!szName || szName[0]==0) sprintf(szText, "Unnamed player (%s) left\n", pclPlayer->GetIPString());
			else sprintf(szText, "Player %s (%s) left\n", szName, pclPlayer->GetIPString());
			m_pclLog->Print(LOG_INFO, szText);

			m_pclPlayerList->erase(m_pclPlayerList->begin()+i);
			delete pclPlayer;
			delete pclSocket; //socket is owned by manager
			break;
		}
	}
}

void C_Manager::ProcessNetMsg()
{
	int i, j, iResult;
	S_NetMsg2 stMsg2;
	S_NetMsg1 stMsg1;
	char      szChatString[4096]; //enough
	C_Player *pclPlayer;
	C_NetMsgServer *pclNetMsg;

	//handle players not in a game
	int iSize = (int)m_pclPlayerList->size();
	for(i=0; i<iSize; i++) {
		pclPlayer = (*m_pclPlayerList)[i];
		pclNetMsg = pclPlayer->GetNetMsg();
		iResult = pclNetMsg->GetMsg(NULL, &stMsg2, NULL, szChatString, sizeof(szChatString));
		if(iResult == 1) {
			switch(stMsg2.iType) {
				case N_MSG2_JOIN:
					pclPlayer->SetName(stMsg2.szPlayerName);
					AddPlayerToGame(pclPlayer, stMsg2.szGameName);
					break;
				case N_MSG2_NEW:
					pclPlayer->SetName(stMsg2.szPlayerName);
					CreateNewGame(pclPlayer, stMsg2.szGameName);
					break;
				case N_MSG2_DISCONNECT:
					delete pclPlayer;
					m_pclPlayerList->erase(m_pclPlayerList->begin()+i);
					break;
			}
			i = 0; iSize = (int)m_pclPlayerList->size(); //message received! restart msg check
		}
	}

	//handle players in a game
	iSize = (int)m_pclGameList->size();
	for(i=0; i<iSize; i++) {
		C_Game *pclGame = (*m_pclGameList)[i];

		for(j=0; j<8; j++) {
			pclPlayer = pclGame->GetPlayer(j);
			if(pclPlayer) {
				pclNetMsg = pclPlayer->GetNetMsg();
				iResult = pclNetMsg->GetMsg(&stMsg1, &stMsg2, NULL, szChatString, sizeof(szChatString));

				if(iResult == 1) {
					switch(stMsg2.iType) {
						//gamesetup msgs
						case N_MSG2_DISCONNECT:
							{
								C_Player *pclToBeDisconnected = pclGame->GetPlayer(stMsg2.iPlayerId); //can be other than sender if master kicks someone
								if(!pclToBeDisconnected) break; //should not happen
								if(pclPlayer == pclToBeDisconnected) { //someone chooses to disconnect
									pclGame->RemovePlayer(pclToBeDisconnected);
									pclGame->SendMsg(&stMsg2);
								} else { //player kicked by master
									pclGame->SendMsg(&stMsg2); //also send to the one disconnected
									pclGame->RemovePlayer(pclToBeDisconnected);
								}
								delete pclToBeDisconnected;
								if(pclGame->GetNumPlayers() == 0) SendRemoveGame(pclGame); //no players in game
								else if(pclGame->StatusChanged()) SendGameStatus(pclGame); //game not full
							}
							break;
						case N_MSG2_LEAVE:
							pclGame->RemovePlayer(pclPlayer);
							stMsg2.iType = N_MSG2_DISCONNECT;
							stMsg2.iPlayerId = pclPlayer->GetIndex();
							pclGame->SendMsg(&stMsg2);
							if(pclGame->GetNumPlayers() == 0) SendRemoveGame(pclGame); //no players in game
							else if(pclGame->StatusChanged()) SendGameStatus(pclGame); //game not full
							NewPlayer(pclPlayer->GetSocket());
							delete pclPlayer;
							break;
						case N_MSG2_OPTIONS:
							iResult = strcmp(pclGame->GetLevelName(), stMsg2.szLevelName);
							pclGame->SendMsg(&stMsg2, pclPlayer);
							if(iResult != 0) SendHiScore(pclGame, NULL); //send hiscorelist if another level
							break;
						case N_MSG2_START:
							pclGame->SendStart();
							if(pclGame->StatusChanged()) SendGameStatus(pclGame);
							break;
						case N_MSG2_SCORE:
							if(pclPlayer->GetIndex()==stMsg2.iPlayerId) HandleScore(pclGame, pclPlayer, stMsg2.iScore);
							pclGame->SendMsg(&stMsg2); //to all
							break;
						//ingame msg
						//id can be other than sender if master for AI
						case N_MSG2_NUMKILLS:
							pclGame->AddKill(stMsg2.iPlayerId, stMsg2.iKilledBy);
							break;
						case N_MSG2_NUMLAPS:
							pclGame->AddLap(stMsg2.iPlayerId);
							break;
						case N_MSG2_CARGO:
							pclGame->HandleCargoRequest(&stMsg2, stMsg2.iPlayerId, pclPlayer);
							break;
						case N_MSG2_CARGORESPAWN:
							pclGame->HandleCargoRespawn(&stMsg2, stMsg2.iPlayerId);
							break;
						case N_MSG2_LIFE:
							pclGame->HandleLifeRequest(&stMsg2, stMsg2.iPlayerId, pclPlayer);
							break;
						case N_MSG2_NEXTLEVEL:
							pclGame->HandleNextLevel();
							break;
					}
					iSize = (int)m_pclGameList->size();
					i = -1;
					break;
				} else if(iResult==3) { //chat
					pclGame->SendChat(szChatString);
				} else if(iResult==4) { //player->players msg
					pclGame->SendMsg1(&stMsg1, pclPlayer);
				}
			}
		}
	}
}

//sends all existing games + status
void C_Manager::NewPlayer(C_Socket *i_pclSocket)
{
	int i, iSize = (int)m_pclGameList->size();
	C_Player *pclPlayer;
	i_pclSocket->SetNoDelay();
	pclPlayer = new C_Player(i_pclSocket);
	m_pclPlayerList->push_back(pclPlayer);

	C_NetMsgServer *pclNetMsg = pclPlayer->GetNetMsg();

	//sends all games to player
	S_NetMsg2 stMsg2;
	stMsg2.iType=N_MSG2_ADDGAME;
	for(i=0; i<iSize; i++) {
		C_Game *pclGame = (*m_pclGameList)[i];
		strcpy(stMsg2.szGameName, pclGame->GetGameName());
		stMsg2.iGameStatus = pclGame->GetStatus();
		pclNetMsg->SendMsg2(&stMsg2);
	}

	//log
	m_pclLog->Print(LOG_INFO, "Player %s connected\n", pclPlayer->GetIPString());
}

void C_Manager::AddPlayerToGame(C_Player *i_pclPlayer, char *i_szGameName)
{
	C_Game *pclGame = SearchGame(i_szGameName);
	int iPlayerIndex;

	if(pclGame != NULL) {
		iPlayerIndex = pclGame->CheckPreState();
		if(iPlayerIndex != -1) {  //status ok, index returned
			C_Player *pclPlayer;
			int i, iSize = (int)m_pclPlayerList->size();
			for(i = 0; i<iSize; i++) {
				pclPlayer = (*m_pclPlayerList)[i];
				if(pclPlayer == i_pclPlayer) {
					m_pclPlayerList->erase(m_pclPlayerList->begin()+i);
					break;
				}
			}
			i_pclPlayer->SetIndex(iPlayerIndex);
			pclGame->JoinToGame(i_pclPlayer);

			if(pclGame->StatusChanged()) SendGameStatus(pclGame); //game full
			SendHiScore(pclGame, i_pclPlayer); //send hiscorelist to new player if the level is set
		} else { //not ok
			//send reply to player
			S_NetMsg2 stMsg2;
			stMsg2.iType = N_MSG2_JOIN_NEW_R;
			stMsg2.iPlayerId = -3; 
			stMsg2.iGameId = -1; //does not matter
			C_NetMsgServer *pclNetMsg = i_pclPlayer->GetNetMsg();
			pclNetMsg->SendMsg2(&stMsg2);
		}
	}
}

void C_Manager::CreateNewGame(C_Player *i_pclPlayer, char *i_szGameName)
{
	S_NetMsg2 stMsg2;
	C_NetMsgServer *pclNetMsg;
	int iSize = (int)m_pclGameList->size();
	int iGameId = -1, iIndex = -1;  //gamename error

	//check if name exists
	if(SearchGame(i_szGameName) == NULL) iIndex = 0;

	if(iIndex == 0) { //game does not already exist
		if(iSize>MAX_GAMES) iIndex = -3; //too many games error
		else {
			//ok to create new
			C_Player *pclPlayer;
			int j, iSize2 = (int)m_pclPlayerList->size();
			for(j = 0; j<iSize2; j++) {
				pclPlayer = (*m_pclPlayerList)[j];
				if(pclPlayer == i_pclPlayer) {
					m_pclPlayerList->erase(m_pclPlayerList->begin()+j);
					break;
				}
			}
			i_pclPlayer->m_bIsMaster = true;

			iGameId = m_iGameIdCounter;
			C_Game *pclNewGame = new C_Game(iGameId, i_szGameName, i_pclPlayer, m_pclLog);

			m_pclGameList->push_back(pclNewGame);

			//send new game to everyone in player list
			stMsg2.iType = N_MSG2_ADDGAME;
			stMsg2.iGameId = iGameId;
			stMsg2.iGameStatus = 0;
			strcpy(stMsg2.szGameName, i_szGameName);

			SendToPlayerList(&stMsg2);

			m_iGameIdCounter++; //increase id for next game to add
		}
	}

	//send reply to player
	stMsg2.iType = N_MSG2_JOIN_NEW_R;
	stMsg2.iPlayerId = iIndex;
	stMsg2.iGameId = iGameId;
	pclNetMsg = i_pclPlayer->GetNetMsg();
	pclNetMsg->SendMsg2(&stMsg2);

	//send addplayer to the player if all ok
	if(iIndex >= 0) {
		stMsg2.iType = N_MSG2_ADDPLAYER;
		stMsg2.iPlayerId = iIndex;
		strcpy(stMsg2.szPlayerName, i_pclPlayer->GetName());
		pclNetMsg = i_pclPlayer->GetNetMsg();
		pclNetMsg->SendMsg2(&stMsg2);
	}
}

void C_Manager::SendGameStatus(C_Game *i_pclGame)
{
	S_NetMsg2 stMsg2;

	//send gamestatus to all in player list
	stMsg2.iType = N_MSG2_GAMESTATUS;
	stMsg2.iGameStatus = i_pclGame->GetStatus();
	strcpy(stMsg2.szGameName, i_pclGame->GetGameName());
	SendToPlayerList(&stMsg2);
}

void C_Manager::SendRemoveGame(C_Game *i_pclGame)
{
	S_NetMsg2 stMsg2;

	//send removegame to all in player list
	stMsg2.iType = N_MSG2_REMOVEGAME;
	strcpy(stMsg2.szGameName, i_pclGame->GetGameName());
	SendToPlayerList(&stMsg2);

	int i, iSize = (int)m_pclGameList->size();
	for(i=0; i<iSize; i++) {
		C_Game *pclGame = (*m_pclGameList)[i];
		if(i_pclGame == pclGame) {
			m_pclGameList->erase(m_pclGameList->begin()+i);
			delete pclGame;
			break;
		}
	}
}

void C_Manager::SendToPlayerList(S_NetMsg2 *i_pstMsg)
{
	C_Player *pclPlayer;
	C_NetMsgServer *pclNetMsg; 
	int i, iSize = (int)m_pclPlayerList->size();

	//send gamestatus to all in player list
	for(i=0; i<iSize; i++) {
		pclPlayer = (*m_pclPlayerList)[i];
		pclNetMsg = pclPlayer->GetNetMsg();
		pclNetMsg->SendMsg2(i_pstMsg);
	}
}

void C_Manager::SendHiScore(C_Game *i_pclGame, C_Player *i_pclPlayer)
{
	int j;
	C_Player *pclPlayer;
	C_NetMsgServer *pclNetMsgServer;
	S_HiScore stHiScore;

	char szFileName[MAX_PATH];
	char *szLevel = i_pclGame->GetLevelName();
	if(!szLevel || szLevel[0]==0) return; //no level set
	sprintf(szFileName, "%s/%s.hiscore", C_Global::szUserPath, szLevel);
	C_HiScoreList *pclHiScore = new C_HiScoreList(szFileName, i_pclGame->GetLevelType()!=MAP_RACE, NULL);
	int iResult = pclHiScore->GetAsData(&stHiScore);
	delete pclHiScore;

	if(!iResult) return; //there is no list yet

	if(i_pclPlayer) {
		//send to player
		pclNetMsgServer = i_pclPlayer->GetNetMsg();
		pclNetMsgServer->SendHiScoreMsg(&stHiScore);
	} else {
		//send to players in a game
		for(j=0; j<8; j++) {
			pclPlayer = i_pclGame->GetPlayer(j);
			if(pclPlayer) {
				pclNetMsgServer = pclPlayer->GetNetMsg();
				pclNetMsgServer->SendHiScoreMsg(&stHiScore);
			}
		}
	}
}

void C_Manager::HandleScore(C_Game *i_pclGame, C_Player *i_pclPlayer, int i_iScore)
{
	int iLevelType = i_pclGame->GetLevelType();

	//no hiscore for dogfight
	if(iLevelType==MAP_DOGFIGHT) return;

	bool bResult;
	char szFileName[MAX_PATH];
	char *szLevel = i_pclGame->GetLevelName();
	if(!szLevel || szLevel[0]==0) return; //no level set
	sprintf(szFileName, "%s/%s.hiscore", C_Global::szUserPath, szLevel);

	if(iLevelType!=MAP_MISSION_COOP || i_pclPlayer->m_bIsMaster) { //only master is allowed to add hiscore in MISSION_COOP
		C_HiScoreList *pclHiScore = new C_HiScoreList(szFileName, iLevelType!=MAP_RACE, &bResult);
		if(!bResult) { //file did not exist, set description
			strcpy(szFileName, i_pclGame->GetLevelName());
			if(iLevelType==MAP_RACE) strcat(szFileName, "   (race time)");
			else strcat(szFileName, "   (score)");
			pclHiScore->SetLevelName(szFileName);
		}

		if(pclHiScore->Add(i_pclPlayer->GetName(), i_iScore)) SendHiScore(i_pclGame, NULL);
		delete pclHiScore;
	}
}

//helper to search for game
C_Game *C_Manager::SearchGame(char* i_szGameName)
{
   int i, iSize = (int)m_pclGameList->size();

   for(i=0; i<iSize; i++) {
      C_Game *pclGame = (*m_pclGameList)[i];
      if(pclGame->CompareGameName(i_szGameName))
       return pclGame;
   }

   return NULL;
}

//info functions
void C_Manager::ResetCurGame()
{
	m_iCurGame = 0;
}

C_Game *C_Manager::GetNextGame()
{
	int iSize = (int)m_pclGameList->size();
	if(m_iCurGame>=iSize) return NULL;

	C_Game *pclGame = (*m_pclGameList)[m_iCurGame++];
	return pclGame;
}

char *C_Manager::GetGameString(C_Game *i_pclGame)
{
	if(!i_pclGame) return NULL;

	strcpy(m_szTempString, i_pclGame->GetGameName());
	if(i_pclGame->GetPlaying()) strcat(m_szTempString, " (playing)");
	else strcat(m_szTempString, i_pclGame->GetStatus()?" (full)":" (open)");
	return m_szTempString;
}

char *C_Manager::GetPlayerString(C_Game *i_pclGame)
{
	m_szTempString[0] = 0; //reset result

	if(!i_pclGame) return NULL;

	int iId = 0;
	int i, iSize = i_pclGame->GetNumPlayers();
	if(iSize == 0) return NULL;
	for(i=0; i<8; i++) {
		C_Player *pclPlayer = i_pclGame->GetPlayer(i);
		if(pclPlayer) {
			strcat(m_szTempString, pclPlayer->GetName());
			if(iId!=iSize-1) {
				if(iId==3) strcat(m_szTempString, "\n"); //break line after no 4
				else strcat(m_szTempString, ", ");
			}
			iId++;
		}
	}
	return m_szTempString;
}

void C_Manager::ResetCurUnnamedPlayer()
{
	m_iCurPlayer = 0;
}

char *C_Manager::GetNextUnnamedPlayerIPString()
{
	int iSize = (int)m_pclPlayerList->size();

	if(m_iCurPlayer>=iSize) return NULL;
	C_Player *pclPlayer = (*m_pclPlayerList)[m_iCurPlayer];
	sprintf(m_szTempString, "%s", pclPlayer->GetIPString());
	m_iCurPlayer++;

	return m_szTempString;
}
